// EMData.h - Steve Ludtke  6/97
// This is the heart of the EMAN project. It does just about all of the
// hard stuff. It reads and writes EM images in various formats, does FFTs,
// filters, etc...
#ifndef EMDATAH
#define EMDATAH

#include <string.h>
#include <unistd.h>
#include "mrc.h"
#include "imagic.h"
#include "spider.h"
#include "util.h"
#include "List.h"
#include "Euler.h"
#include "rfftw.h"

#define EMDATA_COMPLEX	1
#define EMDATA_RI		(1<<1)		// real/imaginary or amp/phase
#define EMDATA_BUSY		(1<<2)		// someone is modifying data
#define EMDATA_SHARED	(1<<3)		// Stored in shared memory
#define EMDATA_SWAPPED	(1<<4)		// Data is swapped (may be offloaded if memory is tight)
#define EMDATA_NEWFFT	(1<<5)		// Data has changed, redo fft
#define EMDATA_HASCTF	(1<<6)		// has CTF info
#define EMDATA_NEEDUPD	(1<<7)		// needs a realupdate()
#define EMDATA_NEEDHIST	(1<<8)		// histogram needs update
#define EMDATA_NEWRFP	(1<<9)		// needs new rotational footprint
#define EMDATA_NODATA	(1<<10)		// no actual data

#define EMDATA_CHANGED	(EMDATA_NEWFFT+EMDATA_NEEDUPD+EMDATA_NEEDHIST+EMDATA_NEWRFP)

#define EMDATA_FLIP		(1<<11)		// a flag only

// used to get statistics on a set of numbers
class floatStat {
int nf,n;
float *v;

public:
floatStat(int s) { nf=s; v=(float *)malloc(nf*sizeof(float)); n=0; }
~floatStat() { free(v); }

void clear() { n=0; }
void add(float f) { 
	if (n>=nf) { nf+=3; v=(float *)realloc(v,nf*sizeof(float)); }
	v[n]=f; 
	n++; 
}
inline void qadd(float f) {
	v[n]=f; 
	n++; 
}
float mean() { float rt=0; for (int i=0; i<n; i++) rt+=v[i]; return rt/(float)n; }
float min() { float rt=1.0e38; for (int i=0; i<n; i++) if (v[i]<rt) rt=v[i]; return rt; }
float max() { float rt=-1.0e38; for (int i=0; i<n; i++) if (v[i]>rt) rt=v[i]; return rt; }
float median() {
	int i,j;
	float s;
	for (i=0; i<=n/2; i++) {
		for (j=i+1; j<n; j++) {
			if (v[i]<v[j]) { s=v[i]; v[i]=v[j]; v[j]=s; }
		}
	}
	if (n%2) return v[n/2];
	return (v[n/2]+v[n/2-1])/2.0;
}

};

int fileCount(const char *filespec,int *type=NULL);	// number of images in a file, should be static
List *readImages(const char *filespec,int n0,int n1,int nodata=0);	// reads a set of images, should be static
SPIDERH *readSpiHed(const char *fsp);

class EMData
{
int flags;
int rocount;
float *rdata;
float *supp;
int nx,ny,nz;		// image size
float dx,dy,dz;		// translation from original image
Euler euler;		// orientation of slice or model
//float daz,dax,daz2;	// angles from original image
int nimg;			// when averaging, # images in average

float min,max;		// image density limits
float mean,sigma;	// image stats
float mean2,sigma2;	// these are statistics for a 4 pixel strip at the edge of the image

float pixel;		// A/pixel (or voxel)

char ptype[4];		// type of parameter sets
float param[4][25];	// parameters

float hist[200];	// image histogram
float histmin,histmax;

long atime;			// access time
int tmp;			// # of temp file when swapped out

mrcH *mrch;			// MRC header when read from MRC file
imagicH *imagich;	// imagic header when read from imagic file
SPIDERH *spiderH;	// spider header when read from spider file

char path[MAXPATH];	// path image was read from
int pathnum;			// image number in file pointed to by path
char name[80];			// descriptive name

EMData *parent;		// if this image was generated from another, this is it
EMData *fft;		// this avoids doing multiple ffts of the same set
EMData *rfp;		// avoids multiple rotational footprints

static List *ralfp;		// rotational alignment basis images
static int rali,ralo,rals;	// rotational alignment inner/outer rad and im size

float ctf[11];		// ctf parameters, check HASCTF flag before using

static int plans;
static int dimplan[6];
static int nplan[6][3];
static int fwplan[6];
static rfftwnd_plan plan[6];

// callback for progress display in long routines, if return!=0, routine should abort
int (*updProgress)(int num,int denom,const char *msg);

void realupdate();

public:
union { int unum; float ufloat; };
union { int unum2; float ufloat2; };

EMData();
~EMData();
void *operator new(size_t size);
void operator delete(void *ptr);

static rfftwnd_plan makeplan(int,int,int,int,int fw,int measure=0);	// generally for internal use in FFT

void setProgCB(int (*cb)(int num,int denom,const char *msg)) { updProgress = cb; }

// general creation
int readImage(const char *filespec, int n,int nodata=0);			
										// reads a single image, 0 on sucess
										// n<0 -> read image as 3d data if possible

EMData *copy(int withfft=0,int withparent=1);		// returns an exact copy of itself
EMData *copyHead();	// copies descriptive info, but not data
EMData *clip(int x0,int y0,int w,int h);		// inclusive clip. Pads if larger than data
EMData *clip(int x0,int y0,int z0,int w,int h,int d);

// read routines for specific formats
int readSAL(const char *filespec,int nodata=0);	//scans-a-lot
int readPGM(const char *filespec,int nodata=0);	//portable greymap (and other pix formats)
int readMRC(const char *filespec,int nodata=0);	
int readIMAGIC(const char *filespec, int n=0,int nodata=0);
int readIMAGIC3d(const char *filespec,int nodata=0);
int readEMIM(const char *filespec, int n=0,int nodata=0);
int readSPIDER(const char *filespec,int n=0,int nodata=0);
int readMRCArea(const char *fsp,int x0,int y0,int w,int h);

// write routines for specific formats
int writeIMAGIC(const char *filespec, int n,int nolock=0);	// n=-1 appends 
								// nolock for internal use only!
int writeMRC(const char *filespec);
int writePGM(const char *filespec,float gmin,float gmax);
int writeSPIDER(const char *filespec, int n);	// n=-1 appends
int writeEMIM(const char *filespec, int n);	// n=-1 appends
int writeVTK(const char *filespec);

// update statistics
void calcHist(float hmin,float hmax);
void realCalcHist();
float *getHist();
void update();

// general parameter set/gets
void setPixel(float pix);
float Pixel();
float Min();
float Max();
float Mean();
float Sigma();
float Mean2();
float Sigma2();
float edgeMean();
float cirMean();
float histMin();
float histMax();
void setParent(EMData *data);
EMData *Parent();
void setName(const char *name);
const char *Name();
void setPath(const char *path);
const char *Path();
void setPathN(int n);
int xSize();
int ySize();
int zSize();
int setSize(int x,int y=1, int z=1);
void setRI(int i);		// sets/resets the RI flag, does not modify data
						// use ap2ri() for conversion
void setComplex(int i);
int isComplex();
int hasCTF();
int getFlags();
void setFlipped(int f);	// !!!! DOESN'T FLIP IMAGE !!!!
float Dx();
float Dy();
float Dz();
float alt();
float az();
float phi();
Euler *getEuler();
void setTAlign(float x,float y,float z=0);
void setRAlign(float Alt,float Az=0,float Phi=0);
void setRAlign(Euler *angle);
void setNImg(int n);
int NImg();
long getATime();

imagicH *getImagicHed();
mrcH *getMRCHed();
SPIDERH *getSpiHed();

// Access to raw data, swapping
float *getData();	// This returns a pointer to the raw float data
float *getDataRO();	// Returns data pointer for reading only
void doneData();	// MUST be called when the data is no longer being
					// modified. Another getData while the data is
					// out will return NULL
int Busy();			// Returns 1 if data is checked out
int waitReady(int ro);	// waits for data to be available
void Wait();		// Why 2 ?    don't recall...
int swapout();		// offloads the data to disk when mem is tight
int swapin();		// reloads the data when necessary

void zero();		// clears the image to zero
void one();			// makes a 'unit' image

// Rendering
int renderAmp8( unsigned char * data , int x , int y, int xsize, int ysize,
	int bpl, float scale, int mingray, int maxgray,float renMin, float renMax);
int renderPha24( unsigned char * data , int x , int y , int xsize , int ysize , 
	int bpl , float scale, float renMin, float renMax);

// for complex images
void ri2ap();			// converts real/imaginary to amp/pha (complex only)
void ap2ri();			// the opposite

// fourier ops
EMData *doFFT();		// obvious, note that result is initially R,I not A,P
EMData *doIFT();		// obvious
void gimmeFFT();		// This gives the caller ownership of the
						// recently obtained fft so it won't be freed automatically
						// this means a new fft will be calculated next time

// wavelet ops
void doDWT(int basis, int level);	// in place DWT, basis/level defined elsewhere
void doIWT(int basis, int level);	// inverse wavelet x-form
void DWTFilt(int basis, int level, float thresh);
						// in place nonlinear threshold wavelet filter
						// thresh is with respect to the 'ideal' threshold

EMData *calcCCF(EMData *with);	// cross correlation between 2 images
								// mirror ACF if with=NULL

// if parent is set, each new interpolation will be done from the
// original parent's data, and additional calls to *Align will modify
// dx,dy,daz cumulatively. With no parent dx,dy,daz are reset each time

// translational alignment using CCF
float transAlign(EMData *with,int useparent,int intonly,int maxshift,float *peak=NULL);

// rotational alignment using angular correlation
float rotAlign(EMData *with,int irad=8,int orad=0);	

// rotational alignment via circular harmonic
float rotAlignCH(EMData *with,int irad=8,int orad=0);	

// rotation alignment with translational independance, +180 degree ambiguity
float rotAlignTI(EMData *with,int usedot=1);	

// translational alignment using CCF
float transAlign3d(EMData *with,int useparent=0,int intonly=0);	

// rotational, translational and flip alignment
EMData *RTFAlign(EMData *with,EMData *flip=NULL,int usedot=1,int maxshift=0);

// rotational and flip alignment
EMData *RFAlign(EMData *with,EMData *flip=NULL,int irad=8,int orad=0);

// rotational, translational alignment used by RTF
EMData *RTAlign(EMData *with,int usedot=0,int maxshift=0);


void refineAlign(EMData *with,int usedot=1);

EMData *makeRFP();

void calcRCF(EMData *with,float *sum,int NS);



void rotateAndTranslate(float scale=1.0,float dxc=0,float dyc=0,float dzc=0);		
								// rotate and translate using current settings
								// behavior changes if parent is/not set

void rotateAndTranslateFast(float scale=1.0);		
								// rotate and translate using current settings
								// behavior changes if parent is/not set

void cmCenter(int intonly=1);				// center at center of mass, ignores old dx,dy

// other maniuplations
void applyMask(int r,int type);	// applies a circular mask to the data
								// type=0 is a step cutoff to the mean value
								// type=4 is a step cutoff to zero
								// type=1 fills in with flatband random noise
									
void applyRadFn(int n,float x0,float dx,float *y,int interp=1);
								// multiplies by a radial function in
								// fourier space

void addRadNoise(int n,float x0,float dx,float *y,int interp);	
								// adds random noise with sigma defined
								// by the radial function

void subtract(EMData *data);	// subtracts 'data' from the current set

void vFlip();					// flips image vertically

void normalize();		// mean -> 0, std dev -> 1

void normalizeTo(EMData *to);	// This will multiply 'this' by a constant so
								// it is scaled to the signal in 'to'

void normalizeMax();		// mean -> 0, std dev -> 1

void invert();			// multiply by -1

void edgeNormalize(int cir=0);	// normalizes so mean -> value at edge
								// of image, if cir=1, uses 2 pixel circular
								// border 

void calcRadDist(int n,float x0,float dx,float *y);
void calcRadDist(int n,float x0,float dx,float *y,float acen,float amwid);
						// calculates radial distribution, works for real
						// and imaginary images. x must be defined on input
						// and y must be allocated. amin,amax are in radians 

void calcAzDist(int n,float a0,float da,float *d,float rmin,float rmax);

void radialAverage();	// makes image circularly symmetric

void radialSubtract();	// subtracts circularly symmetric part of image

void toCorner();	// Translates a centered image to the corner

float dot(EMData *data,int evenonly=0);	// dot product of 2 images (same size !)

float lcmp(EMData *data);	// linear comparison of 2 data sets (smaller better)

void realFilter(int type, float val1=0,float val2=0, float val3=0);
						//a variety of real-space filters

void filter(float highpass,float lowpass,int type);
						// fourier filters an image, real or imaginary
void makeAverage(List *in,EMData *sigma=NULL);
						// averages the images in the list, result in 'this'
						// optionally makes a sigma image as well

void makeAverageIter(List *in);

void makeAverageCTFC(List *in,float filtr);
						// averages the images in the list with CTF correction
						// filtr is in angstroms,

void makeMedian(List *in);
						// Median of the images in the list, result in 'this'

void add(EMData *in);	// adds 'in' to 'this'

void addIncoherent(EMData *in);	// adds 'in' to 'this'
						
void mult(EMData *in);	// multiplies 'in' by 'this'

void multConst(float);	// multiply by constant

void medianShrink(int i);	// reduces the size of the image by a factor of i
							// using a local median filter

void commonLines(EMData *d1,EMData *d2,int mode=0);	
				// finds common lines between 2 complex images
				// mode 0 is a summed dot-product 
				// mode 1 is weighted phase residual

EMData *project3d(float alt,float az, float phi, int mode, float thr=1.0);
				// makes a 3d projection of a volume, mode=-1 does real space
				// projection, 1-5 use fftslice for fourier projection
				// see EMData.h for other important notes

EMData *fftSlice(float alt,float az,float phi,int mode=5);		
													// takes a slice from a 3d complex image
													// for generating projections
// Supplementary functions for fftSlice
void setup4Slice(int redo=1);
void interpFT3d(float x,float y,float z,float *ret,int mode);

// this initializes a new image to receive 'insertSlice' operations
// for fourier volume inversion. If the image has a parent, it will also
// be affected. WARNING: will use 8*size^3 bytes of memory.
void setup4IS(int size);

// This will try to make sure the slice is normalized properly
float normSlice(EMData *slice,float alt,float az,float phi,float *phaseresid=NULL);

// this inserts a complex slice into an image that has been setup4IS()
void insertSlice(EMData *slice,  float alt,float az,float phi,int mode=5,float weight=1.0);

// once all of the slices have been added with insertSlice(), this routine
// does a normalization and returns a real 3d volume
void doneSlice();

float *ctfCurve(int type=0,int nsf=0,float *S=NULL,float *SF=NULL);	
	// calculates a CTF curve
	// optional structure factor data for true SNR calc
float *getCTF();
void setCTF(float *c);

void subNoise();

// This does an IFT of a volume reconstructed from slices. Includes a
// real-space correction for the linear interpolation in Fourier space
EMData *iftSlice();

// This gets a single value with no data locking. May be risky!
float valueAt(int x, int y, int z);
float valueAt(int x, int y);
};

inline float EMData::valueAt(int x, int y, int z) { return rdata[x+y*nx+z*nx*ny]; }
inline float EMData::valueAt(int x, int y) { return rdata[x+y*nx]; }
inline float *EMData::getCTF() { return ctf; }
inline void EMData::setCTF(float *f) { flags|=EMDATA_HASCTF; memcpy(ctf,f,11*sizeof(float)); return; }
inline void EMData::setRI(int i) { if (i) flags|=EMDATA_RI; else flags&=~EMDATA_RI; }
inline void EMData::setComplex(int i) { if (i) flags|=EMDATA_COMPLEX; else flags&=~EMDATA_COMPLEX; }
inline int EMData::isComplex() { if (flags&EMDATA_COMPLEX) return 1; else return 0; }
inline void EMData::setFlipped(int f) { if (f) flags|=EMDATA_FLIP; else flags&=~EMDATA_FLIP; }
inline int EMData::getFlags() { return flags; }
inline int EMData::hasCTF() { if (flags&EMDATA_HASCTF) return 1; else return 0; }
inline long EMData::getATime() { return atime; }
inline void EMData::setPixel(float pix) { pixel=pix; }
inline float EMData::Pixel() { return pixel; }
inline void EMData::update() { flags|=EMDATA_CHANGED; }
inline float EMData::Min() { if (flags&EMDATA_NEEDUPD) realupdate(); return min; }
inline float EMData::Max() { if (flags&EMDATA_NEEDUPD) realupdate(); return max; }
inline float EMData::Mean() { if (flags&EMDATA_NEEDUPD) realupdate(); return mean; }
inline float EMData::Mean2() { if (flags&EMDATA_NEEDUPD) realupdate(); return mean2; }
inline float EMData::Sigma() { if (flags&EMDATA_NEEDUPD) realupdate(); return sigma; }
inline float EMData::Sigma2() { if (flags&EMDATA_NEEDUPD) realupdate(); return sigma2; }
inline float EMData::histMin() { return histmin; }
inline float EMData::histMax() { return histmax; }
inline mrcH *EMData::getMRCHed() { return mrch; }
inline imagicH *EMData::getImagicHed() { return imagich; }
inline SPIDERH *EMData::getSpiHed() { return spiderH; }
inline void EMData::setParent(EMData *data) { parent=data; }
inline EMData *EMData::Parent() { return parent; }
inline void EMData::setName(const char *n) { strncpy(name,n,79); name[79]=0; }
inline const char *EMData::Name() { return name; }
inline void EMData::setPath(const char *p) { strncpy(path,p,MAXPATH-1); path[MAXPATH-1]=0; } 
inline void EMData::setPathN(int n) { pathnum=n; }
inline const char *EMData::Path() { return path; }
inline int EMData::xSize() { return nx; }
inline int EMData::ySize() { return ny; }
inline int EMData::zSize() { return nz; }
inline float EMData::Dx() { return dx; }
inline float EMData::Dy() { return dy; }
inline float EMData::Dz() { return dz; }
inline float EMData::phi() { return euler.phi(); }	// daz
inline float EMData::alt() { return euler.alt(); }	// dax
inline float EMData::az() { return euler.az(); }	// daz2
inline Euler *EMData::getEuler() { return &euler; }
inline void EMData::setTAlign(float x,float y,float z) { dx=x; dy=y; dz=z; }
//inline void EMData::setRAlign(float z,float x,float z2) { euler.setAngle(x,z2,z); }
inline void EMData::setRAlign(float Alt,float Az,float Phi) { euler.setAngle(Alt,Az,Phi); }
inline void EMData::setRAlign(Euler *ang) { euler=*ang; }
inline void EMData::setNImg(int n) { nimg=n; }
inline int EMData::NImg() { return nimg; }
inline void EMData::Wait() { while (flags&EMDATA_BUSY) sleep(1); }
inline float *EMData::getHist() { if (flags&EMDATA_NEEDHIST) realCalcHist(); return hist; }
inline void EMData::calcHist(float hmin,float hmax) { 
	flags|=EMDATA_NEEDHIST; 
	histmin=hmin;
	histmax=hmax;
}

#endif
