// EMData.C - Steve Ludtke  6/97
// This is the heart of the EMAN project. It does just about all of the
// hard stuff. It reads and writes EM images in various formats, does FFTs,
// filters, etc...
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "List.h"
#include "mrc.h"
#include "imagic.h"
#include "emim.h"
#include "util.h"
#include "EMData.h"

//#define DEBUG


// subtract 2 angles (radians)
inline float asub(float x,float y) {
float r;
r=fabs(x-y);
if (r>PI) r=2.0*PI-r;
return r;
}

// subtract 2 angles (radians)
inline float asub2(float x,float y) {
float r;

r=fmod(fabs(x-y),PI);
if (r>PI/2.0) r=PI-r;
return r;
}

float acmp(float a0,float a1,float a2) {
if (a0<=a1 && a1<=a2) return 1;
if (a0<=a1+PI && a1+PI<a2) return 1;
if (a0<=a1+2.0*PI && a1+2.0*PI<a2) return 1;
if (a0<=a1-PI && a1-PI<a2) return 1;
if (a0<=a1-2.0*PI && a1-2.0*PI<a2) return 1;
return 0;
}



//#define PI	3.141592654		// silly, but sometimes there are problems with M_PI

// This is used to define the width of the gaussian for 5x5x5 interpolation
// in 3d slices
#define I5G	(10.4/(PI*PI))	// used for 5x5x5 interpolation
#define I4G	(8.8/(PI*PI))	// used for 4 interpolation
#define I3G	(6.4/(PI*PI))	// used for 3 and 5x5x5 interpolation
#define I2G	(4.0/(PI*PI))	// 2 interpolation


List *EMData::ralfp=NULL;
int EMData::rali=0;
int EMData::ralo=0;
int EMData::rals=0;

rfftwnd_plan EMData::plan[6] = { NULL,NULL,NULL,NULL,NULL,NULL };
int EMData::dimplan[6] = { 0,0,0,0,0,0 };
int EMData::fwplan[6] = { 0,0,0,0,0,0 };
int EMData::nplan[6][3] = { 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0, 0 , 0 , 0 , 0 , 0 , 0 };
int EMData::plans=0;

EMData::EMData()
{
int i;

nx=ny=nz=0;
rdata=NULL;
dx=dy=dz=0;
euler.setAngle(0,0,0);
min=max=mean=sigma=0;
pixel=4.667;
for (i=0; i<200; i++) hist[i]=0;
histmin=histmax=0;
mrch=NULL;
imagich=NULL;
spiderH=NULL;
path[0]=0;
pathnum=0;
name[0]=0;
parent=NULL;
fft=NULL;
rocount=0;
rfp=NULL;
flags=EMDATA_NEEDUPD|EMDATA_NEEDHIST;
supp=NULL;
ufloat=ufloat2=0;
updProgress=NULL;

if (!ralfp) {
	ralfp=new List;
	rali=ralo=rals=0;
}
}

EMData::~EMData()
{
if (rdata) Sfree(rdata,1);
if (mrch) Sfree(mrch,1);
if (spiderH) free(spiderH);
if (imagich) Sfree(imagich,1);
if (fft) delete fft;
if (rfp) delete rfp;
if (supp) free(supp);
}

void *EMData::operator new(size_t size)
{
return Smalloc(sizeof(EMData),1);
}

void EMData::operator delete(void *ptr)
{
Sfree(ptr,1);
}

int EMData::swapout()
{
char s[MAXPATH];
FILE *out;

if (flags&EMDATA_SWAPPED) return 0;

tmp=random();
sprintf(s,"/usr/tmp/tmp.%d",tmp);
out=fopen(s,"w");
if (!out) { error("ERROR: Cannot Swap",ERR_CRIT); return -1; }
if(fwrite(rdata,sizeof(float)*nx,ny*nz,out)!=ny*nz) {
	error("ERROR: Swap space full!",ERR_CRIT); 
	fclose(out);
	unlink(s);
	return -2;
}
fclose(out);
Sfree(rdata,1);
flags|=EMDATA_SWAPPED;
return 0;
}

int EMData::swapin()
{
char s[MAXPATH];
FILE *in;

if (!(flags&EMDATA_SWAPPED)) return 0;

sprintf(s,"/usr/tmp/tmp.%d",tmp);
in=fopen(s,"r");
if (!in) { error("ERROR: Swap file missing!",ERR_CRIT); return -1; }
rdata=(float *)Smalloc(nx*ny*nz*sizeof(float),1);
if (!rdata) { error("Cannot swap in, not enough memory!",ERR_CRIT); return -1; }
if(fread(rdata,sizeof(float)*nx,ny*nz,in)!=ny*nz) {
	error("ERROR: Swap file incomplete!",ERR_CRIT); 
	return -1;
}
flags&=~EMDATA_SWAPPED;
return 0;
}

// general creation

// returns number of images in a file, -1 if none
// type 1=spider, 2=imagic, 4=PGM/gif/etc, 0=MRC/other
int fileCount(const char *filespec,int *type)	
{
int i,l;
char *s,*v;
FILE *in;

if (!filespec) { printf("NULL filecount()\n"); return -1; }
l=strlen(filespec);
s=(char *)malloc(l+1);
v=(char *)&i;
strcpy(s,filespec);

// SPIDER file test
//if (!strcmp(filespec+l-4,".DAT") || !strcmp(filespec+l-4,".dat") ||
//	!strcmp(filespec+l-4,".spi") ) 
{
	SPIDERH *sh=readSpiHed(filespec);
	if (sh) {
		if ((sh->type==1.0 || sh->type==3.0) && sh->nslice<65536 &&
		sh->ny<65536) {
			if (type) *type=1;
			if (sh->istack) { int r=sh->maxim; free(sh); return r; }
			free(sh);
			return 1;
		}
		free(sh);
	}
}

// IMAGIC file test
if (!strcmp(filespec+l-4,".hed") || !strcmp(filespec+l-4,".img")) {
	s[l-3]=0;
	strcat(s,"img");
	if (access(s,R_OK)) i=0; else i=1;
	s[l-3]=0;
	strcat(s,"hed");
	if (!access(s,R_OK) && i) {		//ok, both an img and a hed, probably an IMAGIC file
		i=0;
		in=fopen(s,"r");
		fread(&i,4,1,in);
		fread(&i,4,1,in);
		fclose (in);
#ifdef MSB_FIRST
		if (v[0]!=0||v[1]!=0) swapint(v,1);
#else
		if (v[0]==0&&v[1]==0) swapint(v,1);
#endif
		i++;
		if (type) *type=2;
		return i;
	}
}
strcpy(s,filespec);

// check for EMIM
in=fopen(s,"r");
if (!in) return 0;
if (fread(&i,4,1,in)!=1) { fclose(in); return 0; }
if (!strncmp(v,"EMIM",4)) {
	fread(&i,4,1,in);
	if (i==2) l=0; else l=1;
	fread(&i,4,1,in);
	fclose(in);
	if (l) swapint(v,1);
	return i;
}
fclose(in);

if (!strcmp(filespec+l-4,".gif") || !strcmp(filespec+l-4,".tif") || 
	!strcmp(filespec+l-5,".tiff") || !strcmp(filespec+l-4,".pgm") ||
	!strcmp(filespec+l-5,".rgb") || !strcmp(filespec+l-4,".pnm") ||
	!strcmp(filespec+l-5,".ppm") || !strcmp(filespec+l-4,".tga")) {
		if (type) *type=4;
}
else if (type) *type=0;

// Other image types are 1 image only
return 1;
}

// reads a set of images
// This is really inefficient, but for now it will do...
List *readImages(const char *filespec,int n0,int n1,int nodata)	
{
int i,n;
List *rtn = new List();
EMData *data;

n=fileCount(filespec);
if (n==0) return NULL;
if (n0<0) n0=0;
if (n1>=n || n1<0) n1=n-1;

for (i=n0; i<=n1; i++) {
	data=new EMData();
	if (data->readImage(filespec,i,nodata)<0) { delete data; continue; }
	rtn->addObject(data);
}
return rtn;
}

int EMData::readImage(const char *filespec, int n,int nodata)			// reads a single image, 0 on sucess
{
int i,l,r;
char *s,*v;
FILE *in;

if (!filespec) { printf("NULL readimage()\n"); return(-1); }
l=strlen(filespec);
s=(char *)malloc(l+1);
v=(char *)&i;
strcpy(s,filespec);

// SPIDER file test
//if (!strcmp(filespec+l-4,".DAT") || !strcmp(filespec+l-4,".dat") ||
//	!strcmp(filespec+l-4,".spi") ) {
{
	SPIDERH *sh=readSpiHed(filespec);
	if (sh) {
		if ((sh->type==1.0 || sh->type==3.0) && sh->nslice<65536 &&
		sh->ny<65536) {
			free(sh); 
			r=readSPIDER(filespec,n,nodata);
			if (r) { setSize(10,10,1); zero(); }
			return r;
		}
		free(sh);
	}
}

// IMAGIC file test
if (!strcmp(filespec+l-4,".hed") || !strcmp(filespec+l-4,".img")) {
	s[l-3]=0;
	strcat(s,"img");
	if (access(s,R_OK)) i=0; else i=1;
	s[l-3]=0;
	strcat(s,"hed");
	if (!access(s,R_OK) && i) {		//ok, both an img and a hed, probably an IMAGIC file
		i=0;
		in=fopen(s,"r");
		fread(&i,4,1,in);
		fread(&i,4,1,in);
		fclose (in);
#ifdef MSB_FIRST
		if (v[0]!=0) swapint(v,1);
#else
		if (v[0]==0) swapint(v,1);
#endif
		if (n>i) return(-1);
		if (n<0) {
			r=readIMAGIC3d(s,nodata);
			if (r) { setSize(10,10,1); zero(); }
			return r;
		}	
		r=readIMAGIC(s,n,nodata);
		if (r) { setSize(10,10,1); zero(); }
		return r;
	}
}
strcpy(s,filespec);

// SAL file test
if (!strcmp(filespec+l-4,".hdr") || !strcmp(filespec+l-4,".img") || 
	!strcmp(filespec+l-4,".HDR") || !strcmp(filespec+l-4,".IMG")) {
	s[l-3]=0;
	strcat(s,"img");
	if (access(s,R_OK)) i=0; else i=1;
	if (!i) {
		s[l-3]=0;
		strcat(s,"IMG");
		if (access(s,R_OK)) i=0; else i=2;
	}
	
	s[l-3]=0;
	if (i==2) strcat(s,"HDR");
	else strcat(s,"hdr");
	if (!access(s,R_OK) && i) {		//ok, both an img and a hdr, probably a SAL file
		i=0;
		in=fopen(s,"r");
		fread(&i,4,1,in);
		fread(&i,4,1,in);
		fclose (in);
#ifdef MSB_FIRST
		if (v[0]!=0) swapint(v,1);
#else
		if (v[0]==0) swapint(v,1);
#endif
		if (n>i) return(-1);
		r=readSAL(s,nodata);
		if (r) { setSize(10,10,1); zero(); }
		return r;
	}
}
strcpy(s,filespec);

if (!strcmp(filespec+l-4,".gif") || !strcmp(filespec+l-4,".tif") || 
	!strcmp(filespec+l-5,".tiff") || !strcmp(filespec+l-4,".pgm") ||
	!strcmp(filespec+l-5,".rgb") || !strcmp(filespec+l-4,".pnm") ||
	!strcmp(filespec+l-5,".ppm") || !strcmp(filespec+l-4,".tga")) {
	
	if (n>0) return(-1);
	r=readPGM(filespec,nodata);
	if (r) { setSize(10,10,1); zero(); }
	return r;
}

// check for EMIM
strcpy(s,filespec);
in=fopen(s,"r");
if (!in) return -2;
if (fread(&i,4,1,in)!=1) { fclose(in); return -2; }
if (!strncmp(v,"EMIM",4)) {
	fread(&i,4,1,in);
	if (i==2) l=0; else l=1;
	fread(&i,4,1,in);
	fclose(in);
	if (l) swapint(v,1);
	if (n>i) return(-1);
	return readEMIM(filespec,n,nodata);
}

// Ok, at this point we assume MRC
fclose(in);

r=readMRC(filespec,nodata);
if (r) { setSize(10,10,1); zero(); }
return r;
}

EMData *EMData::copy(int withfft,int withparent)		// returns an exact copy of itself
{
EMData *ret;
float *ndata;

getDataRO();
ret=new EMData;
ret->setSize(nx,ny,nz);
ret->setTAlign(dx,dy,dz);
*ret->getEuler()=euler;
ret->setNImg(nimg);
ret->setPixel(pixel);
ndata=ret->getData();
memcpy(ndata,rdata,nx*ny*nz*sizeof(float));
ret->doneData();
ret->calcHist(histmin,histmax);
ret->update();
if (withparent) ret->setParent(this); else ret->setParent(NULL);
ret->setName(name);
ret->setPath(path);
ret->setPathN(pathnum);
ret->unum=unum;
ret->fft=ret->rfp=NULL;
ret->flags=flags&(EMDATA_COMPLEX|EMDATA_RI|EMDATA_HASCTF);
ret->update();
if (fft && withfft) ret->fft=fft->copy();
memcpy(ret->ctf,ctf,11*sizeof(float));
doneData();
return ret;
}

EMData *EMData::copyHead()		// returns an exact copy of itself
{
EMData *ret;
float *ndata;

ret=new EMData;
ret->setSize(nx,ny,nz);
ret->setTAlign(dx,dy,dz);
*ret->getEuler()=euler;
ret->setPixel(pixel);
ret->update();
ret->setParent(this);
ret->setName(name);
ret->setPath(path);
ret->setPathN(pathnum);
ret->flags=flags&(EMDATA_COMPLEX|EMDATA_RI|EMDATA_HASCTF);
ret->fft=ret->rfp=NULL;
ret->setNImg(nimg);
memcpy(ret->ctf,ctf,11*sizeof(float));
ret->update();

return ret;
}

EMData *EMData::clip(int x0,int y0,int w,int h)		// clip. 0 Pads if larger than data
{
EMData *ret;
int i,j,ys0,ys1,yd0,xs0,xs1,xd0;
float *src,*dst;

getDataRO();
if (nz!=1) { error("Tried to 2d clip 3d image",ERR_ERROR); return NULL; }
ret=new EMData;
ret->setSize(w,h,1);
memset(dst=ret->getData(),0,w*h*sizeof(float));
ys0=y0<0?0:y0;
ys1=y0+h>ny?ny:y0+h;
yd0=y0<0?-y0:0;
xs0=x0<0?0:x0;
xs1=x0+w>nx?nx:x0+w;
xd0=x0<0?-x0:0;
src=rdata+ys0*nx+xs0;
dst+=yd0*w+xd0;
for (i=ys0; i<ys1; i++,src+=nx,dst+=w) {
	memcpy(dst,src,(xs1-xs0)*sizeof(float));
}
ret->doneData();
ret->setPixel(pixel);
ret->calcHist(histmin,histmax);
ret->update();
ret->setParent(NULL);
ret->setName(name);
ret->setPath(path);
ret->setPathN(pathnum);
*ret->getEuler()=euler;
ret->setNImg(nimg);
doneData();
return ret;
}

EMData *EMData::clip(int x0,int y0,int z0,int w,int h,int d)		// clip. Pads if larger than data
{
EMData *ret;
int i,j,jj,ys0,ys1,yd0,xs0,xs1,xd0,zs0,zs1,zd0;
float *src,*dst,*d2;

getDataRO();
ret=new EMData;
ret->setSize(w,h,d);
memset(dst=ret->getData(),0,w*h*d*sizeof(float));
zs0=z0<0?0:z0;
zs1=z0+d>nz?nz:z0+d;
zd0=z0<0?-z0:0;
ys0=y0<0?0:y0;
ys1=y0+h>ny?ny:y0+h;
yd0=y0<0?-y0:0;
xs0=x0<0?0:x0;
xs1=x0+w>nx?nx:x0+w;
xd0=x0<0?-x0:0;
for (j=zs0,jj=zd0; j<zs1; j++,jj++) {
	src=rdata+j*nx*ny+ys0*nx+xs0;
	d2=dst+jj*w*h+yd0*w+xd0;
	for (i=ys0; i<ys1; i++,src+=nx,d2+=w) {
		memcpy(d2,src,(xs1-xs0)*sizeof(float));
	}
}
doneData();
ret->doneData();
ret->setPixel(pixel);
ret->calcHist(histmin,histmax);
ret->update();
ret->setParent(NULL);
ret->setName(name);
ret->setPath(path);
ret->setPathN(pathnum);
return ret;
}


// read routines for specific formats
int EMData::readSAL(const char *filespec,int nodata)		//scans-a-lot
{
FILE *in;
char s[81],s2[10],*t;
int i,j,rl=512,bl,sm=0;
unsigned char *cdata;
short *sdata,sw;

setPath(filespec);
if (strrchr(filespec,'/')==NULL) setName("SAL File");
else setName(strrchr(filespec,'/')+1);

nx=ny=-1;
nz=1;

pixel=4.6667;
sprintf((char *)(filespec+strlen(filespec)-4),".HDR");
if (access(filespec,R_OK)) sprintf((char *)(filespec+strlen(filespec)-4),".hdr");

i=0;
if ((in=fopen(filespec,"r"))==NULL) { sprintf(s,"Cannot open %s\n",filespec); error(s,ERR_ERROR); return -1; }
while (fgets(s,80,in)!=NULL) {
	if (s[0]==0) continue;
	if (strncmp(s+1,"NXP",3)==0) {
		sscanf(strchr(s,'=')+1," %d",&nx);
	}
	else if (strncmp(s+1,"NYP",3)==0) {
		sscanf(strchr(s,'=')+1," %d",&ny);
	}
	else if (strncmp(s+1,"AXSCAN",6)==0) {
		t=strrchr(s,'\'');
		if (t[-1]=='Y') i=1;
	}
	else if (strncmp(s+1,"FILE REC",8)==0) {
		sscanf(strchr(s,'=')+1," %d",&rl);
	}
	else if (strncmp(s+1,"SCAN TYPE",9)==0) {
		sscanf(strchr(s,'\'')+1," %s",s2);
		if (s2[0]=='R') sm=1;
	}
	else if (strncmp(s+1,"DELTAX",6)==0) {
		sscanf(strchr(s,'=')+1," %f",&pixel);
		pixel/=3.0;
	}
}	
fclose(in);

if (i) { i=nx; nx=ny; ny=i; }
sprintf(s,"%s is %dx%d",filespec,nx,ny);
error(s,ERR_INFO);
if (nodata) flags|=EMDATA_NODATA;
setSize(nx,ny,1);
cdata=(unsigned char *)rdata;
sdata=(short *)rdata;
if (nx==-1 || ny==-1) { error("Problem reading header.",ERR_ERROR); return -2; }

sprintf((char *)(filespec+strlen(filespec)-4),".IMG");
if (access(filespec,R_OK)) sprintf((char *)(filespec+strlen(filespec)-4),".img");

if ((in=fopen(filespec,"r"))==NULL) { printf("Cannot open %s\n",filespec); return -3; }

if (!nodata) {
bl=(((nx*sizeof(short)-1)/rl)+1)*rl;
//printf("Block length=%d x %d\n",bl,ny);
for (j=0; j<ny; j++) {
	if ((i=fread(&cdata[j*nx*nz*2],bl,nz,in))!=nz) {
		sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
		error(s,ERR_WARN);
	}
//	printf("Line %d is %d bytes\n",j,(((nx*sizeof(short)-1)/rl)+1)*rl);
}

// This reverses odd numbered rows in RASTER scan mode
if (sm) {
//	printf("Flipping data\n");
	for (i=1; i<ny; i+=2) {
		for (j=0; j<nx/2; j++) {
			sw=sdata[i*nx+j];
			sdata[i*nx+j]=sdata[i*nx+nx-j-1];
			sdata[i*nx+nx-j-1]=sw;
		}
	}
}

// Make the data float
for (i=nx*ny*nz-1; i>=0; i--)  
	rdata[i]=((float)((cdata[i*2+1]*256.0)+cdata[i*2])); 
}

parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(0,0,0);
atime=time(0);
update();
//calcHist(Min(),Max());
fclose(in);
return 0;
}

int EMData::readPGM(const char *fsp,int nodata)		//portable greymap (and other pix formats)
{
FILE *in;
int i,j,l,pipe=0;
float f,a,p,div=0;
char s[MAXPATH],*m2,w;
unsigned char *cdata,*cdata2,c;

#ifdef DEBUG
printf("PGM open '%s'\n",fsp);
#endif

in=fopen(fsp,"r");
if (in==NULL) { sprintf(s,"Cannot open file %s\n",fsp); zero(); return -1; }
if (getc(in)!='P'||getc(in)!='5') {
	fclose(in);
	pipe=1;
	m2=(char *)(fsp+strlen(fsp)-4);
	s[0]=0;
	if (!strcmp(m2,".gif") || !strcmp(m2,".GIF")) 
		sprintf(s,"giftopnm %s|ppmtopgm",fsp);
	else if (!strcmp(m2,".tif") || !strcmp(m2,".TIF") || !strcmp(m2-1,".tiff") )
		sprintf(s,"tifftopnm %s | ppmtopgm",fsp);
	else if (!strcmp(m2,".jpg") || !strcmp(m2,".JPG")) 
		sprintf(s,"djpeg -grayscale -pnm %s",fsp);
	else if (!strcmp(m2,".pcx") || !strcmp(m2,".PCX"))
		sprintf(s,"pcxtoppm %s| ppmtopgm",fsp);
	else sprintf(s,"ppmtopgm %s",fsp);

	in=popen(s,"r");
	if (getc(in)!='P'||getc(in)!='5') {
		pclose(in);
		error("Sorry, I do not recognize this file type.",ERR_ERROR);
		zero();
		return -2;
	}
}

if (strrchr(fsp,'/')==NULL) setName(fsp);
else setName(strrchr(fsp,'/')+1);
setPath(fsp);

nx=ny=0;
nz=1;
getc(in);
while (fgets(s,100,in)!=NULL && s[0]=='#') ;
sscanf(s," %d %d  %f",&nx,&ny,&div);
if (div!=255.0 && div!=0) printf("Warning, this could be an invalid pgm (%1.0f).\n",div);
while (getc(in)!=10);
if (nx<=0||ny<=0) { error("Invalid file\n",ERR_ERROR); zero(); return -1; }

#ifdef DEBUG
printf("Image is %dx%d\n",nx,ny);
#endif

if (nodata) flags|=EMDATA_NODATA;
setSize(nx,ny,1);

if (!nodata) {
cdata=(unsigned char *)rdata;

for (j=0; j<ny; j++) {
	if ((i=fread(&cdata[(ny-j-1)*nx*nz],nx,nz,in))!=nz) {
		printf("Incomplete data read %d/%d blocks\n",i,ny*nz);
		zero();
		return -1;
	}
}
if (div<1.0 || div>=65536.0) div=255.0;
for (i=nx*ny*nz-1; i>=0; i--) rdata[i]=(float)cdata[i]/div;

//for (i=nx*ny*nz-1; i>=0; i--) if (rdata[i]>1.0||rdata[i]<0) printf("%d,%g\n",i,rdata[i]);
}

if (pipe) pclose(in); else fclose(in);
pixel=4.66667;
parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(0,0,0);
atime=time(0);
update();
//calcHist(Min(),Max());
return 0;
}

int EMData::readMRC(const char *fsp,int nodata)
{
FILE *in;
int i,j,l,pipe=0;
int ord=0;	// 1 if the file order is MSB first
int mord=0; // 1 if the machine order is MSB first
float f,a,p;
char s[800],*m2,w;
unsigned char *cdata,*cdata2,c,t;
short *sdata;
static char *str[] = { "char", "short", "float", "short complex", "float complex" };
static char *str2[] = { "" , "byte reversed" , "" };
static char *str3[] = { "" , "compressed" };

#ifdef DEBUG
printf("MRC open '%s'\n",fsp);
#endif

if (!mrch) mrch=(mrcH *)Smalloc(sizeof(mrcH),1);

if(fsp==NULL) {in=stdin;}
else {
if (strcmp(&fsp[strlen(fsp)-3],".gz")==0 || strcmp(&fsp[strlen(fsp)-2],".Z")==0) {
	sprintf(s,"zcat %s",fsp);
	in=popen(s,"r");
	pipe=1;
}
else in=fopen(fsp,"r");
if (in==NULL) {
	error("Cannot open file",ERR_ERROR);
	zero();
	return -1;
}

if (strrchr(fsp,'/')==NULL) setName(fsp);
else setName(strrchr(fsp,'/')+1);
setPath(fsp);
}

if (!fread(mrch,sizeof(mrcH),1,in)) { setSize(10,10,1); zero(); return -1; }

// This swaps the byte order in the header on appropriate machines
#ifdef MSB_FIRST
mord=1;
#else
mord=0;
#endif

m2=(char *)mrch;
if (m2[0]==0 && m2[1]==0) ord=1;
if (mord ^ ord) {
	for (i=0; i<(27+MRC_USER)*4; i+=4) {
		w=m2[i]; m2[i]=m2[i+3]; m2[i+3]=w;
		w=m2[i+1]; m2[i+1]=m2[i+2]; m2[i+2]=w;
	}
}

if (mrch->ny>0xffff || mrch->nz>0xffff || mrch->mode>4 || mrch->nlabl>11) {
	setSize(10,10,1); printf("Invalid MRC file\n"); zero(); return -1;
}

//printf("ord=%d mord=%d\n",ord,mord);

if (mrch->ny==1) { printf("This appears to be a 1d file. I only read 2d files.\n"); zero(); return -4; }
//if (mrch->nz!=1) { printf("This appears to be a 3d file. I only read 2d files.\n"); return 0; }

if (mrch->mode==MODE_float_COMPLEX || mrch->mode==MODE_short_COMPLEX) mrch->nx*=2;
nx=mrch->nx;
ny=mrch->ny;
nz=mrch->nz;

if (mrch->nz==1) sprintf(s,"Read a %s %s %dx%d 2D %s MRC File\n",
	str2[ord+mord],str3[pipe],nx,ny,str[mrch->mode]);
else sprintf(s,"Read a %dx%dx%d 3D MRC File\n",nx,ny,nz);

if (mrch->nxstart!=0||mrch->nystart!=0||mrch->nzstart!=0) printf("WARNING: nx/y/z start not zero\n");

//if (header.mapc!=1 || header.mapr!=2 || header.maps!=3) printf("WARNING: map not 1,2,3\n");


if (mrch->xlen==0) mrch->xlen=1.0;
if (mrch->ylen==0) mrch->ylen=1.0;
if (mrch->zlen==0) mrch->zlen=1.0;

if (nodata) flags|=EMDATA_NODATA;

// This reads 2d&3d arrays
//if (header.nz==1) {
setSize(nx,ny,nz);

if (!nodata) {

	fseek(in,sizeof(mrcH)+mrch->nsymbt,SEEK_SET);

	
	getData();
	cdata=(unsigned char *)rdata;
	sdata=(short *)rdata;

	switch(mrch->mode) {
	case MODE_char:
		flags&=~EMDATA_COMPLEX;
		for (j=0; j<ny; j++) {
			if (updProgress && updProgress(j,ny,"Reading MRC file (8 bit)")) break;
			if ((i=fread(&cdata[j*nx*nz],nx,nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		for (i=nx*ny*nz-1; i>=0; i--) rdata[i]=(float)cdata[i]/100.0;
		if (updProgress) updProgress(0,0,NULL);
		break;
	case MODE_short:
		flags&=~EMDATA_COMPLEX;
#ifdef DEBUG
printf("Short file read\n");
#endif	
		for (j=0; j<ny; j++) {
			if (updProgress && updProgress(j,ny,"Reading MRC file (16 bit)")) break;
			if ((i=fread(&cdata[j*nx*nz*sizeof(short)],nx*sizeof(short),nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		if ((ord+mord)&1) {
			for (i=nx*ny*nz-1; i>=0; i--) {
				t=cdata[i*2];
				cdata[i*2]=cdata[i*2+1];
				cdata[i*2+1]=t;
				rdata[i]=(float)(sdata[i]);
			} 
		}
		else {
			for (i=nx*ny*nz-1; i>=0; i--)  {
				rdata[i]=(float)(sdata[i]); 
			}
		}
		if (updProgress) updProgress(0,0,NULL);

#ifdef DEBUG
printf("Short read complete\n");
#endif
		break;
	case MODE_float:
		cdata=(unsigned char *)rdata;
		flags&=~EMDATA_COMPLEX;
		for (j=0; j<ny; j++) {
			if (updProgress && updProgress(j,ny,"Reading MRC file (float)")) break;
			if ((i=fread(&rdata[j*nx*nz],nx*sizeof(float),nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		// swap order if machine and file have different order
		if (ord+mord==1) {
			for (i=nx*ny*nz*4-4; i>=0; i-=4) {
				c=cdata[i]; cdata[i]=cdata[i+3]; cdata[i+3]=c;
				c=cdata[i+1]; cdata[i+1]=cdata[i+2]; cdata[i+2]=c;
			}
		}
		if (updProgress) updProgress(0,0,NULL);
		break;
	case MODE_short_COMPLEX:
		cdata=(unsigned char *)rdata;
		flags|=EMDATA_COMPLEX|EMDATA_RI;
		for (j=0; j<ny; j++) {
			if (updProgress && updProgress(j,ny,"Reading MRC file (16 bit complex)")) break;
			if ((i=fread(&cdata[j*nx*nz*2],nx*sizeof(short),nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		if (ord) {
			for (i=nx*ny*nz-1; i>=0; i--) {
				rdata[i]=(float)((cdata[i*2]<<8)+cdata[i*2+1]); 
			}
		}
		else {
			for (i=nx*ny*nz-1; i>=0; i--) {
				rdata[i]=(float)((cdata[i*2+1]<<8)+cdata[i*2]); 
			}
		}
		if (updProgress) updProgress(0,0,NULL);
		break;
	case MODE_float_COMPLEX:
		cdata=(unsigned char *)rdata;
		flags|=EMDATA_COMPLEX|EMDATA_RI;
		clearerr(in);
		for (j=0; j<ny; j++) {
			if (updProgress && updProgress(j,ny,"Reading MRC file (float complex)")) break;
			if ((i=fread(&rdata[j*nx*nz],nx*sizeof(float),nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_ERROR);
			}
		}
		if (ord+mord==1) {
			for (i=0; i<nx*ny*nz; i+=4) {
				w=cdata[i*4+3];
				cdata[i*4+3]=cdata[i*4];
				cdata[i*4]=w;
				w=cdata[i*4+2];
				cdata[i*4+2]=cdata[i*4+1];
				cdata[i*4+1]=w;
			}
		}
		if (updProgress) updProgress(0,0,NULL);
		if (ferror(in)) { zero(); return -1; }
		break;
	default: 
		printf("Bad MRC file\n"); 
		zero(); 
		return -1;
	}
}

update();
if(fsp!=NULL){
	if (pipe) pclose(in); else fclose(in);
}

pixel=4.66667;
parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(0,0,0);
atime=time(0);
doneData();
update();
//calcHist(Min(),Max());
return 0;
}

// NOTE: This works ONLY for real 2d images. Also the area MUST reside within the image
int EMData::readMRCArea(const char *fsp,int x0,int y0,int w,int h)
{
FILE *in;
int i,j,l,pipe=0;
int ord=0;	// 1 if the file order is MSB first
int mord=0; // 1 if the machine order is MSB first
float f,a,p;
char s[800],*m2,ww;
unsigned char *cdata,*cdata2,c,t;
short *sdata;
static char *str[] = { "char", "short", "float", "short complex", "float complex" };
static char *str2[] = { "" , "byte reversed" , "" };
static char *str3[] = { "" , "compressed" };

#ifdef DEBUG
printf("MRC open '%s'\n",fsp);
#endif

if (!mrch) mrch=(mrcH *)Smalloc(sizeof(mrcH),1);

if (strcmp(&fsp[strlen(fsp)-3],".gz")==0 || strcmp(&fsp[strlen(fsp)-2],".Z")==0) {
	sprintf(s,"zcat %s",fsp);
	in=popen(s,"r");
	pipe=1;
}
else in=fopen(fsp,"r");
if (in==NULL) {
	error("Cannot open file (area)",ERR_ERROR);
	zero();
	return -1;
}

if (strrchr(fsp,'/')==NULL) setName(fsp);
else setName(strrchr(fsp,'/')+1);
setPath(fsp);

if (!fread(mrch,sizeof(mrcH),1,in)) { zero(); return -1; }

// This swaps the byte order in the header on appropriate machines
#ifdef MSB_FIRST
mord=1;
#else
mord=0;
#endif

m2=(char *)mrch;
if (m2[0]==0 && m2[1]==0) ord=1;
if (mord ^ ord) {
	for (i=0; i<(27+MRC_USER)*4; i+=4) {
		ww=m2[i]; m2[i]=m2[i+3]; m2[i+3]=ww;
		ww=m2[i+1]; m2[i+1]=m2[i+2]; m2[i+2]=ww;
	}
}

//printf("ord=%d mord=%d\n",ord,mord);

if (mrch->ny==1) { printf("This appears to be a 1d file. I only read 2d files.\n"); zero(); return 0; }
//if (mrch->nz!=1) { printf("This appears to be a 3d file. I only read 2d files.\n"); return 0; }

if (mrch->mode==MODE_float_COMPLEX || mrch->mode==MODE_short_COMPLEX) {
	printf("ERROR: cannot read a region of a complex image\n");
	zero();
	return -1; 
}
if (mrch->nz!=1) { printf("ERROR: cannot read a region from volume data\n"); zero(); return -1; }
if (x0<0 || x0+w>mrch->nx || y0<0 ||y0+h>mrch->ny) { 
	printf("ERROR: MRC box size extends beyond edge of image\n");
	zero();
	return -1;
} 

nx=w;
ny=h;
nz=1;

if (mrch->nz==1) sprintf(s,"Read a %s %s %dx%d 2D %s MRC File\n",
	str2[ord+mord],str3[pipe],nx,ny,str[mrch->mode]);
else sprintf(s,"Read a %dx%dx%d 3D MRC File\n",nx,ny,nz);

if (mrch->nxstart!=0||mrch->nystart!=0||mrch->nzstart!=0) printf("WARNING: nx/y/z start not zero\n");

//if (header.mapc!=1 || header.mapr!=2 || header.maps!=3) printf("WARNING: map not 1,2,3\n");


if (mrch->xlen==0) mrch->xlen=1.0;
if (mrch->ylen==0) mrch->ylen=1.0;
if (mrch->zlen==0) mrch->zlen=1.0;

// This reads 2d&3d arrays
//if (header.nz==1) {
setSize(nx,ny,nz);

	getData();
	cdata=(unsigned char *)rdata;
	sdata=(short *)rdata;

	switch(mrch->mode) {
	case MODE_char:
		flags&=~EMDATA_COMPLEX;
		for (j=0; j<ny; j++) {
			fseek(in,(j+y0)*mrch->nx+x0+sizeof(mrcH)+mrch->nsymbt,SEEK_SET);
			if ((i=fread(&cdata[j*nx*nz],nx,nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		for (i=nx*ny*nz-1; i>=0; i--) rdata[i]=(float)cdata[i]/100.0;
		break;
	case MODE_short:
		flags&=~EMDATA_COMPLEX;
#ifdef DEBUG
printf("Short file read\n");
#endif	
		for (j=0; j<ny; j++) {
			fseek(in,((j+y0)*mrch->nx+x0)*sizeof(short)+sizeof(mrcH)+mrch->nsymbt,SEEK_SET);
			if ((i=fread(&cdata[j*nx*nz*sizeof(short)],nx*sizeof(short),nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		if ((ord+mord)&1) {
			for (i=nx*ny*nz-1; i>=0; i--) {
				t=cdata[i*2];
				cdata[i*2]=cdata[i*2+1];
				cdata[i*2+1]=t;
				rdata[i]=(float)(sdata[i]);
			} 
		}
		else {
			for (i=nx*ny*nz-1; i>=0; i--)  {
				rdata[i]=(float)(sdata[i]); 
			}
		}

#ifdef DEBUG
printf("Short read complete\n");
#endif
		break;
	case MODE_float:
		cdata=(unsigned char *)rdata;
		flags&=~EMDATA_COMPLEX;
		for (j=0; j<ny; j++) {
			fseek(in,((j+y0)*mrch->nx+x0)*sizeof(float)+sizeof(mrcH)+mrch->nsymbt,SEEK_SET);
			if ((i=fread(&rdata[j*nx*nz],nx*sizeof(float),nz,in))!=nz) {
				sprintf(s,"Incomplete data read %d/%d blocks\n",i,ny*nz);
				error(s,ERR_WARN);
			}
		}
		// swap order if machine and file have different order
		if (ord+mord==1) {
			for (i=nx*ny*nz*4-4; i>=0; i-=4) {
				c=cdata[i]; cdata[i]=cdata[i+3]; cdata[i+3]=c;
				c=cdata[i+1]; cdata[i+1]=cdata[i+2]; cdata[i+2]=c;
			}
		}
		break;
	default: printf("Bad MRC file\n"); zero(); return -1;
	}

mrch->nx=nx;
mrch->ny=ny;
mrch->nz=nz;

update();
if (pipe) pclose(in); else fclose(in);

pixel=4.66667;
parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(0,0,0);
atime=time(0);
doneData();
update();
//calcHist(Min(),Max());
return 0;
}

int EMData::readIMAGIC(const char *fsp3, int n,int nodata)
{
List *list;
EMData *tmp;
int ret;
imagicH *hed;
FILE *in,*inh;
char *fsp,*fsp2,*t;
float *data;
short *sdata;
unsigned char *cdata,jj;
char *lab[] = { "","Byte reversed ","" };
int i,j,count,type,ord=0;

hed=(imagicH *)Smalloc(sizeof(imagicH),1);

fsp=(char *)malloc(strlen(fsp3)+5);
fsp2=(char *)malloc(strlen(fsp3)+5);
strcpy(fsp,fsp3);
strcpy(fsp2,fsp3);
if (strrchr(fsp3,'.')!=NULL) { *(strrchr(fsp,'.'))=0; *(strrchr(fsp2,'.'))=0; }

strcat(fsp2,".hed");
strcat(fsp,".img");

inh=fopen(fsp2,"r");
in=fopen(fsp,"r");
if (inh==NULL || in==NULL) { error("ERROR: cannot open imagic file",ERR_ERROR); Sfree(hed,1); zero(); return -1; }

fread(hed,4,256,inh);
rewind(inh);

if (strrchr(fsp,'/')==NULL) setName(fsp);
else setName(strrchr(fsp,'/')+1);
setPath(fsp);

// check byte order of image (header must match image)
t=(char *)&hed->headrec;
#ifdef MSB_FIRST
if (t[0]==0&&t[1]==0) ord=0; else ord=1;
#else
if (t[0]==0&&t[1]==0) ord=1; else ord=0;
#endif

if (strncmp(hed->type,"REAL",4)==0) type=1;
else if (strncmp(hed->type,"INTG",4)==0) type=0;
else {
	error("Unknown imagic type\n",ERR_ERROR);
	fclose(in);
	fclose(inh);
	Sfree(hed,1);
	zero();
	return -1;
}

if (ord) { 
	swapint((char *)hed,14);
	swapint((char *)&hed->ixold,14);
	swapint((char *)&hed->SPACE,207);
}

count=hed->count+1;
if (n>=count) { fclose(in); fclose(inh); free(hed); zero(); return -1; }

fseek(inh,1024*n,SEEK_SET);
fseek(in,n*hed->nx*hed->ny*(2+2*type),SEEK_SET);
fread(hed,4,256,inh);

if (ord==1) { 
	swapint((char *)hed,14);
	swapint((char *)&hed->ixold,14);
	swapint((char *)hed->SPACE,207);
}

if (nodata) flags|=EMDATA_NODATA;
setSize(hed->nx,hed->ny,1);
imagich=hed;

if (!nodata) {
data=getData();
sdata=(short *)data;
cdata=(unsigned char *)data;

// read the data (remember to flip y
for (i=ny-1; i>=0; i--) {
	if (updProgress && updProgress(ny-i-1,ny,"Reading IMAGIC file")) break;
	fread(cdata+i*nx*(2+2*type),nx*(2+2*type),1,in);
}
if (updProgress) updProgress(0,0,NULL);

if (type) {
	if (ord) swapint((char *)data,nx*ny);
}
else {
	if (ord) {
		for (i=nx*ny-1; i>=0; i--) { 
			jj=cdata[i*2]; 
			cdata[i*2]=cdata[i*2+1];
			cdata[i*2+1]=jj;
			data[i]=sdata[i];
		}
	}
	else {
		for (i=nx*ny-1; i>=0; i--) data[i]=sdata[i];
	}
}

}

fclose(in);
fclose(inh);

free(fsp);
free(fsp2);

pixel=4.66667;
parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(hed->MRC1[1],hed->MRC1[2],hed->MRC1[0]);	// inappropriate use, but what the heck
nimg=hed->MRC2;
atime=time(0);
hed->SPACE[0]=0;
setName(hed->label);
if (name[0]=='!' && name[1]=='-') {
	i=sscanf(name+2," %f %f %f %f %f %f %f %f %f %f %f",
		ctf,ctf+1,ctf+2,ctf+3,ctf+4,ctf+5,ctf+6,ctf+7,ctf+8,ctf+9,ctf+10); 
	if (i==11) flags|=EMDATA_HASCTF;
	else printf("Incomplete CTF info: %s\n",name);
}
//calcHist(Min(),Max());
doneData();
update();
return(0);
}

int EMData::readIMAGIC3d(const char *fsp3,int nodata)
{
List *list;
EMData *tmp;
int ret;
imagicH *hed;
FILE *in,*inh;
char *fsp,*fsp2,*t;
float *data;
short *sdata;
unsigned char *cdata;
char *lab[] = { "","Byte reversed ","" };
int i,j,count,type,ord=0;

hed=(imagicH *)Smalloc(sizeof(imagicH),1);

fsp=(char *)malloc(strlen(fsp3)+5);
fsp2=(char *)malloc(strlen(fsp3)+5);
strcpy(fsp,fsp3);
strcpy(fsp2,fsp3);
if (strrchr(fsp3,'.')!=NULL) { *(strrchr(fsp,'.'))=0; *(strrchr(fsp2,'.'))=0; }

strcat(fsp2,".hed");
strcat(fsp,".img");

inh=fopen(fsp2,"r");
in=fopen(fsp,"r");
if (inh==NULL || in==NULL) { error("ERROR: cannot open imagic file",ERR_ERROR); Sfree(hed,1); zero(); return -1; }

fread(hed,4,256,inh);
rewind(inh);

if (strrchr(fsp,'/')==NULL) setName(fsp);
else setName(strrchr(fsp,'/')+1);
setPath(fsp);

// check byte order of image (header must match image)
t=(char *)&hed->headrec;
#ifdef MSB_FIRST
if (t[0]==0&&t[1]==0) ord=0; else ord=1;
#else
if (t[0]==0&&t[1]==0) ord=1; else ord=0;
#endif

if (strncmp(hed->type,"REAL",4)==0) type=1;
else if (strncmp(hed->type,"INTG",4)==0) type=0;
else {
	error("Unknown imagic type\n",ERR_ERROR);
	fclose(in);
	fclose(inh);
	Sfree(hed,1);
	zero();
	return -1;
}

if (ord) { 
	swapint((char *)hed,14);
	swapint((char *)&hed->ixold,14);
	swapint((char *)&hed->SPACE,207);
}

count=hed->count+1;

if (count!=hed->nx) { printf("Imagic3d invalid %d,%d,%d\n",hed->nx,hed->ny,count); zero(); return(1); }

if (nodata) flags|=EMDATA_NODATA;
setSize(hed->nx,hed->ny,hed->nx);
imagich=hed;

if (!nodata) {
data=getData();
sdata=(short *)data;
cdata=(unsigned char *)data;

printf("Imagic3d -> %d,%d,%d\n",nx,ny,nz);

// read the data
for (j=0; j<nz; j++) {
	if (updProgress && updProgress(j,nx,"Reading IMAGIC 3D file")) break;
	for (i=ny-1; i>=0; i--) {
		fread(cdata+(i+j*ny)*nx*(2+2*type),nx*(2+2*type),1,in);
	}
}

if (updProgress) updProgress(0,0,NULL);

if (type) {
	if (ord) swapint((char *)data,nx*ny*nz);
}
else {
	if (ord) {
		for (i=nx*ny*nz-1; i>=0; i--) { 
			j=cdata[i*2]; 
			cdata[i*2]=cdata[i*2+1];
			cdata[i*2+1]=j;
			data[i]=sdata[i];
		}
	}
	else {
		for (i=nx*ny*nz-1; i>=0; i--) data[i]=sdata[i];
	}
}
}

fclose(in);
fclose(inh);

free(fsp);
free(fsp2);

pixel=4.66667;
parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(0,0,0);
atime=time(0);
//calcHist(Min(),Max());
doneData();
update();
return(0);
}

int EMData::readEMIM(const char *filespec, int n,int nodata)
{
FILE *in;
emimH1 imh1;
emimH2 imh2;
int swap=0;

in=fopen(filespec,"r");
if (!in) { error("EMIM: Cannot open emim file for reading",ERR_ERROR); zero(); return -1; }

// read main header
if (!fread(&imh1,sizeof(emimH1),1,in)) { error("EMIM: Cannot read EMIM",ERR_ERROR); zero(); return -1; }

if (strncmp(imh1.magic,"EMIM",4)) { error("EMIM: File is not an EMIM",ERR_ERROR); zero(); return -1; }

if (imh1.count&0x7f000000) swap=1;	// there are unlikely to be more than 16e6 images in the file

if (swap) swapint((char *)&imh1,7);		// swap order of main header if necessary

if (n>=imh1.count) { error("EMIM: Not enough images",ERR_ERROR); zero(); return -2; }

// move to the image we want
fseek(in,sizeof(emimH1)+n*(sizeof(emimH2)+imh1.nx*imh1.ny*imh1.nz*sizeof(float)),SEEK_SET);

fread(&imh2,sizeof(emimH2),1,in);

if (nodata) flags|=EMDATA_NODATA;
setSize(imh1.nx,imh1.ny,imh1.nz);

if (!nodata) {
getData();
if (fread(rdata,nx*nz*sizeof(float),ny,in)!=ny) { error("EMIM: File truncated?",ERR_ERROR); zero(); return -3; }

if (swap) {
	swapint((char *)&imh2,10);
	swapint((char *)imh2.param,140);
	swapint((char *)rdata,nx*ny*nz);
}

doneData();
}

dx=imh2.dx;
dy=imh2.dy;
dz=imh2.dz;
setRAlign(imh2.alt,imh2.az,imh2.phi);
memcpy(param,imh2.param,25*4*sizeof(float));
memcpy(ptype,imh2.ptype,4);

setName(imh2.name);

flags=0;
flags=EMDATA_NEEDUPD|EMDATA_NEEDHIST;
if (imh1.flag&EMIM_COMPLEX) flags|=EMDATA_COMPLEX+EMDATA_RI;

pixel=imh1.pixel;
parent=NULL;
pathnum=n;
atime=time(0);
update();
//calcHist(Min(),Max());
return 0;
}

// write routines for specific formats
int EMData::writeIMAGIC(const char *filespec, int n,int nolock)	// n=-1 appends 
{
EMData *tmp;
int ret;
imagicH hed;
FILE *in,*inh;
char *fsp,*fsp2;
int i,swap=0,count,size;
time_t itime;
struct tm *tm;
float *cpy;

fsp=(char *)Smalloc(strlen(filespec)+5,0);
fsp2=(char *)Smalloc(strlen(filespec)+5,0);
strcpy(fsp,filespec);
if (strlen(fsp)>4 && (strcmp(fsp+strlen(fsp)-4,".img")==0 || strcmp(fsp+strlen(fsp)-4,".hed")==0)) fsp[strlen(fsp)-4]=0;
strcpy(fsp2,fsp);

strcat(fsp2,".hed");
strcat(fsp,".img");

if (!nolock) file_lock_wait(fsp2);
inh=fopen(fsp2,"r+");
if (inh==NULL) inh=fopen(fsp2,"w+");
if (inh==NULL) { error("Cannot open imagic header.",ERR_ERROR); zero(); return -1; }
if (!nolock) file_lock_wait(fsp);
in=fopen(fsp,"r+");
if (in==NULL) in=fopen(fsp,"w+");
if (in==NULL) { fclose(inh); error("Cannot open imagic file.",ERR_ERROR); zero(); return -1; }

// new file, start from zero
if (fread(&hed,1024,1,inh)!=1) {
	hed.nx=nx;
	hed.ny=ny;
	hed.count=-1;
	hed.headrec=1;
}
rewind(inh);

// check byte order
if (hed.headrec>64) {
	swap=1;		// file has the opposite order we do
	swapint((char *)&hed,256);
}

// make sure the file has the right image size (all must be same)
count=hed.count+1;
size=hed.nx*hed.ny*4;
if (nx*ny*4!=size) { error("Imagic file wrong size",ERR_ERROR); if (!nolock) file_unlock(fsp); if (!nolock) file_unlock(fsp2); zero(); return -2; }
if (n>=count) {
	EMData tmp;
	tmp.setSize(nx,ny,nz);
	tmp.zero();
	tmp.setNImg(-1);
	fclose(inh);
	fclose(in);
	while (n>count) { tmp.writeIMAGIC(filespec,-1,1); count++; }
	n=-1;
	inh=fopen(fsp2,"r+");
	in=fopen(fsp,"r+");
}

// seek to the correct place in the file
if (n<0) {
	fseek(inh,0,SEEK_END);
	fseek(in,0,SEEK_END);
	count++;
	hed.imnum=count;
}
else {
	fseek(inh,1024*n,SEEK_SET);
	fseek(in,size*n,SEEK_SET);
	hed.imnum=n;
}

// set up the header
itime=time(0);
tm=localtime(&itime);
hed.error=0;
hed.headrec=1;
hed.mday=tm->tm_mday;
hed.month=tm->tm_mon;
hed.year=tm->tm_year+1900;
hed.hour=tm->tm_hour;
hed.minute=tm->tm_min;
hed.sec=tm->tm_sec;
hed.reals=nx*ny;
hed.pixels=nx*ny;
hed.ny=ny;
hed.nx=nx;
hed.ixold=hed.iyold=0;
hed.avdens=Mean();
hed.sigma=Sigma();
hed.oldav=0;
hed.max=Max();
hed.min=Min();
hed.MRC1[0]=euler.phi();
hed.MRC1[1]=euler.alt();
hed.MRC1[2]=euler.az();
hed.MRC2=nimg;


// buffering parameters IMAGIC seems to need

hed.lbuf=nx;
hed.inn=1;
hed.iblp=ny;
hed.ifb=0;
hed.lbw=0;
hed.lbr= -1;
hed.lastlr= -1;
hed.lastlw=1;
hed.num=8; // ???
hed.nhalf=nx/2;
hed.ibsd=nx*2;
hed.ihfl=7;
hed.lcbr= -1;
hed.lcbw=1;
hed.imstr= -1;
hed.imstw= -1;
hed.istart=1;
hed.iend=nx;
hed.leff=nx;
hed.linbuf=nx*2;
hed.ntotbuf= -1;
hed.icstart=1;
hed.icend=nx/2;
memset(hed.misc,0,400);

getDataRO();
if (swap) {
	swapint((char *)&hed,256);
	cpy=(float *)Smalloc(nx*ny*sizeof(float),0);
	memcpy(cpy,rdata,nx*ny*sizeof(float));
	swapint((char *)cpy,nx*ny);
}
else cpy=rdata;

if (flags&EMDATA_HASCTF) {
	char c[80];
	sprintf(c,"!-%1.3g %1.3g %1.3g %1.3g %1.3g %1.3g %1.3g %1.3g %1.3g %1.3g %1.3g",
		ctf[0],ctf[1],ctf[2],ctf[3],ctf[4],ctf[5],ctf[6],ctf[7],ctf[8],ctf[9],ctf[10]);
	setName(c);		
}

if (name) strncpy(hed.label,name,80); else name[0]=0;
hed.type[0]='R';
hed.type[1]='E';
hed.type[2]='A';
hed.type[3]='L';

fwrite(&hed,1024,1,inh);

for (i=ny-1; i>=0; i--) {
	fwrite(&cpy[i*nx],4*nx,1,in);
}
if (swap) Sfree(cpy,0);
doneData();

fseek(inh,4,SEEK_SET);
count--;
if (swap) swapint((char *)&count,1);
fwrite(&count,4,1,inh);
//printf("%d records\n",count);
fclose(in);
fclose(inh);
if (!nolock) file_unlock(fsp);
if (!nolock) file_unlock(fsp2);

Sfree(fsp,0);
Sfree(fsp2,0);

return 0;
}

int EMData::writeMRC(const char *filespec)
{
struct tm *t;
time_t t0=time(0);
FILE *out;


if(filespec==NULL){out=stdout;}
else{
	file_lock_wait(filespec);
	out=fopen(filespec,"w");
	if (!out) { error("Could not write MRC",ERR_ERROR); return -1; }
}

t=localtime(&t0);

if (!mrch) {
	mrch=(mrcH *)Smalloc(sizeof(mrcH),1);
	memset(mrch,0,sizeof(mrcH));
	mrch->mx=nx-1;
	mrch->my=ny-1;
	mrch->mz=nz-1;
	mrch->xlen=mrch->ylen=mrch->zlen=pixel;
	mrch->alpha=mrch->beta=mrch->gamma=90.0;
	mrch->mapc=1;
	mrch->mapr=2;
	mrch->maps=3;
	mrch->nxstart=mrch->nystart=mrch->nzstart=0;
	mrch->nlabl=1;
	sprintf(&mrch->labels[0][0],"EMAN %d/%02d/%02d %d:%02d %s",
		t->tm_mon,t->tm_mday,t->tm_year,t->tm_hour,t->tm_min,name);
	mrch->labels[0][79]=0;
	mrch->labels[1][0]=0;
}
else if (mrch->nlabl<MRC_NUM_LABELS-1) {
	sprintf(&mrch->labels[mrch->nlabl][0],"EMAN %d/%02d/%02d %d:%02d %s",
		t->tm_mon,t->tm_mday,t->tm_year,t->tm_hour,t->tm_min,name);
	mrch->labels[0][79]=0;
	mrch->labels[1][0]=0;
	mrch->nlabl++;
}
mrch->nx=nx;
mrch->ny=ny;
mrch->nz=nz;
mrch->ispg=0;
mrch->nsymbt=0;
mrch->amin=Min();
mrch->amax=Max();
mrch->amean=Mean();
if (flags&EMDATA_COMPLEX) { mrch->mode=MODE_float_COMPLEX; mrch->nx/=2; }
else mrch->mode=MODE_float;

fwrite(mrch,sizeof(mrcH),1,out);
fwrite(getData(),nx*nz,ny*sizeof(float),out);
doneData();
if(filespec!=NULL){
	fclose(out);
	file_unlock(filespec);
}
return 0;
}

int EMData::writeVTK(const char *filespec)
{
struct tm *t;
time_t t0=time(0);
FILE *out;


if(filespec==NULL){ out=stdout; }
else{
	file_lock_wait(filespec);
	out=fopen(filespec,"w");
	if (!out) { error("Could not write MRC",ERR_ERROR); return -1; }
}

t=localtime(&t0);

fprintf(out,"# vtk DataFile Version 2.0\n");
fprintf(out,"%s\n",Name());
fprintf(out,"BINARY\n");
fprintf(out,"DATASET STRUCTURED_POINTS\nDIMENSIONS %0d %0d %0d\nORIGIN %0d %0d %0d\nSPACING 1 1 1\n",
	nx,ny,nz,-nx/2,-ny/2,-nz/2);
fprintf(out,"POINT_DATA %0d\nSCALARS density float 1\n",nx*ny*nz);
fwrite(getData(),nx*nz,ny*sizeof(float),out);
doneData();

if(filespec!=NULL){
	fclose(out);
	file_unlock(filespec);
}

return 0;
}


int EMData::writePGM(const char *filespec,float gmin,float gmax)
{
FILE *out;
unsigned char *tmp;

if (nz!=1) { error("Cannot write 3d image as PGM",ERR_ERROR); return -2; }

out=fopen(filespec,"w");
if (!out) { error("Could not write PGM",ERR_ERROR); return -1; }

tmp=(unsigned char *)malloc(nx*ny);
renderAmp8(tmp,0,0,nx,ny,nx,1.0,0,255,gmin,gmax);

fprintf(out,"P5\n%d %d\n255\n",nx,ny);
fwrite(tmp,nx,ny,out);
free(tmp);
fclose(out);

return 0;
}

int EMData::writeEMIM(const char *filespec, int n)	// n=-1 appends
{
return -1;
}

// read the initial spider header
SPIDERH *readSpiHed(const char *fsp) {
int i,swap,imsize;
SPIDERH *sh;
FILE *in=fopen(fsp,"r");
if (!in) return NULL;

sh=(SPIDERH *)malloc(sizeof(SPIDERH));
// read initial header
if (!fread(sh,sizeof(SPIDERH),1,in)) { free(sh); return NULL; }
fclose (in);

if (sh->nslice>0&&sh->nslice<10000.0 && sh->nslice==floor(sh->nslice)) swap=0; else swap=1;
if (swap) swapint((char *)sh,211);

// Basic tests to see if the header is valid
for (i=0; i<160; i++) {
	if (!isprint(sh->name[i]) && sh->name[i]!=0 ) break;
}
if (i!=160 || sh->nslice!=floor(sh->nslice) || sh->nslice<1.0 ||sh->nslice>8192 ||
	(sh->nslice!=1.0 && sh->nslice!=3.0)) { free(sh); return NULL; }

// a few return values we shove into unused areas
if (sh->istack!=2) sh->u8[158]=1;	// # images in file
else sh->u8[158]=sh->maxim;
sh->u8[159]=swap;					// byte order change necessary ?
sh->u8[160]=sh->headlen+sh->nx*sh->ny*sh->nslice*sizeof(float); // total image+hed len

return sh;
}

// read a spider file
int EMData::readSPIDER(const char *fsp, int n, int nodata) {
FILE *in;
int swap=0,sz,stack;
SPIDERH *sh;

if (!spiderH) spiderH=(SPIDERH *)calloc(sizeof(SPIDERH),1);

sh=readSpiHed(fsp);
if (!sh) { zero(); return -1; }
if (sh->istack) stack=sh->maxim-1; else stack=0;
if (n>stack) { free(sh); zero(); return -1; }
swap=(int)sh->u8[159];

in=fopen(fsp,"r");

if (strrchr(fsp,'/')==NULL) setName(fsp);
else setName(strrchr(fsp,'/')+1);
setPath(fsp);

// read header
fseek(in,(int)(sh->u8[160]*n+(stack?sh->headlen:0)),SEEK_SET);
if (!fread(spiderH,sizeof(SPIDERH),1,in)) { free(sh); zero(); return -1; }
if (swap) swapint((char *)spiderH,211);
fseek(in,(int)spiderH->headlen-sizeof(SPIDERH),SEEK_CUR);

setSize(spiderH->nx,spiderH->ny,spiderH->nslice);

if (!nodata) {
	getData();
	if (fread(rdata,nx*ny,4*nz,in)!=4*nz) { printf("Incomplete SPIDER data read\n"); zero(); return -1; }

	if (swap) swapint((char*)rdata,nx*ny*nz);
	doneData();
}
fclose(in);

pixel=4.66667;
parent=NULL;
pathnum=0;
dx=dy=dz=0;
setRAlign(0,0,0);
atime=time(0);
spiderH=sh;
return 0;
}

// write a spider file, if n>nimg in file || n==-1 appends
int EMData::writeSPIDER(const char *fsp, int n) {
SPIDERH *sh;
int hr=(1+1023/(nx*4));
int hl=hr*nx*4;
FILE *out=NULL;

sh=readSpiHed(fsp);
if (sh) {
	if (sh->u8[159]) { printf("Cannot write to swapped order spider file\n"); return -1; }
	if (sh->nx!=nx || sh->ny!=ny || sh->nslice!=nz) { printf("Wrong image size spider write\n"); exit(1); }
	if (!sh->istack) { printf("Cannot write to single-image Spider files\n"); return -2; }
	out=fopen(fsp,"r+");
	if (!out) return (-4);
	if (n==-1||n>=sh->maxim) { 
//		n=sh->u8[158]; 
		n=sh->maxim;
		sh->maxim++; 
		fwrite(sh,sizeof(SPIDERH),1,out);		
	}
}
else {
	sh=(SPIDERH *)calloc(hl,1);
	sh->mmvalid=0;
	sh->sigma=-1.0;
	sh->angvalid=0;
	sh->scale=1.0;
	sh->nslice=nz;
	sh->nx=nx;
	sh->ny=ny;
	sh->type=(nz==1?1:3);
	sh->reclen=nx*4;
	sh->headrec=hr;
	sh->headlen=hl;
	sh->inuse=1;	//?
	sh->istack=2;
	sh->maxim=1;
	
	out=fopen(fsp,"w");
	if (!out) { printf("Cannot open '%s'\n",fsp); return -3; }
	fwrite(sh,hl,1,out);
}

if (!spiderH) {
	spiderH=(SPIDERH *)calloc(sizeof(SPIDERH),1);
	spiderH->mmvalid=0;
	spiderH->sigma=-1.0;
	spiderH->angvalid=0;
	spiderH->scale=1.0;
}

spiderH->nslice=nz;
spiderH->nx=nx;
spiderH->ny=ny;
spiderH->type=nz==1?1:3;
spiderH->reclen=nx*4;
spiderH->headrec=(1+1023/(nx*4));
spiderH->headlen=spiderH->headrec*spiderH->reclen;
spiderH->inuse=1;
spiderH->dx=dx;
spiderH->dy=dy;
spiderH->dz=dz;
//spiderH->phi= 0;
//spiderH->theta= 0;
//spiderH->gamma= 0;
strcpy(spiderH->name,name);

fseek(out,hl+n*(hl+nx*ny*nz*sizeof(float)),SEEK_SET);

fwrite(spiderH,sizeof(SPIDERH),1,out);
char *t=(char *)calloc(hl-sizeof(SPIDERH),1);
fwrite(t,hl-sizeof(SPIDERH),1,out);
free(t);
getDataRO();
fwrite(rdata,nx*ny,nz*sizeof(float),out);
doneData();
return 0;
}


// update statistics
void EMData::realCalcHist()
{
int i,j,k,p0,p1,di;
float u,v,w,norm=0;
static int prime[] = { 1,3,7,11,17,23,37,59,127,253,511 } ;

for (i=0; i<200; i++) hist[i]=0;


// decide on the sampling based on how big the data set is
if (nx*ny*nz<300000) p0=p1=0;
else if (nx*ny*nz<2000000) { p0=2; p1=3; }
else if (nx*ny*nz<8000000) { p0=4; p1=6; }
else { p0=7; p1=9; }

// real images don't need as much sampling
if (!(flags&EMDATA_COMPLEX)&&p0>0) { p0++; p1++; }

for (k=p0; k<=p1; k++) {
	if (flags&EMDATA_COMPLEX) di=prime[k]*2; else di=prime[k];
	norm+=nx*ny*nz/(float)di;
	w=199.999/(histmax-histmin);
	for (i=nx*ny*nz-di; i>=0; i-=di) {
		v=(rdata[i]-histmin)*w;	
		u=v-floor(v);
		j=(int)floor(v);
		if (j<0||j>199) continue;
		hist[j+1]+=u;
		hist[j]+=1.0-u;
	}
}

for (i=0; i<200; i++) hist[i]=hist[i]*100.0/norm;

}

void EMData::realupdate() {
int i,j,k,l,di,n=0;
double u,v,w;
double m,s,t,m2,s2;

if (nx==0 || ny==0 || nz==0) { setSize(2,2,2); zero(); }

flags&=~EMDATA_NEEDUPD;

max=-1.0e20;
min=-max;
m=m2=0;
s=s2=0;

// getDataRO();		//dangerous not to, but causes problems

for (i=0; i<200; i++) hist[i]=0;
if ((flags&EMDATA_COMPLEX)&& !(flags&EMDATA_RI)) di=2; else di=1;

// pass 1, statistics
for (j=0,i=0; j<nz; j++) {
	for (k=0; k<ny; k++) {
		for (l=0; l<nx; l+=di,i+=di) {
			v=rdata[i];
//	if (v>1e5) { printf(" %d,%d=%f\n",i%nx,i/nx,v); continue; }
			if (v>max) max=v;
			if (v<min) min=v;
			m+=v;
			s+=v*v;
			if (v!=0) n++;

/*			if (k<4 || l<4 || k>ny-5 || l>nx-5) {
				m2+=v;
				s2+=v*v;
			}*/
		}
	}
}
mean=m*di/(double)(nx*ny*nz);
sigma=s*di/(double)(nx*ny*nz)-mean*mean;
if (sigma<0) sigma=0;	// in case of roundoff errors for v small sigma
//if (fabs(sigma)<1.0e-10) sigma=0;		// could be that a 'zero' is a bit off due to rounding
sigma=sqrt(sigma);

if (n==0) n=1;
mean2=m*di/(double)(n);
sigma2=s*di/(double)(n)-mean2*mean2;
if (sigma2<0) sigma2=0;	// in case of roundoff errors for v small sigma
sigma2=sqrt(sigma2);

histmax=max;
histmin=min;

//	printf("!!! %d %g %g %g %g\n",(nx*ny*nz),min,max,(float)m,(float)s);

if (!finite(mean) || !finite(sigma)) {
	char st[120];
	sprintf(st,"ERROR - Bad statistics on this data set! %d %g %g",(nx*ny*nz),min,max);
//	for (i=0; i<110; i++) printf("%5.3g\t",rdata[i+6050]);
//	printf("\n");
	error(st,ERR_ERROR);
	zero();
	return;
}

if (histmax-mean>sigma*3.0) histmax=mean+sigma*3.0;
if (mean-histmin>sigma*3.0) histmin=mean-sigma*3.0;
if (histmax==histmin) histmax+=.0001;
if (max==min) max+=.0001;

// pass 2, histogram
calcHist(histmin ,histmax);

return;
}

// 2d only
float EMData::edgeMean() {
int i,j,k,di;
float edgemean;

getData();
// this finds the average around the (1 pixel) edge of the image
edgemean=0;
for (i=0,j=(ny-1)*nx; i<nx; i++,j++) edgemean+=rdata[i]+rdata[j]; 
for (i=0, j=nx-1; i<nx*ny; i+=nx,j+=nx) edgemean+=rdata[i]+rdata[j];
edgemean/=(double)(nx*2+ny*2);
doneData();

return edgemean;
}

//circular edgeMean
float EMData::cirMean() {
static int busy=0;
static EMData *mask=NULL;
int i,n;
float r,*d;

while (busy) ;
busy=1;
if (!mask || mask->xSize()!=nx || mask->ySize()!=ny || mask->zSize()!=nz) {
	if (!mask) mask=new EMData;
	mask->setSize(nx,ny,nz);
	mask->zero();
	mask->applyMask(ny/2-2,7);
	n=0;
	d=mask->getDataRO();
	for (i=0; i<nx*ny*nz; i++) if (d[i]) n++;
	mask->doneData();
	mask->multConst(1.0/(float)n);
//mask->writeMRC("cir.mrc");
}
r=dot(mask);
busy=0;
return r;
}

// general parameter set/gets
int EMData::setSize(int x,int y, int z) {
if (waitReady(0)) return -1; 

if (x<=0) x=1;
if (y<=0) y=1;
if (z<=0) z=1;
if (y==1&&z!=1) { y=z; z=1; }
if (x==1&&y!=1) { x=y; y=1; }
nx=x;
ny=y;
nz=z;
if (flags&EMDATA_NODATA) {
	if (rdata) free(rdata);
	rdata=NULL;
}
else rdata=(float *)Srealloc(rdata,x*y*z*sizeof(float),1);
update();
flags&= ~EMDATA_SWAPPED;
if (supp) { free(supp); supp=NULL; }
return 0;
}

// Access to raw data
float *EMData::getData()	// This returns a pointer to the raw float data
{
//if (flags&EMDATA_BUSY || rocount>0 || !rdata) return NULL;
waitReady(0);
if (flags&EMDATA_SWAPPED) swapin();
flags|=EMDATA_BUSY;
return rdata;
}

// Read only access to raw data
float *EMData::getDataRO()	// This returns a pointer to the raw float data
{
//if (flags&EMDATA_BUSY || !rdata) return NULL;
waitReady(1);
if (flags&EMDATA_SWAPPED) swapin();
rocount++;
return rdata;
}

void EMData::doneData()	// MUST be called when the data is no longer being
						// modified. Another getData while the data is
						// out will return NULL
{
if (flags&EMDATA_BUSY) {
	flags&=(~EMDATA_BUSY);
	update();
}
else {
	rocount--;
	if (rocount<0) rocount=0;
}

return;
}

int EMData::Busy()			// Returns 1 if data is checked out
{
if (flags&EMDATA_BUSY || rocount) return 1; else return 0;
}

// Waits for data to become available
int EMData::waitReady(int ro) 
{
/*
if (ro) {
	while (flags&EMDATA_BUSY) sleep(1);
}
else {
	while ((flags&EMDATA_BUSY) || (rocount)) sleep(1);
}
*/

// for debugging only
if (ro) {
	while (flags&EMDATA_BUSY) {
		printf("Busy 1 (%p)\n",this);
#ifdef ALPHA
		sleep(1);
#else
		usleep(100);
#endif
	}
}
else {
	if ((flags&EMDATA_BUSY) || (rocount)) {
		printf("Busy 0.%d (%p)\n",rocount,this);
		while ((flags&EMDATA_BUSY) || (rocount)) {
#ifdef ALPHA
		sleep(1);
#else
		usleep(100);
#endif
		}
		printf("Busy done 0.%d (%p)\n",rocount,this);	
	}
}

return 0;
}

// Rendering
int EMData::renderAmp8( unsigned char * data , int x0 , int y0, int ixsize, 
	int iysize, int bpl, float scale, int mingray, int maxgray,float renMin,float renMax) {
int i,j,k,l,ll,ymin=0,xmin,ymax,lmax,xsize,ysize;
int dsx,dsy; 	// used for integer scalings
int remx,remy,br,addi,addr;	// used for non integer scaling
float x,y,gs,t,rm;
unsigned char *tmp;

if (flags&EMDATA_COMPLEX) ri2ap();
if (renMax<=renMin) renMax=renMin+.01;
rm=renMin;

getDataRO();
scale=1.0/scale;
ysize=iysize;
xsize=ixsize;
if (iysize*scale>ny) ymin=(int)(iysize-ny/scale);
gs=(maxgray-mingray)/(renMax-renMin);
if (renMax<renMin) { gs=0;  rm=1.0e20; }
if (scale==floor(scale)) { dsx=(int)scale; dsy=(int)(scale*nx); } else dsx=-1;
if (dsx==-1) { remx=remy=0; addi=(int)floor(scale); addr=(int)(100000.0*(scale-floor(scale))); }
if (x0<0) { xmin=(int)(-x0/scale); xsize-=(int)floor(x0/scale); x0=0; } 
else { xmin=0; }
if ((xsize-xmin)*scale>(nx-x0)) xsize=(int)((nx-x0)/scale+xmin);
if (y0<0) { ymax=(int)(ysize+y0/scale-1); ymin+=(int)floor(y0/scale); y0=0; } else ymax=ysize-1;

if (xmin<0) xmin=0;
if (ymin<0) ymin=0;
if (xsize>ixsize) xsize=ixsize;
if (ymax>iysize) ymax=iysize;

lmax=nx*ny-1;

// complex version very similar to real version
if (flags&EMDATA_COMPLEX) {
// integer scalings are fast
	if (dsx!=-1) {
		for (j=ymax,l=y0*nx; j>=ymin; j--) {
			for (i=xmin,ll=x0; i<xsize; i++,ll+=dsx) {
				if (l+ll>lmax || ll>=nx-2) break;
	// +2 here is to prevent displaying qx=0 twice
				if (ll>=nx/2) {
					if (l>=(ny-scale)*nx) k=2*(ll-nx/2)+2;
					else k=2*(ll-nx/2)+l+2+nx;
				}
				else k=nx*ny-(l+2*ll)-2;
				t=rdata[k];
				if (t<=rm) k=mingray;	// renMin, but deals with renMin>renMax
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
			}
			l+=dsy;
		}
	}
// non integer scalings a bit slower
	else {
		remy=10; // a small fixed fraction
		for (j=ymax,l=y0*nx; j>=ymin; j--) {
			br=l;
			remx=10;
			for (i=xmin,ll=x0; i<xsize-1; i++) {
				if (l+ll>lmax || ll>=nx-2) break;
				if (ll>=nx/2) {
					if (l>=(ny*nx-nx)) k=2*(ll-nx/2)+2;
					else k=2*(ll-nx/2)+l+2+nx;
				}
				else k=nx*ny-(l+2*ll)-2;

				t=rdata[k];
				if (t<=rm) k=mingray;
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
				ll+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; ll++; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx; }
		}
	}
}
else {
// integer scalings are fast
	if (dsx!=-1) {
		for (j=ymax,l=x0+y0*nx; j>=ymin; j--) {
			br=l;
			for (i=xmin; i<xsize; i++,l+=dsx) {
				if (l>lmax) break;
				t=rdata[l];
				if (t<=rm) k=mingray;
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
			}
			l=br+dsy;
		}
	}
// non integer scalings a bit slower
	else {
		remy=10;
		for (j=ymax,l=x0+y0*nx; j>=ymin; j--) {
			br=l;
			remx=10;
			for (i=xmin; i<xsize; i++) {
				if (l>lmax) break;
				t=rdata[l];
				if (t<=rm) k=mingray;
				else if (t>=renMax) k=maxgray;
				else {
					k=(int)(gs*(t-renMin));
					k+=mingray;
				}
				data[i+j*bpl]=k;
				l+=addi;
				remx+=addr;
				if (remx>100000) { remx-=100000; l++; }
			}
			l=br+addi*nx;
			remy+=addr;
			if (remy>100000) { remy-=100000; l+=nx; }
		}
	}
}

doneData();
return 1;
}

int EMData::renderPha24( unsigned char * data , int x , int y, int xsize, int ysize, 
	int bpl, float scale, float renMin, float renMax)
{
return -1;
}

//
void EMData::ri2ap() {		// converts real/imaginary to amp/pha (complex only)
int i;
float f;
float *data;

if (!(flags&EMDATA_COMPLEX) || !(flags&EMDATA_RI)) { /* printf("already ap\n"); */  return; }
data=getData();

for (i=0; i<nx*ny*nz; i+=2) {
	f=hypot(data[i],data[i+1]);
	if (data[i]==0 && data[i+1]==0) data[i+1]=0;
	else data[i+1]=atan2(data[i+1],data[i]);
	data[i]=f;
}

doneData();
flags&=~EMDATA_RI;
update();
}

void EMData::ap2ri() {		// the opposite
int i;
float f,*data;

if (!(flags&EMDATA_COMPLEX) || (flags&EMDATA_RI)) return;
data=getData();

for (i=0; i<nx*ny*nz; i+=2) {
	f=data[i]*sin(data[i+1]);
	data[i]=data[i]*cos(data[i+1]);
	data[i+1]=f;
}

doneData();
flags|=EMDATA_RI;
update();
}

// real images only
void EMData::invert() {
int i,di=1;
float f,*data;

if ((flags&EMDATA_COMPLEX)) return;
data=getData();

for (i=0; i<nx*ny*nz; i+=di) {
	data[i]=-data[i];
}

doneData();
update();
}

// 'this' should be noisy, 'to' should be less noisy
void EMData::normalizeTo(EMData *im2) {
float *d1,*d2;
int i,n;
float a,b;

// This section normalizes the image to make the subtraction work better
n=xSize()*ySize()*zSize();
d1=getData();
d2=im2->getDataRO();
a=b=0;
for (i=0; i<n; i++) {
	if (d2[i]>im2->Sigma()*.7) {		// .7 is arbitrary 
		a+=d1[i]/d2[i];
		b+=1.0;
	}
}
for (i=0; i<n; i++) d1[i]=d1[i]*b/a;

doneData();
im2->doneData();

}

void EMData::normalize() {
int i,di=1;
float f,*data;
float nor;

if ((flags&EMDATA_COMPLEX) && !(flags&EMDATA_RI)) di=2;
if ( !finite(Sigma())) { zero(); return; }
if (Sigma()==0 ) return;
data=getData();


nor=Sigma2();	// this normalizes to sigma excluding pixels which are exactly zero

//printf("%f %f %f\n",nor,Sigma(),Sigma2());

for (i=0; i<nx*ny*nz; i+=di) {
	data[i]=(data[i]-Mean())/nor;
}

doneData();
update();
}

// real only
void EMData::normalizeMax() {
int i,di=1;
float f,*data;
float sc,d;

if ( !finite(Sigma())) { zero(); return; }
if (Sigma()==0 ) return;
sc=(Max()-Min())/2.0;
d=(Max()+Min())/2.0;

data=getData();

for (i=0; i<nx*ny*nz; i+=di) {
	data[i]=(data[i]-d)/sc;
}

doneData();
update();
}

void EMData::edgeNormalize(int cir) {
int i,di=1;
float f,*data,em,s;

if ((flags&EMDATA_COMPLEX) && !(flags&EMDATA_RI)) di=2;
if (Sigma()==0 || !finite(Sigma())) return;

if (cir) em=cirMean();
else em=edgeMean();
s=Sigma();
data=getData();

for (i=0; i<nx*ny*nz; i+=di) {
	data[i]=(data[i]-em)/s;
}

doneData();
update();
}

// makes plans for both forward and backwards transforms
rfftwnd_plan EMData::makeplan(int dim,int nx,int ny, int nz,int fw,int measure) {
int i;

// see if we've got one already
for (i=0; i<plans; i++) {
//	printf("%d. %d %d   %d %d   %d %d   %d %d   %d %d\n",i,nplan[i][0],nz,nplan[i][1],ny,nplan[i][2],nx,dimplan[i],dim,fwplan[i],fw);
	if (nplan[i][0]==nz && nplan[i][1]==ny && nplan[i][2]==nx 
		&& dimplan[i]==dim && fwplan[i]==fw) break;
}
if (i!=plans) return plan[i];

//printf("New plan %d %d %d\n",nx,ny,nz);
while (dimplan[0]==-1) ;	// make sure we don't make 2 plans at once


if (plan[5]) rfftwnd_destroy_plan(plan[5]);
for (i=5; i>0; i--) {
	plan[i]=plan[i-1];
	dimplan[i]=dimplan[i-1];
	fwplan[i]=fwplan[i-1];
	nplan[i][0]=nplan[i-1][0];
	nplan[i][1]=nplan[i-1][1];
	nplan[i][2]=nplan[i-1][2];
}
dimplan[0]=-1;

if (measure) measure=FFTW_MEASURE; else measure=FFTW_ESTIMATE;

nplan[0][0]=nz;
nplan[0][1]=ny;
nplan[0][2]=nx;
fwplan[0]=fw;
if (fw) plan[0]=rfftwnd_create_plan(dim, nplan[0]+(3-dim),FFTW_REAL_TO_COMPLEX,
	measure|FFTW_OUT_OF_PLACE|FFTW_USE_WISDOM|FFTW_THREADSAFE);
else plan[0]=rfftwnd_create_plan(dim, nplan[0]+(3-dim),FFTW_COMPLEX_TO_REAL,
	measure|FFTW_IN_PLACE|FFTW_USE_WISDOM|FFTW_THREADSAFE);
dimplan[0]=dim;
if (plans<6) plans++;
return plan[0];
}

EMData *EMData::doFFT()	{	// obvious, note that result is R,I not A,P
EMData *dat;
float *d,re,im,norm;
int i,j,k,l,ii,jj,dim;
char s[80],*t;
mrcH *newh;
rfftwnd_plan cplan;

if (flags&EMDATA_COMPLEX) return this;	// already complex, just return

if (nx!=npfa(nx) || (ny!=npfa(ny)&&ny!=1) || (nz!=npfa(nz)&&nz!=1)) { error("Invalid FFT size!",ERR_ERROR); return NULL; }

getData();	// we open for writing so nobody else does a fft at the same time

//flags|=EMDATA_NEWFFT;		// disable caching

if (fft) {
	if (!(flags&EMDATA_NEWFFT)) { 
		i=flags; 
		doneData(); 
		flags=i&~EMDATA_BUSY;
//		printf("old ok\n"); 
		return fft; 
	}
	fft->setSize(nx+2,ny,nz);
	dat=fft;
//	printf("old fft\n");
}
else {
//	printf("new fft\n");
	dat=copyHead();
//	sprintf(s,"FFT %s",name);
//	dat->setName(s);
	fft=dat;
	fft->setSize(nx+2,ny,nz);
}


if (ny==1) dim=1;
else if (nz==1) dim=2;
else dim=3;

cplan=makeplan(dim,nx,ny,nz,1);

// 1d
if (ny==1) {
	j=0;
	k=1;
	d=dat->getData();
	rfftwnd_one_real_to_complex(cplan,(fftw_real *)rdata,(fftw_complex *)d);
	dat->doneData();
}
// 2d
else if (nz==1) {
	d=dat->getData();

	rfftwnd_one_real_to_complex(cplan,(fftw_real *)rdata,(fftw_complex *)d);

	// This 'rotates' the data vertically by ny/2
	l=ny/2*(nx+2);
	for (i=0; i<ny/2; i++) {
		for (j=0; j<nx+2; j++) {
			k=j+i*(nx+2);
			re=d[k];
			d[k]=d[k+l];
			d[k+l]=re;
		}
	}
	dat->doneData();
}
// 3d
else {
	d=dat->getData();
	
	rfftwnd_one_real_to_complex(cplan,(fftw_real *)rdata,(fftw_complex *)d);

	// in the y,z plane, swaps quadrants I,III and II,IV
	// this is the 'rotation' in y and z
	t=(char *)malloc(sizeof(float)*(nx+2));

	k=(nx+2)*ny*(nz+1)/2;
	l=(nx+2)*ny*(nz-1)/2;
	jj=(nx+2)*sizeof(float);
	for (j=ii=0; j<nz/2; j++) {
		for (i=0; i<ny; i++,ii+=nx+2) {
			memcpy(t,d+ii,jj);
			if (i<ny/2) {
				memcpy(d+ii,d+ii+k,jj);
				memcpy(d+ii+k,t,jj);
			}
			else {
				memcpy(d+ii,d+ii+l,jj);
				memcpy(d+ii+l,t,jj);
			}
		}
	}
	dat->doneData();
	free(t);
}

fft=dat;
dat->multConst(1.0/(nx*ny*nz));
dat->update();
dat->flags|=EMDATA_COMPLEX|EMDATA_RI;
i=flags;	// we didn't acutally change 'this', so we don't need an update
doneData();
flags=i&~(EMDATA_NEWFFT|EMDATA_BUSY);
return dat;
}

EMData *EMData::doIFT()		// obvious, source must be R,I
{
EMData *dat;
float *d,re,im,norm;
int i,j,k,l,ii,jj,fl,dim;
char s[80],*t;
mrcH *newh;
rfftwnd_plan cplan;

if (!(flags&EMDATA_COMPLEX)) return NULL;

//printf("%d,%d,%d %d,%d,%d\n",nx,ny,nz,npfa(nx-2),npfa(ny),npfa(nz));

if ((nx-2)!=npfa(nx-2) || (ny!=npfa(ny)&&ny!=1) || (nz!=npfa(nz)&&nz!=1)) { error("Invalid IFT size !",ERR_ERROR); return NULL; }

getData();

if (fft) {
	if (!(flags&EMDATA_NEWFFT)) { doneData(); printf("Old fft\n"); return fft; }
	fft->setSize(nx,ny,nz);
	dat=fft;
	printf("old pointer\n");
}
else {
	dat=copyHead();
	if (strlen(name)>75) name[76]=0;
	sprintf(s,"IFT %s",name);
	dat->setName(s);
	fft=dat;
	fft->setSize(nx,ny,nz);
}

ap2ri();

if (ny==1) dim=1;
else if (nz==1) dim=2;
else dim=3;

cplan=makeplan(dim,nx-2,ny,nz,0);


// 1d
if (ny==1) {
	j=nx-2;
	k=1;
	d=dat->getData();
	rfftwnd_one_complex_to_real(cplan, (fftw_complex *)rdata,(fftw_real *)d);
	dat->doneData();
}
// 2d
else if (nz==1) {
	d=dat->getData();

	memcpy((char *)d,(char *)rdata,nx*ny*sizeof(float));

	// This 'rotates' the data vertically by ny/2
	l=ny/2*(nx);
	for (i=0; i<ny/2; i++) {
		for (j=0; j<nx; j++) {
			k=j+i*(nx);
			re=d[k];
			d[k]=d[k+l];
			d[k+l]=re;
		}
	}

	
	rfftwnd_one_complex_to_real(cplan, (fftw_complex *)d,(fftw_real *)d);

	for (i=1; i<ny; i++) memcpy((char *)&d[i*(nx-2)],(char *)&d[i*nx],(nx-2)*sizeof(float));
	dat->doneData();


//printf("size -> %d %d %d\n",nx-2,ny,nz);
}
// 3d
else {
	d=dat->getData();

	memcpy((char *)d,(char *)rdata,nx*ny*nz*sizeof(float));

	// in the y,z plane, swaps quadrants I,III and II,IV
	// this is the 'rotation' in y and z
	t=(char *)malloc(sizeof(float)*(nx+2));

	k=nx*ny*(nz+1)/2;
	l=nx*ny*(nz-1)/2;
	jj=nx*sizeof(float);
	for (j=ii=0; j<nz/2; j++) {
		for (i=0; i<ny; i++,ii+=nx) {
			memcpy(t,d+ii,jj);
			if (i<ny/2) {
				memcpy(d+ii,d+ii+k,jj);
				memcpy(d+ii+k,t,jj);
			}
			else {
				memcpy(d+ii,d+ii+l,jj);
				memcpy(d+ii+l,t,jj);
			}
		}
	}
	free(t);
		
	rfftwnd_one_complex_to_real(cplan, (fftw_complex *)d,(fftw_real *)d);

	for (i=0; i<ny*nz; i++) {
		j=i*(nx-2);
		k=i*nx;
		for (l=j; l<j+nx-2; l++,k++) d[l]=d[k];
	}
	dat->doneData();
}

fft->setSize(nx-2,ny,nz);

dat->update();
dat->setComplex(0);

i=flags;	// we didn't acutally change 'this', so we don't need an update
doneData();
flags=i&~(EMDATA_NEWFFT|EMDATA_BUSY);

return dat;
}

// If a program wishes to modify the 
void EMData::gimmeFFT()
{
fft=NULL;
}

// Calculates a CCF. If with==NULL, does mirror ACF
EMData *EMData::calcCCF(EMData *with) {
EMData *f1,*f2,*cf;
float *rdata1,*rdata2;
int i,j,nx2,ny2,nz2;
float re,im,norm;

// get the ffts
f1=doFFT();
if (!f1) return NULL;
f1->ap2ri();			// make sure image is real/imag

if (with) {
	cf=with->doFFT();	// this is where the result will go
	if (!cf) return NULL;
	cf=cf->copy(0,0);		// get a copy so we can modify it freely
	cf->ap2ri();		//make sure image is real/imag
}
else cf=f1->copy(0,0);		// already r/i, so don't worry

//writeMRC("b1.img");
//with->writeMRC("b2.img");

// make sure the images are the same size
nx2=cf->xSize();
ny2=cf->ySize();
nz2=cf->zSize();
if (with && (f1->xSize()!=nx2 || f1->ySize()!=ny2 || f1->zSize()!=nz2)) 
	{ error("CCF: Images must be the same size.",ERR_ERROR); return NULL; }

//f1->writeMRC("a1.img");
//cf->writeMRC("a2.img");

rdata1=f1->getDataRO();
rdata2=cf->getData();

if (with) {    
//	norm=nx2*ny2*nx2*ny2*Sigma()*with->Sigma();
	norm=nx2*ny2*nz2*nx2*ny2*nz2;
	for (i=0; i<nx2*ny2*nz2; i+=2) {
		re=(rdata1[i]*rdata2[i]+rdata1[i+1]*rdata2[i+1])/norm;
		im=(rdata1[i+1]*rdata2[i]-rdata1[i]*rdata2[i+1])/norm;
		rdata2[i]=re;
		rdata2[i+1]=im;
	}
}
else {                           
//	norm=nx2*ny2*nx2*ny2*(Mean()+Sigma())*(Mean()+Sigma());
	norm=nx2*ny2*nz2*nx2*ny2*nz2;
	for (i=0; i<nx2*ny2*nz2; i+=2) {
		re=(rdata1[i]*rdata2[i]-rdata1[i+1]*rdata2[i+1])/norm;
		im=(rdata1[i+1]*rdata2[i]+rdata1[i]*rdata2[i+1])/norm;
		rdata2[i]=re;
		rdata2[i+1]=im;
	}
}

f1->doneData();		// ok done with the original data
cf->doneData();
f2=cf->doIFT();		// the result
cf->gimmeFFT();		// we're keeping the ift, don't want it freed
delete cf;			// cf is a copy, so get rid of it
f2->setName("CCF");
f2->setPath("/tmp/eman.ccf");
//f2->writeMRC("af.img");

//if (f2->Sigma()) f2->multConst(1.0/f2->Sigma());

return f2;
}

// move center to center of mass (more or less)
// eliminates parent if any, integer shifts only
void EMData::cmCenter(int intonly) {
float xm=0,ym=0,zm=0,m=0;
int i,j,k;

normalize();

getData();
for (i=0; i<nx; i++) {
	for (j=0; j<ny; j++) {
		for (k=0; k<nz; k++) {
//			if (hypot(i-nx/2.0,j-ny/2.0)>=nx/2) continue;
			if (rdata[i+nx*j+k*nx*ny]<Sigma()*.75+Mean()) continue; // may help since protein is denser
			xm+=(float)i*rdata[i+nx*j+k*nx*ny];
			ym+=(float)j*rdata[i+nx*j+k*nx*ny];
			zm+=(float)k*rdata[i+nx*j+k*nx*ny];
			m+=rdata[i+nx*j+k*nx*ny];
		}
	}
}	
doneData();
xm/=m;
ym/=m;
zm/=m;

//printf("%f\t%f\t%f\n",xm,ym,zm);

if (intonly) {
	dx=-(floor(xm+.5)-nx/2);
	dy=-(floor(ym+.5)-ny/2);
	if (nz>1) dz=-(floor(zm+.5)-nz/2);
}
else {
	dx=-(xm-nx/2);
	dy=-(ym-ny/2);
	if (nz>1) dz=-(zm-nz/2);
}
parent=NULL;
rotateAndTranslate();
}

// Intonly will only make whole integer translations to avoid interpolation
// maximum allowed translation +-nx/4
float EMData::transAlign(EMData *with,int useparent,int intonly,int maxshift, float *peak) {
EMData *cf;
float *rdata3;
int i,j,ii,jj,nx2,flag,acf=0;
float re,im,r2,f,to,am,ph,t;
float tx,ty,ret;
char *tmp;

if (with && (with->xSize()!=nx || with->ySize()!=ny)) { error("ALIGN: images must be same size",ERR_ERROR); return -1; }
//if (!with) printf("Warning: ACF\n");

if (useparent&&parent) { cf=parent->calcCCF(with); }
else cf=calcCCF(with);

if (maxshift<=0) maxshift=ny/4;
if (maxshift>nx/2-1) maxshift=nx/2-1;
if (maxshift>ny/2-1) maxshift=ny/2-1;

//writeIMAGIC("ccf.img",-1);
//with->writeIMAGIC("ccf.img",-1);
//cf->writeIMAGIC("ccf.img",-1);

cf->toCorner();				// shift center to center from corner
cf->applyMask(0,-1);
rdata3=cf->getData();

/*
t=cf->Max()-cf->Sigma();
tx=ty=f=0;
for (j=ny/2-maxshift; j<ny/2+maxshift; j++) {
	for (i=nx/2-maxshift; i<nx/2+maxshift; i++) {
		if (rdata3[i+j*nx]>t) {
			tx+=rdata3[i+j*nx]*i;
			ty+=rdata3[i+j*nx]*j;
			f+=rdata3[i+j*nx];
		}
	}
}

ii=floor(tx/f+.5);
jj=floor(ty/f+.5);
*/

if (cf->Mean()-cf->Min()<cf->Max()-cf->Mean()) { flag=1;  }
else { flag=-1; }

// locate the peak
ii=jj=nx/2;
r2=-1.0e20;
f=1.0e20;
for (j=ny/2-maxshift; j<ny/2+maxshift; j++) {
	for (i=nx/2-maxshift; i<nx/2+maxshift; i++) {
		if (rdata3[i+j*nx]*flag>r2) { r2=rdata3[i+j*nx]*flag; ii=i; jj=j; }
		if (rdata3[i+j*nx]<f) f=rdata3[i+j*nx];
	}
}

if (peak) *peak=r2;

// now find the peak to subpixel resolution
/*
tx=-p5peak(rdata3[ii-2+jj*nx],rdata3[ii-1+jj*nx],rdata3[ii+jj*nx],
	rdata3[ii+1+jj*nx],rdata3[ii+2+jj*nx])+(float)(nx/2-ii);

ty=-p5peak(rdata3[ii+jj*nx-2*nx],rdata3[ii+jj*nx-nx],rdata3[ii+jj*nx],
	rdata3[ii+jj*nx+nx],rdata3[ii+jj*nx+nx*2])+(float)(ny/2-jj);

if (intonly) { tx=floor(tx+.5); ty=floor(ty+.5); }
*/

// no subpixel at the moment
tx=nx/2-ii;
ty=ny/2-jj;


if (useparent&&parent) {
	ret=hypot(tx-dx,ty-dy);
	//printf("\t%7.3f \t%7.3f \t%5.3g\n",tx-dx,ty-dy,r2);
}
else {
	ret=hypot(tx,ty);
	//printf("\t%7.3f \t%7.3f \t%5.3g\n",tx,ty,r2);
}

/*if (!parent||useparent) {
	dx=tx;
	dy=ty;
	euler.phi()=0;
}
else {
	dx+=tx*cos(euler.phi())+ty*sin(euler.phi());
	dy+=-tx*sin(euler.phi())+ty*cos(euler.phi());
}*/
dx=tx;
dy=ty;
setRAlign(0,0,0);

if (!with) { dx/=2.0; dy/=2.0; }	// aligning to self, only 1/2 shift

if (intonly) { dx=floor(dx+.5); dy=floor(dy+.5); }

cf->doneData();
delete cf;

return ret;
}

// Intonly will only make whole integer translations to avoid interpolation
// maximum allowed translation +-nx/4
float EMData::transAlign3d(EMData *with,int useparent,int intonly) {
EMData *cf;
float *rdata3;
int i,j,k,ii,jj,kk,nx2,flag,acf=0;
float re,im,r2,f,to,am,ph;
float tx,ty,tz,ret;
char *tmp;

if (with && (with->xSize()!=nx || with->ySize()!=ny || with->zSize()!=nz)) { error("ALIGN: images must be same size",ERR_ERROR); return -1; }
if (!with) printf("Warning: ACF\n");

if (useparent&&parent) { cf=parent->calcCCF(with); }
else cf=calcCCF(with);

cf->toCorner();				// shift center to center from corner
rdata3=cf->getData();

if (cf->Mean()-cf->Min()<cf->Max()-cf->Mean()) { flag=1; /*printf("positive CCF ");*/ }
else { flag=-1; /*printf("negative CCF ");*/ }

// locate the peak
ii=jj=nx/2;
r2=-1.0e20;
f=1.0e20;
for (k=nx/4; k<3*nx/4; k++) {
	for (j=ny/4; j<3*ny/4; j++) {
		for (i=nx/4; i<3*nx/4; i++) {
			if (rdata3[i+j*nx+k*nx*ny]*flag>r2) { r2=rdata3[i+j*nx+k*nx*ny]*flag; ii=i; jj=j; kk=k; }
			if (rdata3[i+j*nx+k*nx*ny]<f) f=rdata3[i+j*nx+k*nx*ny];
		}
	}
}

// no subpixel at the moment
tx=nx/2-ii;
ty=ny/2-jj;
tz=nz/2-kk;

if (useparent&&parent) {
	ret=hypot3(tx-dx,ty-dy,tz-dz);
	//printf("\t%7.3f \t%7.3f \t%5.3g\n",tx-dx,ty-dy,r2);
}
else {
	ret=hypot3(tx,ty,tz);
	//printf("\t%7.3f \t%7.3f \t%5.3g\n",tx,ty,r2);
}

dx=tx;
dy=ty;
dz=tz;

if (!with) { dx/=2.0; dy/=2.0; dz/=2.0; }	// aligning to self, only 1/2 shift

cf->doneData();
delete cf;

return ret;
}


// This is a translationally independant rotational alignment (in theory)
// The data sets are padded by a factor of 2 and transformed. The center
// 1/4 of the transform is then rotationally correlated
// result does not destiguish between 180 degree rotations
float EMData::rotAlignTI(EMData *with,int usedot) {
EMData *d1,*d2;
float a,da,t,besta=0;
//float best=0;
float best=1.0e20;	// arbitrary big number
float worst=-1.0e20;
int xs;

// make 'rotational footprints' of the images
if (parent) d2=parent->makeRFP(); else d2=makeRFP();
if (with->Parent()) d1=with->Parent()->makeRFP(); else d1=with->makeRFP();

if (usedot) { best*=-1.0; worst*=-1.0; }
xs=d1->xSize();
da=2.0/(float)xs;

//da/=6.0;

// copy the second so we can perform multiple rotations without degredation
d2=d2->copy();

for (a=da*1.5; a<PI; a+=da*3.0) {
	d2->setRAlign(a,0,0);
	d2->rotateAndTranslate();

	if (usedot) {
		t=d1->dot(d2);
		if (t>best) { best=t; besta=a; }
		if (t<worst) worst=t;
	} else {
		t=d1->lcmp(d2);
		if (t<best) { best=t; besta=a; }
		if (t>worst) worst=t;
	}
		
//d2->writeIMAGIC("tst.hed",-1);
//	printf("R %f\t%f\n",a*180.0/PI,t);
}

for (a=besta-da*3.0; a<=besta+da*3.0; a+=da/2.0) {
	d2->setRAlign(a,0,0);
	d2->rotateAndTranslate();

	if (usedot) {
		t=d1->dot(d2);
		if (t>best) { best=t; besta=a; }
		if (t<worst) worst=t;
	} else {
		t=d1->lcmp(d2);
		if (t<best) { best=t; besta=a; }
		if (t>worst) worst=t;
	}
	
//	printf("r %f\t%f\n",a*180.0/PI,t);
}

//delete d2->Parent();	
delete d2;
//delete d1;

setRAlign(besta,0,0);

//printf("%f - %f \t %f\n",worst,best,besta*180.0/PI);

return best;
}

// shrinks a factor by an integral factor using a local median filter
// note that sf MUST be a factor of the image size
void EMData::medianShrink(int sf) {
int i,j,l,i2,j2,l2,k,nx2,ny2;
float *old,*mbuf,f;

if (sf<=1) return;
if ((nx%sf) || (ny%sf)) { printf("Image size not divisible by shrink size\n"); return; }
if (nz>1) { 
	if (nz%sf) { printf("Image size not divisible by shrink size\n"); return; }
	mbuf=(float *)malloc(sf*sf*sf*sizeof(float));
	old=(float *)malloc(nx*ny*nz*sizeof(float));
	memcpy(old,getDataRO(),nx*ny*nz*sizeof(float));
	doneData();	
	nx2=nx;
	ny2=ny;
	setSize(nx/sf,ny/sf,nz/sf);
	getData();

	// note that nx,ny,nz are now /sf
	for (l=0; l<nz; l++) {
		for (j=0; j<ny; j++) {
			for (i=0; i<nx; i++) {
		// read out the data for 1 result pixel median filter
				k=0;
				for (l2=l*sf; l2<(l+1)*sf; l2++) {
					for (j2=j*sf; j2<(j+1)*sf; j2++) {
						for (i2=i*sf; i2<(i+1)*sf; i2++,k++) {
							mbuf[k]=old[i2+j2*nx2+l2*nx2*ny2];
						}
					}
				}
		
				// sort into the 1st half of the array
				for (k=0; k<sf*sf*sf/2+1; k++) {
					for (i2=k+1; i2<sf*sf*sf; i2++) {
						if (mbuf[i2]<mbuf[k]) { f=mbuf[i2]; mbuf[i2]=mbuf[k]; mbuf[k]=f; }
					}
				}
		
				rdata[i+j*nx+l*nx*ny]=mbuf[sf*sf*sf/2];
			}
		}
	}
	doneData();

	free(old);
	free(mbuf);

	return;
}

mbuf=(float *)malloc(sf*sf*sizeof(float));
old=(float *)malloc(nx*ny*sizeof(float));
memcpy(old,getDataRO(),nx*ny*sizeof(float));
doneData();
nx2=nx;
setSize(nx/sf,ny/sf,1);
getData();

// note that nx is now /sf
for (i=0; i<nx; i++) {
	for (j=0; j<ny; j++) {
		// read out the data for 1 result pixel median filter
		k=0;
		for (i2=i*sf; i2<(i+1)*sf; i2++) {
			for (j2=j*sf; j2<(j+1)*sf; j2++,k++) {
				mbuf[k]=old[i2+j2*nx2];
			}
		}
		
		// sort into the 1st half of the array
		for (k=0; k<sf*sf/2+1; k++) {
			for (i2=k+1; i2<sf*sf; i2++) {
				if (mbuf[i2]<mbuf[k]) { f=mbuf[i2]; mbuf[i2]=mbuf[k]; mbuf[k]=f; }
			}
		}
		
		rdata[i+j*nx]=mbuf[sf*sf/2];
	}
}
doneData();

free(old);
free(mbuf);

return;
}

// Generates a small version of the image used for rotational alignment
EMData *EMData::makeRFP() {
EMData *tmp,*tmp2;
int cs;

if (!(flags&EMDATA_NEWRFP) && rfp) { return rfp; }
if (nx&1) return NULL;
if (rfp) delete rfp;

cs=((nx/4)&0xfffff8);
//while ((nx+2*cs)%8!=0) cs+=2;
if (nz==1) tmp2=clip(-cs,-cs,nx+2*cs,ny+2*cs);
else tmp2=clip(-cs,-cs,-cs,nx+2*cs,ny+2*cs,nz+2*cs);
tmp2->applyMask(nx/2,4);
tmp=tmp2->calcCCF(tmp2);
delete tmp2;

//tmp=calcCCF(this);
tmp->toCorner();

if (nz==1) {
	tmp2=tmp->clip(cs,cs,nx,ny);
}
else {
	tmp2=tmp->clip(cs,cs,cs,nx,ny,nz);
}

if (nz==1) tmp2->radialSubtract();
tmp2->applyMask(-1,4);

rfp=tmp2;
flags&=~EMDATA_NEWRFP;
delete tmp;

if (nz==1 && isvalidfft(nx/2)) rfp->medianShrink(2);

return rfp;
}

/*
EMData *EMData::makeRFP() {
EMData *ret,*tmp,*tmpf;
float *dat,*ddat,sig;
int i,j,k,s,o,nnx,nny;

if (!(flags&EMDATA_NEWRFP) && rfp) { return rfp; }
if (rfp) delete rfp;

// volume RFP
if (nz>1) {
	s=ny/2;

	ret=new EMData;
	ret->setSize(s,s,s);
	ret->zero();
	ddat=ret->getData();

	// pads the image to twice its size for better fft
//	tmp=clip(-ny/2,-ny/2,-ny/2,ny*2,ny*2,ny*2);
	tmp=copy();
	tmp->normalize();
	tmp->applyMask(ny/2,0);
	tmp->toCorner();
	tmpf=tmp->doFFT();
	tmpf->ri2ap();
	nnx=tmpf->xSize();
	nny=tmpf->ySize();
	dat=tmpf->getData();
	dat[ny*(ny+2)/2+ny*ny*(ny+2)/2]=0;	// complex origin ->0


	// right half
	for (i=s/2; i<s; i++) {
		for (j=1; j<s; j++) {
			for (k=1; k<s; k++) {
				ddat[i+j*s+k*s*s]=dat[(i-s/2)*2+(j+ny/4)*nnx+(k+ny/4)*nnx*nny];
			}
		}
	}

	// left half is point mirrored
	for (i=1; i<s/2; i++) {
		for (j=1; j<s; j++) {
			for (k=1; k<s; k++) {
				ddat[i+j*s+k*s*s]=ddat[(s-i)+(s-j)*s+(s-k)*s*s];
			}
		}
	}

	tmpf->doneData();
	delete tmp; //also gets tmpf

	ret->doneData();
	ret->applyMask(-1,0);
	rfp=ret;
	flags&=~EMDATA_NEWRFP;
	
//	ret->writeMRC("rfp.mrc");
	return ret;
}

// 2d images
s=ny/2;
o=ny/4*(ny+2);

ret=new EMData;
ret->setSize(s,s,1);
ret->zero();
ddat=ret->getData();

// pads the image to twice its size for better fft
//tmp=clip(-ny/2,-ny/2,ny*2,ny*2);
tmp=copy();
tmp->normalize();
//tmp->applyMask(ny/2,0);
tmp->radialSubtract();
tmp->toCorner();
tmpf=tmp->doFFT();
tmpf->ri2ap();
dat=tmpf->getData();
dat[(ny+2)*ny/2]=0;	// complex origin ->0


// right lower half
for (i=s/2; i<s; i++) {
	for (j=1; j<s; j++) {
		ddat[i+j*s]=dat[(i-s/2)*2+j*(ny+2)+o];
	}
}

// left half is point mirrored
for (i=1; i<s/2; i++) {
	for (j=1; j<s; j++) {
		ddat[i+j*s]=ddat[(s-i)+(s-j)*s];
	}
}


tmpf->doneData();
delete tmp; //also gets tmpf

ret->doneData();
//ret->applyMask(-1,4);
rfp=ret;
flags&=~EMDATA_NEWRFP;

//ret->multConst(1.0/sig);
//ret->writeIMAGIC("rfp.hed",-1);

//ret->normalize();
//ret->applyMask(-1,4);
//ret->radialSubtract();
return ret;
}
*/


#define RALRAD	5
#define RALANG	8

// This does a rotational alignment using circular harmonics
float EMData::rotAlignCH(EMData *with,int irad,int orad) {
int i,j,x,y,na;
float cen,wid,a,r;
EMData *d1,*d2;
float *dat1,*dat2;

if (nx!=ny) return(-1.0);

if (irad<=0) irad=6;
if (orad<=2 || orad>nx/2) orad=nx/2-2;

//printf("%%%%%%%%%%%%%%%%%\n");

// here we regenerate circular harmonics if necessary
if (nx!=rals || irad!=rali || orad!=ralo) {
//	printf("Recalc basis %d %d %d\n",rals,ralo,rali);
	for (i=0; i<ralfp->count(); i++) 
		delete (EMData *)(ralfp->objectAt(i));
	ralfp->empty();
	
	rals=nx;
	ralo=orad;
	rali=irad;
	
	wid=(ralo-rali)/(RALRAD+1);
	for (i=0; i<RALANG; i++) {
		for (j=0; j<RALRAD; j++) {
			cen=(j+i)*wid+rali;
			na=1<<(i+1);
			if (cen*PI<na) continue;
		
			d1=new EMData;
			d1->setSize(nx,ny,1);
			d2=new EMData;
			d2->setSize(nx,ny,1);
			
			d1->ufloat=cen;
			d1->unum2=na;
			
			dat1=d1->getData();
			dat2=d2->getData();

			for (x=0; x<nx; x++) {
				for (y=0; y<ny; y++) {
					if (y==ny/2 && x==nx/2) a=0;
					else a=atan2((float)(y-ny/2),(float)(x-nx/2));
					r=hypot((float)(y-ny/2),(float)(x-nx/2))-cen;
					dat1[x+y*nx]=sin(a*na)*exp(-r*r/(wid*wid/2.4));
					dat2[x+y*nx]=cos(a*na)*exp(-r*r/(wid*wid/2.4));
				}
			}

			d1->doneData();
			d2->doneData();
			
			ralfp->addObject(d1);
			ralfp->addObject(d2);

		}
	}
}

// OK now do the alignment
float ta,tb,a1,p1,a2,p2;
float ca,aa=0,ndot=0;
float dl,ep;

for (i=0; i<ralfp->count(); i+=2) {
	d1=(EMData *)ralfp->objectAt(i);
	d2=(EMData *)ralfp->objectAt(i+1);

	ta=dot(d1);
	tb=dot(d2);
	a1=hypot(ta,tb);
	if (ta==0 && tb==0) p1=0;
	else p1=atan2(ta,tb);
	
	ta=with->dot(d1);
	tb=with->dot(d2);
	a2=hypot(ta,tb);
	if (ta==0 && tb==0) p2=0;
	else p2=atan2(ta,tb);
	
// this complicated little mess could probably be simplified  :^)
	ca=p2-p1;							// relative phase of 2 images
	ca=ca-floor(ca/(2.0*PI))*2.0*PI;	// range 0 - 2 PI
	if (ca>PI) ca-=2.0*PI;				// range -PI to PI
	if (ndot>0) {
		dl=PI*2.0/(float)d1->unum2;			// size of one oscillation in radians
		ep=aa/ndot;
		ep=(ep-floor(ep/dl)*dl)*2.0*PI/dl;	// phase predicted by current average
		ca-=ep;								// phase relative to current average
		ca=ca-floor(ca/(2.0*PI))*2.0*PI;	// range 0 - 2 PI
		if (ca>PI) ca-=2.0*PI;				// range -PI to PI
		ca=aa/ndot+ca/(float)d1->unum2;
//		printf("%6.2f  %4d\t%4.1f  %4.1f  %4.1f\n",d1->ufloat,d1->unum2,ca*180.0/PI,aa/ndot*180.0/PI,a2);
	}
//	else printf("%6.2f  %4d\t%4.1f\n",d1->ufloat,d1->unum2,ca*180.0/PI);
//	printf("%f\t%d\n",ca*180.0/PI,i);

	aa+=ca*d1->ufloat*a2;
	ndot+=d1->ufloat*a2;

	printf("%f\t%d\n",ca*180.0/PI,i);
}
printf("%f\t%d\n",aa/ndot*180.0/PI,i+5);
setRAlign(aa/ndot,0,0);
return aa/ndot;
}

// Doesn't use irad or orad currently
float EMData::rotAlign(EMData *with,int irad,int orad) {
EMData *r1,*r2;
float a,ba,bd,d,ba2,dam;

r1=makeRFP()->copy();
r2=with->makeRFP();

dam=atan2(1.0,nx/2.0)/2.0;

ba=0;
bd=-1.0e10;
for (a=PI/36.0; a<PI; a+=PI/18.0) {
	r1->setTAlign(0,0,0);
	r1->setRAlign(a,0,0);
	r1->rotateAndTranslate();
	d=r1->dot(r2);
	if (d>bd) { bd=d; ba=a; }
}

ba2=ba;
for (a=ba2-PI/18.0; a<ba2+PI/18.0; a+=dam) {
	r1->setTAlign(0,0,0);
	r1->setRAlign(a,0,0);
	r1->rotateAndTranslate();
	d=r1->dot(r2);
	if (d>bd) { bd=d; ba=a; }
}

delete r1;
setRAlign(ba,0,0);
return ba;
}

// Despite interpolation, ROTSMP=360 is really only good to ~ +-.5 deg
//#define ROTSMP 360
/*
#define ROTDA (2.0*PI/ROTSMP)

float EMData::rotAlign(EMData *with,int irad,int orad) {
int i,j,k,l,nx2;
int ROTSMP;
float *dat,*dat2,*sum;
EMData *dd;
List *list;
float min,max,max2,norm,lim,ret;
char s[80];
// maximum radius to use
nx2=nx*9/20;	// 90% rad
if (fabs(dx)<fabs(dy)) lim=fabs(dy); else lim=fabs(dx);
nx2-=(int)floor(lim);

if (irad<1) irad=8;
if (orad<irad) orad=nx/2;
if (nx2>orad) nx2=orad;
ROTSMP=(int)floor(orad*2*PI/4.0)*4;
dat=(float *)malloc((ROTSMP+2)*sizeof(float));
dat2=(float *)malloc((ROTSMP+2)*sizeof(float));
sum=(float *)malloc((ROTSMP+2)*sizeof(float));

memset((char *)sum,0,sizeof(float)*ROTSMP);		// clear the sum array before we start

// split into radial 'shells' 6 pixels wide
for (i=irad; i<nx2; i+=4) {
	calcAzDist(ROTSMP,0,ROTDA,dat,i,i+6);			// azimuthal distributions in this slice
	with->calcAzDist(ROTSMP,0,ROTDA,dat2,i,i+6);

//	if (i>nx2-6) {
//		sprintf(s,"a1.%d.dat",i);
//		save_data(0,ROTDA,dat,ROTSMP,s);
//		sprintf(s,"a2.%d.dat",i);
//		save_data(0,ROTDA,dat2,ROTSMP,s);
//	}

	l=1;
	j=0;
	k=ROTSMP;
	odfft_(dat,&k,&l,&j);		// 1d fft of each
	odfft_(dat2,&k,&l,&j);
	
	dat[0]=dat[1]=0;	// This eliminates any circularly symmetric part
	dat2[0]=dat2[1]=0;	// (so each ring only contributes the appropriate amount)
	
	// CCF in Fourier space
	for (j=0; j<ROTSMP+2; j+=2) {		
		max=dat[j]*dat2[j]+dat[j+1]*dat2[j+1];
		max2=dat[j+1]*dat2[j]-dat2[j+1]*dat[j];
		dat[j]=max;
		dat[j+1]=max2;
	}

	// iFFT to get the CCF
	odfft_(dat,&k,&l,&l);
	
	norm=ROTSMP*ROTSMP*(4.0*Sigma())*(4.0*with->Sigma());
	for (j=0; j<ROTSMP; j++)
		sum[j]+=dat[j]*(float)i/norm;
}

//save_data(0,ROTDA,sum,ROTSMP,"test.rcf.dat");

// find the peak value
max=-1.0e25;
j=-1;
for (i=0; i<ROTSMP; i++) if (sum[i]>max) { max=sum[i]; j=i; }
ret=max;
//printf("Max at %d, %g, %g\n",j,dat[j]/PI*180.0,max);

// use neighboring values to get a more accurate peak position
if (j==0) max=sum[ROTSMP-1]; else max=sum[j-1];
if (j==ROTSMP-1) max2=sum[0]; else max2=sum[j+1];
if (sum[j]-max>sum[j]-max2) min=(max-max2)/(max-sum[j]);
else min=(max-max2)/(max2-sum[j]);
min+=j;
min*=ROTDA;
while (min>PI) min-=PI*2.0;
while (min<-PI) min+=PI*2.0;

//printf("Max at %g (%g %g %g)\n",min*180.0/PI,max,sum[j],max2);
//printf("Max at %g %g\n",min,sum[j]);


//if (!parent) {
//	dx=0;
//	dy=0;
//	euler.phi()=min;
//}
//else {
//	euler.phi()+=min;
//}

dx=0;
dy=0;
setRAlign(min,0,0);

free(sum);
free(dat);
free(dat2);

return ret;
}
*/

void EMData::calcRCF(EMData *with,float *sum,int NS) {
int i,j,k,l,nx2;
float *dat,*dat2;
float DA=(2.0*PI/NS);
EMData *dd;
List *list;
float min,max,max2,norm,lim,ret;
char s[80];

rfftwnd_plan fplan=rfftwnd_create_plan(1, &NS,FFTW_REAL_TO_COMPLEX,
	FFTW_MEASURE|FFTW_IN_PLACE|FFTW_USE_WISDOM);
rfftwnd_plan rplan=rfftwnd_create_plan(1, &NS,FFTW_COMPLEX_TO_REAL,
	FFTW_MEASURE|FFTW_IN_PLACE|FFTW_USE_WISDOM);

dat=(float *)malloc((NS+2)*sizeof(float));
dat2=(float*)malloc((NS+2)*sizeof(float));

// maximum radius to use
nx2=nx*9/20;	// 90% rad
if (fabs(dx)<fabs(dy)) lim=fabs(dy); else lim=fabs(dx);
nx2-=(int)floor(lim);

memset((char *)sum,0,sizeof(float)*NS);		// clear the sum array before we start

// split into radial 'shells' 6 pixels wide
for (i=8; i<nx2; i+=6) {
	calcAzDist(NS,0,DA,dat,i,i+6);			// azimuthal distributions in this slice
	with->calcAzDist(NS,0,DA,dat2,i,i+6);
	
	rfftwnd_one_real_to_complex(fplan, dat,(fftw_complex *)dat);
	rfftwnd_one_real_to_complex(fplan, dat2,(fftw_complex *)dat2);
	
	// CCF in Fourier space
	for (j=0; j<NS+2; j+=2) {		
		max=dat[j]*dat2[j]+dat[j+1]*dat2[j+1];
		max2=dat[j+1]*dat2[j]-dat2[j+1]*dat[j];
		dat[j]=max;
		dat[j+1]=max2;
	}

	// iFFT to get the CCF
	rfftwnd_one_complex_to_real(rplan,(fftw_complex *)dat, dat);
	
	norm=NS*NS*(4.0*Sigma())*(4.0*with->Sigma());
	for (j=0; j<NS; j++)
		sum[j]+=dat[j]*(float)i/norm;
}

//save_data(0,ROTDA,sum,ROTSMP,"test.rcf.dat");
//writeMRC("x1.img");
//with->writeMRC("x2.img");

fftwnd_destroy_plan(fplan);
fftwnd_destroy_plan(rplan);

free(dat);
free(dat2);
return;
}

// Rotational alignment with possible flip, particles must be centered already
// (2d only)
EMData *EMData::RFAlign(EMData *with,EMData *flip,int irad,int orad) {
EMData *cop,*cop2;
float d1,d2;

//with->writeIMAGIC("tst.hed",-1);
cop=copy();
cop->rotAlign(with,irad,orad);
cop->rotateAndTranslate();
d1=cop->dot(with);

//printf("------\n");

if (flip) {
	cop2=copy();
	cop2->rotAlign(with,irad,orad);
	cop2->rotateAndTranslate();
	d2=cop2->dot(with);
}
else {
	cop2=copy();
	cop2->vFlip();
	cop2->rotAlign(with,irad,orad);
	cop2->rotateAndTranslate();
	d2=cop2->dot(with);
}

if (cop==NULL) return cop2;
if (cop2==NULL) return cop;

//with->writeIMAGIC("abc",-1);
//cop->writeIMAGIC("abc",-1);
//cop2->writeIMAGIC("abc",-1);

if (d1>d2) { cop->setFlipped(0); delete cop2; return cop; }
else { cop2->setFlipped(1); delete cop; return cop2; }

}

// Rotational and translational alignment with possible flip
// (2d only)
EMData *EMData::RTFAlign(EMData *with,EMData *flip,int usedot,int maxshift) {
EMData *cop,*cop2;
float d1,d2;

//with->writeIMAGIC("tst.hed",-1);
cop=RTAlign(with,usedot);

//printf("------\n");

if (flip) {
	cop2=flip->RTAlign(with,usedot,maxshift);
}
else {
	vFlip();
	cop2=RTAlign(with,usedot,maxshift);
	vFlip();
}

if (cop==NULL) return cop2;
if (cop2==NULL) return cop;

//with->writeIMAGIC("abc",-1);
//cop->writeIMAGIC("abc",-1);
//cop2->writeIMAGIC("abc",-1);


if (usedot) {
	d1=cop->dot(with);
	d2=cop2->dot(with);
//printf("%f %f\n",d1,d2);
	if (d1>d2) { cop->setFlipped(0); delete cop2; return cop; }
	else { cop2->setFlipped(1); delete cop; return cop2; }
}
else {
	d1=cop->lcmp(with);
	d2=cop2->lcmp(with);
	if (d1<d2) { cop->setFlipped(0); delete cop2; return cop; }
	else { cop2->setFlipped(1); delete cop; return cop2; }
}

}


// This will take 2 images that are aligned pretty well and do the
// rest of the job (usedot = -1 is a debugging flag)
void EMData::refineAlign(EMData *with,int usedot) {
float sda=euler.alt(),sdx=dx,sdy=dy,dda=atan(2.0/ny)/4.0,best,f;
int i,j=1;
float daz;

if (usedot) best=dot(with);
else best=-lcmp(with);

if (!parent) printf("Warning: no parent in refineAlign\n");
//printf("Start %f  %f,%f\n",sda*180.0/PI,sdx,sdy);
for (i=0; i<10 && j; i++) {
	j=0;	// break out if nothing changes
	
	for (daz=sda-dda*4.0; daz<=sda+dda*4.0; daz+=dda) {
		if (daz==sda) continue;
		setRAlign(daz,0,0);
		rotateAndTranslate();
		if (usedot) f=dot(with);
		else f=-lcmp(with);
		if (f>best) { best=f; sda=daz; j=1; break; }
	}
	
	for (dx=sdx-.25; dx<=sdx+.25; dx+=.24999) {
		for (dy=sdy-.25; dy<=sdy+.25; dy+=.24999) {
			rotateAndTranslate();
			if (usedot) f=dot(with);
			else f=-lcmp(with);
			if (f>best) { best=f; sdx=dx; sdy=dy; j=1; break; }
		}
	}			
if (usedot<0) printf("-- %f  %f,%f (%1.3g) \n",sda*180.0/PI,sdx,sdy,best);
}
setRAlign(sda,0,0);
dx=sdx;
dy=sdy;
rotateAndTranslate();

return;
}

// Used by RTFAlign
EMData *EMData::RTAlign(EMData *with,int usedot,int maxshift) {
EMData *cop,*cop2,*rfp,*rfpw;
int i;
float cda=0,cda2=0,cdx=0,cdy=0;

cop=copy();
cop2=copy();

rotAlign(with);
cda=cda2=alt();

// first align
cop->setRAlign(cda,0,0);
cop->setTAlign(0,0,0);
cop->rotateAndTranslate();

cop->transAlign(with,0,1,maxshift);
cdx= cop->dx*cos(cda)+cop->dy*sin(cda);
cdy=-cop->dx*sin(cda)+cop->dy*cos(cda);
cop->setRAlign(cda,0,0);
cop->setTAlign(cdx,cdy,0);
cop->rotateAndTranslate();



//cop->writeIMAGIC("abc.hed",-1);

// now try 180 deg rot
cda2+=PI;
cop2->setRAlign(cda2,0,0);
cop2->setTAlign(0,0,0);
cop2->rotateAndTranslate();

cop2->transAlign(with,0,1,maxshift);
cdx= cop2->dx*cos(cda2)+cop2->dy*sin(cda2);
cdy=-cop2->dx*sin(cda2)+cop2->dy*cos(cda2);
cop2->setRAlign(cda2,0,0);
cop2->setTAlign(cdx,cdy,0);
cop2->rotateAndTranslate();


//cop2->writeIMAGIC("abc.hed",-1);

if (cop->dot(with)>cop2->dot(with) ) {
/*	cop->rotAlign(with);
	cda+=cop->alt();
	cda=cda-floor(cda/(2.0*PI))*2.0*PI;
	cop->setRAlign(cda,0,0);
	cop->setTAlign(cdx,cdy,0);
	cop->rotateAndTranslate();*/

	delete cop2;
	return cop;
}

/*cop2->rotAlign(with);
cda2+=cop2->alt();
cda2=cda2-floor(cda2/(2.0*PI))*2.0*PI;
cop2->setRAlign(cda2,0,0);
cop2->setTAlign(cdx,cdy,0);
cop2->rotateAndTranslate();*/

delete cop;
return cop2;

//printf("%5.1f\n",cda*180.0/PI);
/*
// iterate translate/rotate 3 times
for (i=0; i<5; i++) {
	cop->transAlign(with,0,1,0);
	cdx+= cop->dx*cos(cda)+cop->dy*sin(cda);
	cdy+=-cop->dx*sin(cda)+cop->dy*cos(cda);
	cop->setRAlign(cda,0,0);
	cop->setTAlign(cdx,cdy,0);
	cop->rotateAndTranslate();
cop->writeIMAGIC("abc.hed",-1);

	cop->rotAlign(with);
printf("%5.1f\n",cop->alt()*180.0/PI);
	cda+=cop->alt();
	cda=cda-floor(cda/(2.0*PI))*2.0*PI;
	cop->setRAlign(cda,0,0);
	cop->setTAlign(cdx,cdy,0);
	cop->rotateAndTranslate();
cop->writeIMAGIC("abc.hed",-1);
printf("%5.1f\t%5.1f\t%5.2f\n",cdx,cdy,cda*180.0/PI);
}
*/
//return cop;
}

// Used by RTFAlign
/*
EMData *EMData::RTAlign(EMData *with,int usedot) {
EMData *cop,*cop2;
float v,a;
float tx,ty;
float dot,dot2;

//with->writeIMAGIC("tst.hed",-1);

cop=copy();
cop->setRAlign(0,0,0);
cop->setTAlign(0,0,0);
cop->rotAlignTI(with,usedot);
a=cop->euler.alt();

cop->rotateAndTranslate();
cop->transAlign(with,0,0,0);
cop->setRAlign(a,0,0);
tx=cop->Dx()*cos(a)+cop->Dy()*sin(a);
ty=-cop->Dx()*sin(a)+cop->Dy()*cos(a);
cop->setTAlign(tx,ty,0);
cop->rotateAndTranslate();
if (usedot) dot2=cop->dot(with);
else dot2=-cop->lcmp(with);
cop2=cop;
//cop->writeIMAGIC("tst.hed",-1);

cop=copy();
cop->setTAlign(0,0,0);
if (a>PI) a=a-PI,0,0; else a=a+PI;
cop->setRAlign(a,0,0);
cop->rotateAndTranslate();
cop->transAlign(with,0,0,0);
cop->setRAlign(a,0,0);
tx=cop->Dx()*cos(a)+cop->Dy()*sin(a);
ty=-cop->Dx()*sin(a)+cop->Dy()*cos(a);
cop->setTAlign(tx,ty,0);
cop->rotateAndTranslate();
if (usedot) dot=cop->dot(with);
else dot=-cop->lcmp(with);
//cop->writeIMAGIC("tst.hed",-1);

//printf("%f\t%f\n",dot2,dot);

if (dot>dot2) { delete cop2; return cop; }

delete cop;
return cop2;
}
*/

void EMData::zero() {
getData();
memset(rdata,0,nx*ny*nz*sizeof(float));
doneData();
update();
}

void EMData::one() {
if (isComplex()) {
	ap2ri();
	getData();
	for (int i=0; i<nx*ny*nz; i+=2) { rdata[i]=1.0; rdata[i+1]=0; }
}
else {
getData();
for (int i=0; i<nx*ny*nz; i++) rdata[i]=1.0;
}
doneData();
update();
}
