// helixhunter.c 1/99 Steve Ludtke
// This program will search a med-res (~8A) map for alpha helices

//N helixhunter
//F This program is used to locate possible helices in intermediate resolution maps.
//1
//P <input>	Input 3D map at 5-8A resolution
//P <output>	Output 3d map of helix positions
//P <a/pix>	Angstroms/pixel in 3D map. Must be <2.
//P [cyl]	Also outputs a 3d map containing the cylinder used in the search.
//P [int=<file>]	Outputs the raw correlation files describing helix locations.
//P [da=<angle>]	Angular step used in the hunt.
//P [keep=<sigmamult>]	Specifies the peak strength that defines a helix.
//P [filt=<res>]	Prefilter resolution in angstroms.
//D This program is classified as experimental at this point. Expect
//E a forthcoming publication with more details. We do not suggest
//E using it until it has been better characterized.

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

float DA=(5.0*PI/180.0);
#define HLXL	(15.0/2.0)
//#define HLXW	(7.0/2.0)

main(int argc, char *argv[]) {
EMData *model,*cyl,*out,*ccf,*best,*balt,*baz;
int nz,i,j,k,l,ps,ps2;
int ii,jj,kk,nn=0,kcyl=0;
float alt,az,dlaz,apix,x,y,z,s,bm,bs,sigallow=.5;
float *d,*d2,*dalt,*daz,*dt,*dt2,*dt3;
float *a[4],dl,filt=0.0;
char *intfile=NULL;

appinit(argc,argv);

if (argc<4) { 
	printf("helixhunter - the search for alpha helices\nUsage:\nhelixhunter <input> <output> <a/pix> [cyl] [int=<file>] [da=<angle>] [keep=<sigmamult>] [filt=<res>]\n");
	exit(0); 
}

for (i=4; i<argc; i++) {	
	if (strncmp(argv[i],"int=",4)==0) intfile=&argv[i][4];
	else if (strncmp(argv[i],"da=",3)==0) DA=atof(&argv[i][3])*PI/180.0;
	else if (strncmp(argv[i],"cyl",3)==0) kcyl=1;
	else if (strncmp(argv[i],"keep=",5)==0) sigallow=atof(&argv[i][5]);
	else if (strncmp(argv[i],"filt=",5)==0) filt=atof(&argv[i][5]);
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
} 

apix=atof(argv[3]);
if (apix<=0) { printf("A/pix must be >=0\n"); exit(1); }
if (apix>2) { printf("Helixhunter will not work properly if A/pix < 2\n"); exit(1); }
if (apix>1.5) { printf("Helixhunter works better if A/pix < 1.5\n"); }

model=new EMData;
if (model->readImage(argv[1],-1)) { printf("Error reading image\n"); exit(1); }
model->normalize();
nz=model->zSize();
if (nz!=model->xSize() || nz!=model->ySize()) {
	printf("The model MUST be in a square box (nx=ny=nz)\n");
	exit(1);
}

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

out=new EMData;
out->setSize(nz,nz,nz);
out->zero();

best=new EMData;
best->setSize(nz,nz,nz);
best->zero();

balt=new EMData;
balt->setSize(nz,nz,nz);
balt->zero();

baz=new EMData;
baz->setSize(nz,nz,nz);
baz->zero();

cyl=new EMData;
cyl->setSize(nz,nz,nz);

ps=(int)floor(HLXL/apix*2);
ps2=(int)floor(HLXL/apix);

//model->toCorner();

for (alt=0; alt<PI+DA/2.0; alt+=DA) {
	i=floor(sin(alt)*PI/DA);
	dlaz=PI/(float)(i+1);
	for (az=0; az<PI-dlaz/2; az+=dlaz) {
		// generate cylinder for ccf
		cyl->zero();
		d=cyl->getData();
		for (k=-ps; k<=ps; k++) {
			for (j=-ps; j<=ps; j++) {
				for (i=-ps; i<=ps; i++) {
					l=(i+nz/2)+(j+nz/2)*nz+(k+nz/2)*nz*nz;
					x= i*cos(az)         +j*sin(az);
					y=-i*sin(az)*cos(alt)+j*cos(alt)*cos(az)+k*sin(alt);
					z= i*sin(alt)*sin(az)-j*sin(alt)*cos(az)+k*cos(alt);
					dl=hypot(x*apix,y*apix);
					if (z>ps2) d[l]=exp(-(SQR(dl/6.0)+SQR((z-ps2)*apix/2.0)))*cos(dl*PI/5.0);
					else if (z<-ps2) d[l]=exp(-(SQR(dl/6.0)+SQR((z+ps2)*apix/2.0)))*cos(dl*PI/5.0);
					else d[l]=exp(-SQR(dl/6.0))*cos(dl*PI/5.0);
				}
			}
		}
		cyl->doneData();
		if (kcyl && az==0 && alt==0) cyl->writeMRC("cyl.mrc");
//		cyl->toCorner();

		
		
		// now do a CCF
		ccf=model->calcCCF(cyl);
		ccf->multConst(1.0e9);
		ccf->toCorner();
//		if (ccf->Max()<fabs(ccf->Min())) { ccf->multConst(-1.0); printf("Inv  "); }
//		else printf("     ");

//		char t[80];
//		sprintf(t,"ccf.%1.0f.%1.0f.mrc",alt*180/PI,az*180/PI);
//		cyl->writeMRC(t);

		// Now keep track of the best cylinder at each voxel
		d=ccf->getDataRO();
		d2=best->getData();
		dalt=balt->getData();
		daz=baz->getData();

		for (k=0; k<nz; k++) {
			for (j=0; j<nz; j++) {
				for (i=0; i<nz; i++) {					
					if (*d>*d2) {
						*d2=*d;
						*dalt=alt;
						*daz=az;
					}
					d++;
					d2++;
					dalt++;
					daz++;
				}
			}
		}
		ccf->doneData();
		best->doneData();
		balt->doneData();
		baz->doneData();

//		char t[80];
//		sprintf(t,"ccf.%d.mrc",nn++);
//		best->writeMRC(t);

		printf("%1.1f,%1.1f\tccf max=%g sig=%g\n",alt*180.0/PI,az*180.0/PI,ccf->Max(),ccf->Sigma());
//		printf(" %f\t%f\r",alt*180.0/PI,az*180.0/PI); fflush(stdout);
		delete ccf;
	}
}

printf("ccf max=%g sig=%g\n",best->Max(),best->Sigma());

best->multConst(1.0/best->Sigma());
if (filt>=1.0) model->filter(-5,apix*model->ySize()/filt,2);
best->add(model);
model->realFilter(2,best->Mean()+best->Sigma()*sigallow);
best->mult(model);

bm=best->Max();
bs=best->Sigma();

if (intfile) {
	best->writeMRC(intfile);
	LOG(Ref,intfile,LOG_OUTFILE,"intermediate");
}


// Now make a map out of the best cylinders
a[0]=out->getData();
a[1]=best->getDataRO();
a[2]=balt->getDataRO();
a[3]=baz->getDataRO();
for (s=bm-bs*8.0; s<=bm+bs/4.0; s+=bs/2.0) {
//for (s=bm-bs/2.0; s<=bm+bs/4.0; s+=bs/2.0) {
d=a[0];
d2=a[1];
dalt=a[2];
daz=a[3];
for (k=0; k<nz; k++) {
	for (j=0; j<nz; j++) {
		for (i=0; i<nz; i++) {
			// if we have a big peak, draw a line in the output file
			if (*d2>s) {
				dt3=NULL;
				for (dl=-ps/2; dl<=ps/2; dl+=.25) {
					ii=dl*sin(*dalt)*sin(*daz);
					jj=-dl*sin(*dalt)*cos(*daz);
					kk=dl*cos(*dalt);
					if (i+ii<0||i+ii>=nz || j+jj<0||j+jj>=nz || k+kk<0||k+kk>=nz) continue;
					dt2=d+(ii)+(jj)*nz+(kk)*nz*nz;
					if (dt2==dt3) continue;
					dt3=dt2;
					(*dt2)=s;
				}
			}
			d++;
			d2++;
			dalt++;
			daz++;
		}
	}
}
}
out->doneData();
best->doneData();
balt->doneData();
baz->doneData();

out->writeMRC(argv[2]);
LOG(Ref,argv[2],LOG_OUTFILE,NULL);
LOG(Ref,NULL,LOG_END,NULL);
}
