// proc2d.C 8/97  Steve Ludtke

//N proc2d
//F This program performs various processing operations on images and sets of images
//1
//P <input file>	Source file
//P <output file>	Destination file, appends unless 'inplace' specified
//P [first=<n>]	Specify the first image in the input to process (0 - n-1)
//P [last=<n>]	Specify the last image in the input to process
//P [inplace]	Output overwrites input, USE SAME FILENAME, DO NOT 'clip' images.
//P [edgenorm]	Output is always normalized, this will set 0 to the circular edge value rather than the overall mean
//P [nonorm]	Disable the automatic normalization.
//P [invert]	Inverts the density
//P [rot=<angle>]	Rotate by an angle (degrees)
//P [center]	Center image using center of mass
//P [scale=<sca>]	Scale by specified scaling factor, must also specify clip
//P [clip=<newsize x&y>]	Define the output image size (always square)
//P [shrink=<n>]	Reduce an image size by an integral scaling factor, uses median filter. Don't use clip.
//P [apix=<A/pix>]	Specify the Angstr/pixel for S scaling
//P [lp=<filt r>]	Gaussian lowpass filter, r in angstroms if apix defined, pixels if undef
//P [hp=<filt r>]	Arctan highpass filter
//P [mask=<radius>]	Circular mask
//P [imask=<radius>]	Circular inner mask
//P [noisemask]	Fill the masked region with 1-sigma noise rather than zero
//P [mrc]	Output a set of MRC files rather than a single IMAGIC file
//P [spider]	Output a Spider stack file rather than an IMAGIC file
//P [sharphp=<pixels>]	A highpass with a sharp cutoff (beware artifacts)
//P [norefs]	Skip any input images which are marked as references (usually used with classes.*)
//P [rotav]	Each image is rotationally averaged 
//P [split=n]	Splits the input file into a set of n output files
//P [interlv=<file 2 interleave>]	Specifies a 2nd input file. Output will be 2 files interleaved.
//P [sym=<Cn>]	Apply an n-fold rotational symmetry to each image
//P [rfp]	Experimental
//P [mraprep]	Experimental
//P [phot]	Experimental
//P [rsub]	Experimental
//U proc2d 3244.img start.hed invert apix=5 lp=18
//U proc2d start.hed start.hed invert inplace
//D IMPORTANT NOTE: proc2d normalizes output images by default. The images are adjusted so they
//E have a standard deviation of 1, and a mean of zero. Use the 'nonorm' flag to disable this.
//D This program is used to perform a variety of generic 2D image processing tasks. Like all of the
//E EMAN programs, it will read any supported file format. By default it outputs to an Imagic file although
//E there are options to change this. Note that, by default, images are appended to the end of the output
//E file if it already exists. All images in one file MUST have the same size. ie - if you try to copy 50x50
//E images to the end of a file that already contains 100x100 images, an error will result. Note the 'first='
//E and 'last=' options allow a subset of the input file to be processed. The 'inplace' keyword may be used
//E to cause output images to overwrite the input images rather than being appended to the end of the file.
//D Most of the processing is explained in sufficient detail above, but
//E there are a few notes:
//E <UL>
//E <LI> The scale operation does NOT resize the image in pixels. Generally you want to specify both scale and clip.
//E <LI> If apix is not specified, all filters are specified in pixels in Fourier space. If apix
//E is specified, all filters are specified in Angstroms. ie - 'apix=4.5 lp=8' would filter
//E the image using a Gaussian in Fourier space with a 1/e width of 1/8 reciprocal angstroms.
//E <LI> The 'center' option generally doesn't produce very good results on noisy images. Use the 'cenalignint'
//E program instead.
//E <LI> You might notice that 'mask'ing an image often produces a mask that doesn't match the density just inside
//E the mask very well. The 'edgemean' value normalizes the images so the mean value matches the edge of the image
//E rather than the overall mean value of the image.
//E </UL>

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

main(int argc, char *argv[]) {
float lp=0,hp=0,hp2=0,anoise=0,scale=1.0,rot=0,apix=-1.0;
int mask=0,noise=6,invert=0,center=0,rfp=0,split=1,norefs=0,phot=0,clip=-1,mrc=0,edgenorm=0;
int i,j,n0=0,n1=-1,rotav=0,shrink=1,rsub=0,imask=-1,csym=0,inpl=0,spider=0,mraprep=0,norm=1;
char outfile[1024],*il=NULL;
EMData *d,*e,*f;
float *dat;

appinit(argc,argv);

if (argc<3) { 
	printf("Preprocess images\nUsage: proc2d <input file> <output file> [mraprep] [center] [scale=<sca>]\n\
 [clip=<newsize x&y>] [invert] [rfp] [apix=<A/pix>] [lp=<pixels>] [hp=<pixels>] [sharphp=<pixels>] [shrink=<n>] [inplace]\n\
 [mask=<radius>] [imask=<radius>] [noisemask] [first=<n>] [last=<n>] [split=n] [mrc] [spider]\n\
 [norefs] [phot] [edgenorm] [nonorm] [rotav] [rsub] [interlv=<file 2 interleave>] [sym=<c?>] [rot=<angle>]\n");
	exit(0); 
}

for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"center",6)==0) center=1;
	else if (strncmp(argv[i],"invert",6)==0) invert=1;
	else if (strncmp(argv[i],"phot",4)==0) phot=1;
	else if (strncmp(argv[i],"apix=",5)==0) apix=atof(&argv[i][5]);
	else if (strncmp(argv[i],"lp=",3)==0) lp=atof(argv[i]+3);
	else if (strncmp(argv[i],"hp=",3)==0) hp=atof(argv[i]+3);
	else if (strncmp(argv[i],"sharphp=",8)==0) hp2=atof(argv[i]+8);
	else if (strncmp(argv[i],"rot=",4)==0) rot=atof(argv[i]+4);
	else if (strncmp(argv[i],"mask=",5)==0) mask=(int)atof(&argv[i][5]);
	else if (strncmp(argv[i],"imask=",6)==0) imask=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"first=",6)==0) n0=atoi(argv[i]+6);
	else if (strncmp(argv[i],"scale=",6)==0) scale=atof(argv[i]+6);
	else if (strncmp(argv[i],"last=",5)==0) n1=atoi(argv[i]+5);
	else if (strncmp(argv[i],"rfp",3)==0) rfp=1;
	else if (strncmp(argv[i],"noisemask",9)==0) noise=1;
	else if (strncmp(argv[i],"mraprep",7)==0) mraprep=1;
	else if (strncmp(argv[i],"addnoise=",9)==0) anoise=atof(&argv[i][9]);
	else if (strncmp(argv[i],"split=",6)==0) split=atoi(argv[i]+6);
	else if (strncmp(argv[i],"shrink=",7)==0) shrink=atoi(argv[i]+7);
	else if (strncmp(argv[i],"norefs",6)==0) norefs=1;	
	else if (strncmp(argv[i],"clip=",5)==0) clip=atoi(argv[i]+5);
	else if (strncmp(argv[i],"mrc",3)==0) mrc=1;
	else if (strncmp(argv[i],"spider",3)==0) spider=1;
	else if (strncmp(argv[i],"rotav",5)==0) rotav=1;	
	else if (strncmp(argv[i],"inplace",7)==0) inpl=1;	
	else if (strncmp(argv[i],"rsub",5)==0) rsub=1;	
	else if (strncmp(argv[i],"interlv=",8)==0) il=argv[i]+8;
	else if (strncmp(argv[i],"edgenorm",8)==0) edgenorm=1;
	else if (strncmp(argv[i],"nonorm",6)==0) norm=0;
	else if (strncmp(argv[i],"sym=",4)==0) {
		if (tolower(argv[i][4])!='c') { printf("Only C symmetry currently supported!\n"); exit(1); }
		csym=atoi(&argv[i][5]);
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

if (spider && mrc) { printf("Specify either Spider or MRC, not both\n"); exit(1); }

if (!norm && edgenorm) printf("Warning: nonorm and edgenorm specifcied\n");

if (clip>0 && !edgenorm) printf("Warning: suggest using edgenorm with clip\n");

if ((i=fileCount(argv[1]))<=n1 || n1<0) n1=i-1;

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

strcpy(outfile,argv[2]);

if (mraprep) {
	char ss[1024];
	for (i=0; i<10000; i++) {
		sprintf(ss,"%s%04d.hed",outfile,i);
		if (access(ss,F_OK)) break;
		unlink(ss);
		sprintf(ss,"%s%04d.img",outfile,i);
		unlink(ss);
	}
}

d=new EMData;
for (i=n0; i<=n1; i++) {
	if (split>1) {
		strcpy(outfile,argv[2]);
		outfile[strlen(outfile)-4]=0;
		sprintf(outfile+strlen(outfile),".%02d.img",i%split);
	}
	if ((i+1)%10==0) { printf("\t%d complete\r",i+1); fflush(stdout); }
	d->readImage(argv[1],i);
	if (d->Sigma()==0 || !finite(d->Sigma())) continue;
	if (norefs && d->NImg()<0) continue;
	if (norm) {
		if (edgenorm) d->edgeNormalize(1);
		else d->normalize();
	}
	if (invert) d->invert();

// Phil's Butterworth mod
//	if (lp) d->filter(hp,lp==0?100000.0:lp,1);
//	if (hp) d->filter(hp,0,3);
	int y=d->ySize();

	if (apix>0 && lp) lp=y*apix/lp;
	if (apix>0 && hp) hp=y*apix/hp;	
	if (apix>0 && hp2) hp2=y*apix/hp2;	

	if (lp || hp) d->filter(hp==0?-10.0:hp,lp==0?100000.0:lp,1);
	if (hp2) d->filter(hp2,100000.0,0);

	if (mask) d->applyMask(mask,noise);
		// uses correlation with 180 deg rot for centering
//	if (center) { d->transAlign(NULL,0); d->rotateAndTranslate(); }
		// uses center of mass to center
	if (imask>0) d->applyMask(imask,5);
	if (center) d->cmCenter(1); 
	if (phot) d->toCorner();
//	if (d->Sigma()==0 || !finite(d->Sigma())) continue;

	if (anoise) {
		dat=d->getData();
		for (j=0; j<d->xSize()*d->ySize(); j++) dat[j]+=grand(anoise,anoise/2);
		d->doneData();
	}

	if (rfp) { 
		e=d->makeRFP(); 
		e->writeIMAGIC("rfp.hed",-1); 
	}

	if (scale<1.0) { d->setRAlign(0,0,0); d->setTAlign(0,0,0); d->rotateAndTranslate(scale); }
	if (clip>0) {
		e=d->clip((d->xSize()-clip)/2,(d->ySize()-clip)/2,clip,clip);
		delete d;
		d=e;
	}
	if (scale>1.0) { d->setRAlign(0,0,0); d->setTAlign(0,0,0); d->rotateAndTranslate(scale); }

	if (shrink>1) d->medianShrink(shrink);

	if (rotav) {
		d->radialAverage();
	}

	if (rot!=0) {
		d->setParent(NULL);
		d->setRAlign(rot*PI/180.0,0,0);
		d->setTAlign(0,0,0);
		d->rotateAndTranslate();
	}

	if (csym>1) {
		f=d->copy();
		f->setParent(NULL);
		e=f->copy();
		for (i=1; i<csym; i++) {
			e->setRAlign(i*PI*2.0/(float)csym,0,0);
			e->rotateAndTranslate();
			d->add(e);
		}
		d->multConst(1.0/(float)csym);
		delete e;
		delete f;
	}

	if (rsub) {
		d->radialSubtract();
	}

	if (mrc) {
		// single images go to outfile
		if (n1==0) {
			d->writeMRC(outfile);
		}
		// multi-images get numbered
		else {
			sprintf(outfile,"%03d.%s",i+100,argv[2]);
			d->writeMRC(outfile);
		}
	}
	else if (spider) {
		d->writeSPIDER(outfile,-1);
	}
	else {
		if (inpl) d->writeIMAGIC(outfile,i);
		else if (mraprep) {
			char nf[1024];
			sprintf(nf,"%s%04d.hed",outfile,i);
			d->writeIMAGIC(nf,0);
		}
		else d->writeIMAGIC(outfile,-1);
	}

	// interleaved images not processed
	if (il) {
		d->readImage(il,i);
		d->writeIMAGIC(outfile,-1);
	}

//	if (rfp) delete e;	// auotmatically freed by emdata
}

LOG(Ref,argv[2],LOG_OUTFILE,NULL);
LOG(Ref,NULL,LOG_END,NULL);
printf("%d images\n",fileCount(argv[2]));
}
