// refine.C 1997  Steve Ludtke

//N refine
//F This is the main refinement loop in EMAN. It has a tremendous number of options.
//1
//P <total iter>	Total number of iterations to be run in the current directory
//P [proc=<min proc>,<maxproc>]	Number of proceesors to use for parallel processing
//P [hard=<maxpr>]	Hard limit for make3d
//P [ctfc=<res in A>]	Performs CTF correction filtered to the specified resolution
//P [median]	Specify this when CTF correction is NOT being performed
//P [sym=<c2,c4,etc>]	Imposes symmetry on the model
//P [ang=<dang>]	Angular step for projections (in degrees)
//P [mask=<rad>]	Applies an outside mask in several stages of refinement
//P [pad=<size>]	Pads the model during Fourier reconstruction
//P [classkeep=<sigma coef>]	Classkeep value for <a href=classalignall.html>classalignall</a>
//P [classiter=<iter>]	Iterations for <a href=classalignall.html>classalignall</a>
//P [filt3d=<lp rad>]	Apply a lowpass filter to the model after each iteration, this occurs before 3dit
//P [sep=<n>]	This option is passed to <a href=classesbymra.html>classesbymra</a>
//P [xfiles=<a/pix>,<mass in kd>,<ali to>]	Used to generate 3d models that are easy to evaluate
//P [3dit=<n it>]	After constructing a model with Fourier reconstruction, this runs several iterations of real-space model 'cleansing' 
//P [3dit2=<n it>]	This is real-space refinement with a special maximum entropy-like feature extraction method 
//P [sigfilt]	Applies a 1-sigma real-space filter to images at various points
//P [euler2=<oversmp>]	Reassigns Euler angles to class averages with an optionally reduced angular increment. ie - 2 will use ang/2
//P [tree=<2,3>]	Faster classifications for large numbers of projections
//P [precen]	Indicates that raw particles have been accurately precentered
//P [imask=<rad>]	Applies an inside mask in several stages of refinement
//P [goodbad]	Save files with good and bad particles, for diagnostic purposes
//P [shrink=<n>]	Experimental, shrinks images at several points for faster runs
//P [ca3]	Experimental. Uses classalign3 instead of classalign2 
//P [classfp=<msa#>]	Experimental, classification using rotational/translational invariants
//P [msa=<proj>]	Experimental
//P [mra2]	Experimental
//P [3dr[=<range>]]	Experimental
//P [d3s=<sigmult>]	Experimental
//U refine 5 ang=3 mask=28 pad=96 hard=25 classkeep=.3 classiter=8 dot proc=12 sym=d7 ctfc=12 sep=2 xfiles=4.2,1000,40 tree=3 3dit=5
//D This is the main refinement command in EMAN. You should have your starting
//E images in 'start.img' and your initial 3D model in threed.0a.mrc. The
//E 'refine' command will then refine this model iteratively. Note that the
//E 'total number of iterations' specifies the
//E number of iterations the program will finish with. If 5 iterations have
//E been run, then 'refine 8 ...' is run, the program will pick up after 5
//E iterations and run only 3 more cycles. If you want to start over again
//E from the first iteration, you must remove all of the 'threed.?.mrc'
//E files (except for 0a) and all of the 'classes.?.mrc' files.
//D Most of the options are simply passed through to other programs. Typical options
//E for this program can be determined by running <a href=eman.html>eman</a> and following the reconstruction
//E procedure it provides.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <wait.h>
#include "util.h"
#include "EMData.h"

void *cmra(void *fnum);

main(int argc, char *argv[]) {
int i,j,n;
int iter=5,restart=0,mask=-1,imask=-1,pad=0,classiter=8,d3r=0,classfp=0,goodbad=0,precen=0;
int minproc=0,maxproc=0,lp=0,alito=0,median=0,dot=1,mode=2,msa=0,shrink=1,mra2=0,can=2,sigfilt=0;
int noweight=0,ntk=1,tree=0,d3it=0,d3it2=0;
float ang=7.5,keep=1.2,xfiles=0,appix=0,d3s=0,hard=0,range=90.0,ctfc=0,euler2=0.0;
char s[1024],*sym=NULL;
FILE *log;
time_t ti;

appinit(argc,argv);

log=fopen("refine.log","a");

if (argc<2) { printf("Usage:\nPut initial model in 'threed.0a.mrc'. Particles should\n\
be in 'start.hed/img'.\n\nrefine <total iter> [proc=<nproc>] [hard=<maxpr>] [precen] [imask=<rad>]\
[sym=<c2,c4,etc>]\n\t[ang=<dang>] [mask=<rad>] [pad=<size>] [classkeep=<sigma coef>] [mra2]\n\
\t[classiter=<iter>] [filt3d=<lp rad>] [3dr[=<range>]] [d3s=<sigmult>] [median] [sep=<n>]\n\
\t[xfiles=<a/pix>,<mass>,<ali to>] [classfp=<msa#>] [msa=<proj>] [goodbad] [3dit=<n it>] [3dit2=<n it>]\n\
\t[mode=<1-6>] [ctfc=<res in A>] [shrink=<n>] [ca3] [sigfilt] [euler2] [tree=<2,3>]\n"); exit(0); }

int Ref=LOGbegin(argc,argv);
LOG(Ref,"start.hed",LOG_INFILE,NULL);
LOG(Ref,"refine.log",LOG_OUTFILE,NULL);

ti=time(0);
if (log) fprintf(log,"******** refine started %s",ctime(&ti));
for (i=0; i<argc; i++) fprintf(log,"%s ",argv[i]);
fprintf(log,"\n");

iter=atoi(argv[1]);

for (i=2; i<argc; i++) {
	if (strncmp("proc=",argv[i],5)==0) sscanf(&argv[i][5],"%d,%d",&minproc,&maxproc);
	else if (strncmp("tree=",argv[i],5)==0) tree=atoi(&argv[i][5]);
	else if (strncmp("mask=",argv[i],5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp("3dit=",argv[i],5)==0) d3it=atoi(&argv[i][5]);
	else if (strncmp("3dit2=",argv[i],6)==0) d3it2=atoi(&argv[i][6]);
	else if (strncmp("imask=",argv[i],6)==0) imask=atoi(&argv[i][6]);
	else if (strncmp("mode=",argv[i],5)==0) mode=atoi(&argv[i][5]);
	else if (strncmp("hard=",argv[i],5)==0) hard=atof(&argv[i][5]);
	else if (strncmp("classkeep=",argv[i],10)==0) keep=atof(&argv[i][10]);
	else if (strncmp("classiter=",argv[i],10)==0) classiter=atoi(&argv[i][10]);
	else if (strncmp("pad=",argv[i],4)==0) pad=atoi(&argv[i][4]);
	else if (strncmp("filt3d=",argv[i],7)==0) lp=atoi(&argv[i][7]);
	else if (strncmp("shrink=",argv[i],7)==0) shrink=atoi(&argv[i][7]);
	else if (strncmp("sigfilt",argv[i],7)==0) sigfilt=1;
	else if (strncmp("euler2",argv[i],6)==0) {
		if (argv[i][6]!='=') euler2=2.0;
		else euler2=atof(&argv[i][7]);
	}
	else if (strncmp("noweight",argv[i],8)==0) noweight=1;
	else if (strncmp("sep=",argv[i],4)==0) ntk=atoi(&argv[i][4]);
	else if (strncmp("ang=",argv[i],4)==0) ang=atof(&argv[i][4]);
	else if (strncmp("msa=",argv[i],4)==0) msa=atoi(&argv[i][4]);
	else if (strncmp("mra2",argv[i],4)==0) mra2=1;
	else if (strncmp("ca3",argv[i],3)==0) can=3;
	else if (strncmp("d3s=",argv[i],4)==0) d3s=atof(&argv[i][4]);
	else if (strncmp("3dr",argv[i],3)==0) {
		d3r=1;
		if (argv[i][3]=='=') range=atof(&argv[i][4]);
	}
	else if (strncmp("sym=",argv[i],4)==0) sym=&argv[i][4];
//	else if (strncmp("restart",argv[i],7)==0) restart=1;
	else if (strncmp("precen",argv[i],6)==0) precen=1;
	else if (strncmp("ctfc=",argv[i],5)==0) ctfc=atof(&argv[i][5]);
	else if (strncmp("goodbad",argv[i],7)==0) goodbad=1;
	else if (strncmp("classfp=",argv[i],8)==0) classfp=atoi(&argv[i][8]);
	else if (strncmp("dot",argv[i],3)==0) dot=1;
	else if (strncmp("median",argv[i],6)==0) median=1;
	else if (strncmp("xfiles=",argv[i],7)==0) {
		xfiles=atof(&argv[i][7]);
		sscanf(&argv[i][7]," %f , %f , %d",&appix,&xfiles,&alito);
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

if (minproc<=0) minproc=1;
if (maxproc<=minproc) maxproc=minproc;

if (xfiles) {
	sprintf(s,"proc3d threed.0a.mrc x.0.mrc clip=%d,%d,%d mask=%d",mask*2,mask*2,mask*2,mask);
	if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
	run(s);

	if (xfiles>0) {
		sprintf(s,"volume x.0.mrc %f set=%f",appix,xfiles);
		if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
		
		run(s);
	}
}


for (i=1; i<=iter; i++) {
	ti=time(0);
	if (log) { fprintf(log,"Iteration %d beginning at %s",i,ctime(&ti)); fflush(log); }
	
// project3d, classesbymra, classalignall
	sprintf(s,"classes.%d.hed",i);
	if (restart || access(s,R_OK)) {
		// project3d
		char *t = tmpnam(NULL);
		FILE *out = fopen(t,"w");
		if (!out) { printf("Cannot open temporary file!\n"); exit(1); }

		for (j=0; j<minproc; j++) {
			sprintf(s,"project3d threed.%da.mrc prop=%1.3f mask=%d mode=-1 frac=%d/%d",i-1,ang,mask,j,minproc);
			if (sym) sprintf(s+strlen(s)," sym=%s",sym);
			fprintf(out,"%s\n",s);
			if (log && !j) { ti=time(0); fprintf(log,"%s > %s  %s",s,t,ctime(&ti)); fflush(log); }
		}
		fclose(out);

		sprintf(s,"runpar proc=%d file=%s",minproc,t);
		run(s);

		if (classfp) {
			if (access("start.fp",R_OK)) { 
				if (log) { ti=time(0); fprintf(log,"footprintr start.hed start.fp   %s",ctime(&ti)); fflush(log); } 
				run("footprintr start.hed start.fp");
			}
			if (log) { ti=time(0); fprintf(log,"footprintr proj.hed proj.fp   %s",ctime(&ti)); fflush(log); } 
			run("footprintr proj.hed proj.fp");
			
//			run("msa2 proj.fp start.fp proj.msa.fp start.msa.fp %d 100",classfp);
			sprintf(s,"msa2 proj.fp start.fp proj.msa.fp start.msa.fp %d 100",classfp);
			if (log) { ti=time(0); fprintf(log,"%s    %s",s,ctime(&ti)); fflush(log); } 
			run(s);
//			if (log) { ti=time(0); fprintf(log,"msa2 proj.fp start.fp proj.msa.fp start.msa.fp 6 100  %s",ctime(&ti)); fflush(log); } 

			run("classesrefs start.msa.fp proj.msa.fp dot");
			if (log) { ti=time(0); fprintf(log,"classesrefs start.msa.fp proj.msa.fp   %s",ctime(&ti)); fflush(log); } 			
		}
		else {
			n=fileCount("start.hed");
			if (n/maxproc<50) n=n/maxproc;
			else n/=50;

			char *t = tmpnam(NULL);
			FILE *out = fopen(t,"w");
			if (!out) { printf("Cannot open temporary file!\n"); exit(1); }

			if (tree) {
				sprintf(s,"projtree proj.hed proj.tree level=%d",tree);
				if (sym) sprintf(s+strlen(s)," sym=%s",sym);
				run(s);
			}
			
			run("proc2d proj.hed cls mraprep");	//prepares references in cls files
			for (j=0; j<n; j++) {
				sprintf(s,"classesbymra  start.hed proj.hed split frac=%d/%d dot mask=%d imask=%d shrink=%d%s%s%s%s%s",
					j,n,mask,imask,shrink,precen?" precen":"",
					j==0?"":" quiet",can==3?" ca3":"",sigfilt?" sigfilt":"",ctfc?" ctfc":"");
				if (ntk) sprintf(s+strlen(s)," sep=%d",ntk);
				if (tree) strcat(s," tree1=proj.tree");
				else if (mra2) s[12]='2';
				fprintf(out,"%s\n",s);
				if (log && !j) { ti=time(0); fprintf(log,"%s > %s  %s",s,t,ctime(&ti)); fflush(log); }
			}
			fclose(out);
			
			sprintf(s,"runpar proc=%d,%d file=%s",minproc,maxproc,t);
			run(s);
 			if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
			unlink(t);
			
			out = fopen(t,"w");
			if (!out) { printf("Cannot open temporary file!\n"); exit(1); }

			if (tree) {
				int fn;
				FILE *tree=fopen("proj.tree","r");
				if (!tree) { printf("Tree file vanished !\n"); exit(1); }

				while (fgets(s,256,tree)) {
					if (sscanf(s," %d",&fn)!=1) continue;			
					sprintf(s,"classesbymra  tcls%04d.hed proj.hed split dot mask=%d imask=%d shrink=%d%s%s%s%s%s",
						fn,mask,imask,shrink,precen?" precen":"",
						j==0?"":" quiet",can==3?" ca3":"",sigfilt?" sigfilt":"",ctfc?" ctfc":"");
					strcat(s," tree2=proj.tree");
					fprintf(out,"%s\n",s);
					if (log && !j) { ti=time(0); fprintf(log,"%s > %s  %s",s,t,ctime(&ti)); fflush(log); }
				}
				fclose(tree);
				fclose(out);
			
				sprintf(s,"runpar proc=%d,%d file=%s",minproc,maxproc,t);
				run(s);
	 			if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
				unlink(t);
			}
		}
		
		sprintf(s,"classalignall ref1 finalref saverefs keep=%1.2f %d proc=%d,%d mask=%d imask=%d %s%s%s",
			keep,classiter,minproc,maxproc,mask,imask,median?"median":"",dot?" dot":"",precen?" precen":"");
		if (ctfc!=0) sprintf(s+strlen(s)," ctfc=%f",ctfc);
		if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
		run(s);

		sprintf(s,"classes.%d.img",i);
		rename("classes.img",s);
		sprintf(s,"classes.%d.hed",i);
		rename("classes.hed",s);
		if (log) fprintf(log,"mv classes.hed/img to %s/img\n",s);
	}

// make3d
	if (!restart) {
		sprintf(s,"threed.%da.mrc",i);
		if (!access(s,R_OK)) { 
			if (log) { fprintf(log,"Iteration %d already complete\n",i); fflush(log); }
			continue;
		}
	}

// This reassigns all of the euler angles
// only use 5 threads at most
	if (euler2>0) {
	
		// first, reproject the model with finer spacing
		sprintf(s,"project3d threed.%da.mrc prop=%1.3f mask=%d mode=-1",i-1,ang/euler2,mask);
		if (sym) sprintf(s+strlen(s)," sym=%s",sym);
		run(s);

		char *t = tmpnam(NULL);
		FILE *out = fopen(t,"w");
		if (!out) { printf("Cannot open temporary file!\n"); exit(1); }
		
		for (j=0; j<maxproc; j++) {
			sprintf(s,"eulerproj classes.%d.hed proj.hed writeali frac=%d/%d mask=%d %s %s",i,j,maxproc,mask,precen?" precen":"",sigfilt?"sigfilt":"");
			fprintf(out,"%s\n",s);
			if (log && !j) { ti=time(0); fprintf(log,"%s > %s  %s",s,t,ctime(&ti)); fflush(log); }
		}
		fclose(out);
			
		sprintf(s,"runpar proc=%d,%d file=%s",minproc,maxproc,t);
		run(s);
		if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
		unlink(t);
	}

	if (msa) {
		sprintf(s,"cloud classes.%d.hed classes.msa.hed %d 1",i,msa);
		if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
		run(s);
	}


	if (d3r) sprintf(s,"make3dr classes.%d.%shed out=threed.%d.mrc mask=%d range=%f %s",i,msa?"msa.":"",i,mask,range,goodbad?"goodbad":"");
	else sprintf(s,"make3d classes.%d.%shed out=threed.%d.mrc mask=%d",i,msa?"msa.":"",i,mask);
	if (hard) sprintf(s+strlen(s)," hard=%1.2f",hard);
	if (d3s) sprintf(s+strlen(s)," keep=%1.2f",d3s);
	if (pad) sprintf(s+strlen(s)," pad=%d",pad);
	if (sym) sprintf(s+strlen(s)," sym=%s",sym);
	if (noweight) sprintf(s+strlen(s)," noweight");
	if (mode) sprintf(s+strlen(s)," mode=%d",mode);
	if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
	run(s);

// proc3d
	sprintf(s,"proc3d threed.%d.mrc threed.%da.mrc norm mask=%d",i,i,mask);
	if (lp) sprintf(s+strlen(s)," lp=%d",lp);
	if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
	run(s);
	
// make3diter if selected
	if (d3it || d3it2) {
		char *t = tmpnam(NULL);
		char *t2 = tmpnam(NULL);
		FILE *out = fopen(t,"w");
		FILE *out2 = fopen(t2,"w");
		if (!out) { printf("Cannot open temporary file!\n"); exit(1); }
		
		for (j=0; j<minproc; j++) {
			fprintf(out,"make3diter threed.%da.mrc classes.%d.hed threed.%da.mrc tozero frac=%d/%d\n",i,i,i,j,minproc);
			fprintf(out2,"make3diter threed.%da.mrc classes.%d.hed threed.%da.mrc frac=%d/%d\n",i,i,i,j,minproc);
			if (log && !j) { ti=time(0); fprintf(log,"%s > %s  %s",s,t,ctime(&ti)); fflush(log); }
		}
		fclose(out);
		fclose(out2);
			
		for (j=0; j<d3it2; j++) {
			sprintf(s,"runpar proc=%d file=%s",minproc,t);
			run(s);

			if (sym) {
				sprintf(s,"proc3d threed.%da.mrc threed.%da.mrc sym=%s",i,i,sym);
				run(s);
			}
		}

		for (j=0; j<d3it; j++) {
			sprintf(s,"runpar proc=%d file=%s",minproc,t2);
			run(s);

			if (sym) {
				sprintf(s,"proc3d threed.%da.mrc threed.%da.mrc sym=%s",i,i,sym);
				run(s);
			}
		}
		sprintf(s,"make3diter threed.%da.mrc classes.%d.hed threed.%da.mrc testonly",i,i,i);
		run(s);

		unlink(t);
		unlink(t2);
	}

// x-files
	if (xfiles) {
		sprintf(s,"proc3d threed.%da.mrc x.%d.mrc clip=%d,%d,%d",i,i,mask*2,mask*2,mask*2);
		if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
		run(s);

		if (i>alito) {
			sprintf(s,"align3d x.%0d.mrc x.%d.mrc x.%d.mrc",alito,i,i);
			if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
			run(s);
		}
		else if (log) fprintf(log,"Alignment skipped\n");
		
		if (xfiles>0) {
			sprintf(s,"volume x.%d.mrc %f set=%f",i,appix,xfiles);
			if (log) { ti=time(0); fprintf(log,"%s   %s",s,ctime(&ti)); fflush(log); }
			run(s);
		}
	}

	
}

if (log) { ti=time(0); fprintf(log,"%s  Refine complete\n",ctime(&ti)); }
fclose(log);
LOG(Ref,NULL,LOG_END,NULL);
}

void *cmra(void *s)
{
char ss[1024];

strcpy(ss,(char *)s);
run(ss);
return NULL;
}


