// runpar.C Copyright 1999 Steve Ludtke
// This will run a list of programs in parallel on a specified number of
// processors.

//N runpar
//F Run a list of programs in parallel on one or several machines
//1
//P [proc=<min>[,<max>]]	Number of processors to use
//P [file=<command file>]	File containing list of programs to run
//P [test]	This will test all of the listed processors
//U runpar test
//U runpar file=script proc=6,12
//D This program is for coarse grained parallelization on virtually any
//E platform. It permits parallel execution on a network of workstations or
//E a single multiprocessor machine. This was specifically done in a way 
//E to ensure maximum portability,
//E without requiring the installation of a large parallel processing package
//E (like MPI). For this to work, the user must have his account(s) set up so
//E rsh can execute programs on all of the referenced machines. Each machine
//E must also have remote access to the run directory, although this may be a
//E different location on each machine.
//D The machine parameters are specified in a file called .mparm located in
//E the directory runpar is run from. If this file does not exist, dynamic load
//E usage (min and max processors) will not function, and only 'min' procssors
//E will be used.
//D The .mparm file is 5 tab separated columns, with one machine specified per line.
//E This is a sample entry for one machine:
//D rsh 2 1 localhost /homes/stevel/tst
//D The first column is always 'rsh'. Column 2 specifies the total number of procssors
//E on the machine. Column 3 is a relative speed factor (currently unused). Column
//E 4 is the machine name. 'localhost' is a special machine name, and should
//E always be present. The final column is the path to the run directory on that
//E machine.
//D For example, when running on a single 8 processor machine, the file would contain
//E one line:
//D rsh 8 1 localhost /homes/stevel/tst
//D If there are 4 workstations named alpha, beta, gamma and delta, each with one
//E processor, with the job being submitted on alpha, the file might contain:
//D rsh 1 1 localhost /homes/stevel/test
//D rsh 1 1 beta /hosts/alpha/disk1/stevel/test
//D rsh 1 1 gamma /hosts/alpha/disk1/stevel/test
//D rsh 1 1 delta /hosts/alpha/disk1/stevel/test
//D When the job is submitted, the least loaded machines will be chosen to run the
//E specified jobs. The 'min' number of processors will be used at all times, regardless
//E of load. If enough unloaded processors are available, runpar will slowly begin to use more
//E processors until 'max' are used. Periodically, the usage is reduced back to 'min', to
//E allow other users jobs to take precedence. If the machine becomes loaded during
//E this time, the load will shift to another processor if available.
//D 'runpar test' can be used to test the .mparm file, and the configuration of the individual
//E machines. This will also give an indication of the current load levels on each machine
//E as determined by runpar.
//D Note that this program can be used for parallel submission of any jobs, but it supports
//E only simple coarse grained parallelization with no inter-process communication.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <pthread.h>
#include <ctype.h>
#include <wait.h>
#include <string.h>
#include <util.h>

//#define DEBUG


int stl=0;

enum { MODE_RSH,MODE_SSH };
enum { T_INIT,T_RUN,T_DONE,T_FAIL };

// machine parameters
struct MPARM {
int mode;
int maxproc;
int speed;
float load;
int jobs;
char name[80];
char path[1024];
};

MPARM mparm[100];	// should probably be dynamic
int nparm = 0;

struct CPARM {
char *command;
int pid;
int mnum;
int status;
int jnum;
};

void update_load();
int do_run(CPARM *cp);
int pick_proc();
int avail_proc();

main(int argc, char *argv[]) {
int i,j,minproc=1,maxproc=0,test=0;
char *file=NULL;
pthread_t *thr;
CPARM *cparm;
FILE *in;
char s[1065],t[40];
char com[1024];

appinit(argc,argv);

if (argc<2) { printf("runpar usage:\nrunpar proc=<min>[,<max>] [file=<command file>] [test]\n"); exit(1); }

for (i=1; i<argc; i++) {
	if (strncmp("proc=",argv[i],5)==0) sscanf(&argv[i][5]," %d , %d",&minproc,&maxproc);
	else if (strncmp("file=",argv[i],5)==0) file=&argv[i][5];
	else if (strncmp("test",argv[i],5)==0) test=1;
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}
if (!maxproc) maxproc=minproc;


if (!test && !file) { printf("Please specify a command file or 'test'\n"); exit(1); }
if (test && file) { printf("Please specify only one of 'test' or 'file'\n"); exit(1); } 

in=fopen(".mparm","r");
if (!in) {
	mparm[0].mode=MODE_RSH;
	mparm[0].maxproc=1024;
	mparm[0].speed=1;
	mparm[0].jobs=0;
	strcpy(mparm[0].name,"localhost");
	strcpy(mparm[0].path,getenv("PWD"));
	nparm=1;
}
else {
	while (fgets(s,1064,in)) {
		sscanf(s," %s %d %d %s %s",t,&mparm[nparm].maxproc,&mparm[nparm].speed,
			mparm[nparm].name,mparm[nparm].path);
		mparm[nparm].jobs=0;
		if (!strcmp(t,"rsh")) mparm[nparm].mode=MODE_RSH;	
		else { printf("Unknown mode '%s'\n",t); exit(1); }
		nparm++;
	}
	fclose(in);
}

update_load();

if (test) {
	int a=0,t=0;
	for (i=0; i<nparm; i++) {
		j=mparm[i].maxproc-(int)floor(mparm[i].load+.4);
		if (j<0) j=0;
		printf("%s\tload=%1.2f\tavail=%d\n",mparm[i].name,mparm[i].load,j);
		a+=j;
		t+=mparm[i].maxproc;
	}
	printf("\n%d processors available of %d on %d machines\n",a,t,nparm);
	exit(0);
}

thr=(pthread_t *)malloc(sizeof(pthread_t)*maxproc);
cparm=(CPARM *)malloc(sizeof(CPARM)*maxproc);

for (i=0; i<maxproc; i++) {
	cparm[i].command=NULL;
	cparm[i].status=T_INIT;
}

in=fopen(file,"r");
if (!in) { printf("cannot open command file\n"); exit(1); }

int Ref=LOGbegin(argc,argv);

int running=0;
int lrun=0;
int finished=0;
int rising=1;		// periodically uses maxproc, then drifts down to min and back up
int count=0;
int job=1;
int mn;

while (1) {
	int stat;
	int spid=waitpid(-1,&stat,WNOHANG);
	
	running=0;
	for (i=0; i<maxproc; i++) {
		if (cparm[i].status==T_RUN) {
			running++;
			mn=cparm[i].mnum;
			if (spid==cparm[i].pid) {
				mparm[mn].load-=1.0;
				mparm[mn].jobs--;
				if (!WIFEXITED(stat)) {
					printf("Error, running '%s' on %s\n",cparm[i].command,mparm[cparm[i].mnum].name);
					exit(1);
				}
				else {
					cparm[i].status=T_INIT; printf("Job %d(%d) complete\n",spid,i);
				}
			}
		}
	}

	if (finished && running==0) break;

#ifdef DEBUG
	printf("%d. running=%d %d-%d finished=%d 0=%1.1f 1=%1.1f\n",count,running,minproc,maxproc,finished,mparm[0].load,mparm[1].load);
#endif
	
	if (!finished && (running<minproc || (running<maxproc && rising==1 && avail_proc()))) {
		for (i=0; i<maxproc; i++) { 
			if (cparm[i].status==T_INIT) break; 
		}
		if (i<maxproc) {
			if (!fgets(com,1023,in)) { finished=1; continue; }
			com[strlen(com)-1]=0;
			cparm[i].mnum=mn=pick_proc();
			cparm[i].command=(char *)realloc(cparm[i].command,strlen(com)+1);
			strcpy(cparm[i].command,com);
			cparm[i].jnum=job;
			job++;
			cparm[i].status=T_RUN;
			cparm[i].pid=do_run(&cparm[i]);
			printf("(%s) Start process %d(%d)\n",mparm[mn].name,cparm[i].pid,i);
		}
	}
	
	if (rising==1 && (avail_proc()==0 || running==maxproc)) rising=-20;		// ~20 second delay before we start going up again
	if (rising!=1 && running<=minproc) rising++;
	
	sleep(1);	
	count++;
	if (count%300==0) update_load();		// update the loads every 5 minutes
//	if (count%20==0 && lrun!=running) { lrun=running; printf("Running on %d processors\n",running); }
	if (lrun!=running) { lrun=running; printf("Running on %d processors\n",running); }
}

printf("#### RUNPAR COMPLETE\n");
LOG(Ref,NULL,LOG_END,NULL);

}

// Number of processors currently available
int avail_proc() {
int a=0,i,j;
for (i=0; i<nparm; i++) {
	j=mparm[i].maxproc-(int)floor(mparm[i].load+.4);
	if (j<0) j=0;
	a+=j;
}
return a;
}

// selects a processor to use for the next thread
int pick_proc() {
int i,l;
float b;

// first look for a machine that definitely isn't overloaded
for (i=0; i<nparm; i++) {
//	if (mparm[i].load <0) continue;
	l=(int)floor(mparm[i].load+.4);
	if (l<mparm[i].maxproc) break;
}

if (i!=nparm) return i;

// if there aren't any, find the machine that's the least overloaded
b=mparm[0].maxproc-mparm[0].load;
l=0;
for (i=1; i<nparm; i++) {
	if (mparm[i].load <0) continue;
	if (mparm[i].maxproc-mparm[0].load>b) { b=mparm[i].maxproc-mparm[0].load; l=i; }
}

// if the machine is overloaded, but not by us alone, use it
if (mparm[l].jobs<mparm[l].maxproc) return l;

// otherwise look for one that WE haven't overloaded
for (i=0; i<nparm; i++) {
	if (mparm[i].jobs<mparm[i].maxproc) break;
}

// if there aren't any, the user made a mistake, but go ahead and use one
if (i==nparm) {
	printf("Overloading %s with %1.1f jobs\n",mparm[l].name,mparm[l].load+1.0);
	return l;
}	

printf("Possibly verloading %s with %1.1f jobs\n",mparm[i].name,mparm[i].load+1.0);
return i;
}

// This starts the individual process
int do_run(CPARM *cp) {
int mn = cp->mnum;
char *c = (char *) malloc(strlen(cp->command)+256);
pid_t f=fork();

if (f) {
	mparm[mn].jobs++;
	mparm[mn].load+=1.0;
	return f;
}


if (!strcmp(mparm[mn].name,"localhost")) execl("/bin/sh","/bin/sh","-c",cp->command,0);

sprintf(c,"rsh %s 'cd %s; %s'",mparm[mn].name,mparm[mn].path,cp->command);
printf("%s\n",c);
execl("/bin/sh","/bin/sh","-c",c,0);

return 0;
}

// Updates the machine load values for each machine in the list
void update_load() {
int i;
float l,m;
char s[120],*t;
FILE *in;

#ifdef DEBUG
printf("update_load\n");
#endif

for (i=0; i<nparm; i++) {
	if (!strcmp(mparm[i].name,"localhost")) strcpy(s,"uptime");
	else sprintf(s,"rsh %s uptime",mparm[i].name);
	in=popen(s,"r");
	if (!in) { printf("%s is unavailable.\n",mparm[i].name); l=m=1.0e5; }
	fgets(s,119,in);
	pclose(in);
	t=strstr(s,"average:");
	if (!t) { printf("%s is unavailable.\n",mparm[i].name); l=m=1.0e5; }
	else sscanf(t+8," %f , %f",&m,&l);
	mparm[i].load=l>m?l:m;
//	mparm[i].load=l;
}

}
