// ttest.C Copyright 1998  Steve Ludtke
// This program splits the data in half and does 2 independent
// reconstructions. This is used to test the resolution of
// a reconstruction.

//N ttest
//F This program performs a 2 way T-test to determine the resolution of a reconstruction.
//D This program's options parallel <a href=refine.html>refine</a>. 'ttest' options are a
//E subset of refine's options. ttest is used to perform a fairly robust resolution
//E determination on a reconstruction. It is run in the same directory as a completed
//E refine command. Note that the refine command MUST terminate normally. Chances are
//E good that if refine is interrupted in the middle of a run, ttest will not produce
//E valid results. ttest splits each cls file into even and odd halves, then proceeds to
//E build an independent model based on each half of the data. By calculating a Fourier
//E shell correlation between the two resulting maps, a fairly robust resolution determination
//E can be made. For a proper determination, the data should be split into 2, 4, 8, etc sets,
//E then the resolution can be determined by interpolating to the entire data set. The
//E estimate made by comparing 2 halves of the data is actually slightly underestimating
//E the final resolution. Due to the rapid nonlinear falloff in signal to noise ratio
//E at resolution limiting spatial frequencies, this underestimation should be
//E relatively slight in most cases.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <pthread.h>
#include <ctype.h>
#include "util.h"
#include "EMData.h"

void *cmra(void *fnum);

main(int argc, char *argv[]) {
int i,j;
int restart=0,mask=-1,imask=-1,pad=0,classiter=8,d3r=0,precen=0;
int proc=1,median=0,dot=0,mode=2,mra2=0;
float ang=7.5,keep=1.2,d3s=0,hard=0,range=90.0,ctfc=0;
char s[1024],*sym=NULL;
pthread_t *thr;
FILE *log;
time_t t;

appinit(argc,argv);

log=fopen("refine.log","a");

if (argc<2) { printf("Usage:\nttest [proc=<nproc>] [hard=<maxpr>] [precen] [imask=<rad>]\
[sym=<c2,c4,etc>]\n\t[mask=<rad>] [pad=<size>] [classkeep=<sigma coef>]\n\
\t[classiter=<iter>] [3dr[=<range>]] [d3s=<sigmult>] [median]\n\
\t[mode=<1-6>] [dot] [ctfc=<res in A>]\n"); exit(0); }

int Ref=LOGbegin(argc,argv);
LOG(Ref,"start.hed",LOG_INFILE,NULL);
LOG(Ref,"refine.log",LOG_OUTFILE,NULL);

t=time(0);
if (log) fprintf(log,"******** refine started %s",ctime(&t));
for (i=0; i<argc; i++) fprintf(log,"%s ",argv[i]);
fprintf(log,"\n");

for (i=1; i<argc; i++) {
	if (strncmp("proc=",argv[i],5)==0) proc=atoi(&argv[i][5]);
	else if (strncmp("mask=",argv[i],5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp("imask=",argv[i],6)==0) imask=atoi(&argv[i][6]);
	else if (strncmp("mode=",argv[i],5)==0) mode=atoi(&argv[i][5]);
	else if (strncmp("hard=",argv[i],5)==0) hard=atof(&argv[i][5]);
	else if (strncmp("classkeep=",argv[i],10)==0) keep=atof(&argv[i][10]);
	else if (strncmp("classiter=",argv[i],10)==0) classiter=atoi(&argv[i][10]);
	else if (strncmp("pad=",argv[i],4)==0) pad=atoi(&argv[i][4]);
	else if (strncmp("d3s=",argv[i],4)==0) d3s=atof(&argv[i][4]);
	else if (strncmp("3dr",argv[i],3)==0) {
		d3r=1;
		if (argv[i][3]=='=') range=atof(&argv[i][4]);
	}
	else if (strncmp("sym=",argv[i],4)==0) sym=&argv[i][4];
	else if (strncmp("precen",argv[i],6)==0) precen=1;
	else if (strncmp("ctfc=",argv[i],5)==0) ctfc=atof(&argv[i][5]);
	else if (strncmp("dot",argv[i],3)==0) dot=1;
	else if (strncmp("median",argv[i],6)==0) median=1;
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

thr=(pthread_t *)malloc(sizeof(pthread_t)*proc);

t=time(0);
if (log) { fprintf(log,"Beginning T-test %s\n",ctime(&t)); fflush(log); }

// align even
sprintf(s,"classalignall even ref1 finalref saverefs keep=%1.2f %d proc=%d mask=%d imask=%d %s%s%s",
	keep,classiter,proc,mask,imask,median?"median":"",dot?" dot":"",precen?" precen":"");
if (ctfc!=0) sprintf(s+strlen(s)," ctfc=%f",ctfc);
if (log) { t=time(0); fprintf(log,"%s   %s",s,ctime(&t)); fflush(log); }
system(s);

rename("classes.img","classes.te.img");
rename("classes.hed","classes.te.hed");
if (log) fprintf(log,"mv classes.hed/img to classes.te.hed/img\n");


// align odd
sprintf(s,"classalignall odd ref1 finalref saverefs keep=%1.2f %d proc=%d mask=%d imask=%d %s%s%s",
	keep,classiter,proc,mask,imask,median?"median":"",dot?" dot":"",precen?" precen":"");
if (ctfc!=0) sprintf(s+strlen(s)," ctfc=%f",ctfc);
if (log) { t=time(0); fprintf(log,"%s   %s",s,ctime(&t)); fflush(log); }
system(s);

rename("classes.img","classes.to.img");
rename("classes.hed","classes.to.hed");
if (log) fprintf(log,"mv classes.hed/img to classes.to.hed/img\n");


// make3d

if (d3r) sprintf(s,"make3dr classes.te.hed out=threed.te.mrc mask=%d range=%f",mask,range);
else sprintf(s,"make3d classes.te.hed out=threed.te.mrc mask=%d",mask);
if (hard) sprintf(s+strlen(s)," hard=%1.2f",hard);
if (d3s) sprintf(s+strlen(s)," keep=%1.2f",d3s);
if (pad) sprintf(s+strlen(s)," pad=%d",pad);
if (sym) sprintf(s+strlen(s)," sym=%s",sym);
if (mode) sprintf(s+strlen(s)," mode=%d",mode);
else sprintf(s+strlen(s)," mode=2");
if (log) { t=time(0); fprintf(log,"%s   %s",s,ctime(&t)); fflush(log); }
system(s);

if (d3r) sprintf(s,"make3dr classes.to.hed out=threed.to.mrc mask=%d range=%f",mask,range);
else sprintf(s,"make3d classes.to.hed out=threed.to.mrc mask=%d",mask);
if (hard) sprintf(s+strlen(s)," hard=%1.2f",hard);
if (d3s) sprintf(s+strlen(s)," keep=%1.2f",d3s);
if (pad) sprintf(s+strlen(s)," pad=%d",pad);
if (sym) sprintf(s+strlen(s)," sym=%s",sym);
if (mode) sprintf(s+strlen(s)," mode=%d",mode);
else sprintf(s+strlen(s)," mode=2");
if (log) { t=time(0); fprintf(log,"%s   %s",s,ctime(&t)); fflush(log); }
system(s);

system("proc3d threed.te.mrc threed.to.mrc fsc=fsc.ttest");
if (log) { t=time(0); fprintf(log,"%s  proc3d threed.te.mrc threed.to.mrc fsc=fsc.ttest\n",ctime(&t)); }

if (log) { t=time(0); fprintf(log,"%s  T-test complete\n",ctime(&t)); }
fclose(log);
LOG(Ref,NULL,LOG_END,NULL);
}

void *cmra(void *s)
{
char ss[1024];

strcpy(ss,(char *)s);
system(ss);
return NULL;
}
